(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion

    /**
     * Static Singleton DatePickerController
     */
    var DatePickerController = function(){
        var self = this;
        //the picker UI
        var picker;
        //the id for the current culture
        var cultureID;
        //the context for the current active control element
        var ctx;
        //current calendar display mode
        var viewMode = 'days';
        //the date object we can use to manipulate
        var viewDate = new Date(0);
        var modes = {
            days:{navField: 'Month', navStep: 1, navInc:7, out:'months', inp:'days'},
            months:{navField: 'FullYear', navStep: 1, navInc:4, out:'years', inp:'days'},
            years:{navField: 'FullYear', navStep: 10, navInc:4, out:'years', inp:'months'}
        };
        //culture shortcut
        var dates = {};
        var template = '' +
            '<header class="dialog-header"><span class="date-time-header"></span></header>' +
            '<section class="dialog-content">' +
            '<div class="datepicker-header" tabindex="0">' +
            '<table class="dp table-condensed ignore">' +
            '<thead><tr><th class="dp prev"><button class="no-face icon icon-c no-text i-left" data-action="date:prev"></button></th><th colspan="5" class="dp switch"><button class="picker-switch" data-action="date:switch"></button></th><th class="dp next"><button class="no-face no-text icon icon-c i-right" data-action="date:next"></button></th></tr></thead>' +
            '</table>' +
            '</div>' +
            '<div class="datepicker-days" tabindex="0">' +
            '<table class="dp table-condensed ignore">' +
            '<thead>[[DAYS_HEADER]]</thead>' +
            '<tbody></tbody>' +
            '</table>' +
            '</div>' +
            '<div class="datepicker-months" tabindex="0">' +
            '<table class="dp table-condensed ignore">' +
            '<tbody><tr><td class="dp" colspan="7">[[MONTHS]]</td></tr></tbody>' +
            '</table>' +
            '</div>' +
            '<div class="datepicker-years" tabindex="0">' +
            '<table class="dp table-condensed ignore">' +
            '<tbody><tr><td class="dp" colspan="7"></td></tr></tbody>' +
            '</table>' +
            '</div>' +
            '</section>' +
            '<footer class="dialog-footer"><button class="btn-dialog" data-action="dialog-close">Clear</button><button class="btn-dialog margin-l-a" data-action="dialog-close">Done</button></footer>';

        /**
         *
         * @param id
         * @returns {string}
         */
        function getTemplatePart(id){
            var html = '';
            switch(id){
                case 'daysHeader':
                    for (var i = dates.firstDay; i < dates.firstDay + 7;i++) {
                        html += '<th class="dp dow">' + dates.days[(i) % 7] + '</th>';
                    }
                    html = '<tr>' + html + '</tr>';
                    break;
                case 'months':
                    for (var i=0;i<12;i++) {
                        html += '<span class="dp month" data-key="'+(i+1)+'">'+dates.monthsShort[i]+'</span>';
                    }
                    break;
                default:
                    break;
            }
            return html;
        }

        /**
         * Initialize the controller
         */
        function init(){
            //create the global picker
            setupCulture('default');
            //listen for internal system events
            System.inotify.add(notifyHandler);
        }

        /**
         * Display the picker
         */
        function showPicker() {
            //set the viewDate we will use this for all date calculations
            if (ctx && (ctx.el.dataset.dtm === 'datetime.CF') && (u.obs(ctx.datetime).indexOf('/') > 0 || u.obs(ctx.datetime).indexOf('.') > 0 || u.obs(ctx.datetime).indexOf('-') > 0)) {
                //CF Only
                // CF dates are stored in US format MM/DD/YYYY
                if (dates.format.substring(0, 3) === 'd/M' || dates.format.substring(0, 4) === 'dd/M') {
                    viewDate.setTime(Date.parse(Date.getISODate(ctx.datetime.split('/')[0], ctx.datetime.split('/')[1], ctx.datetime.split('/')[2].substring(0, 4)) || Date.today.toISOString()));
                } else if (dates.format.substring(0, 3) === 'd.M' || dates.format.substring(0, 4) === 'dd.M') {
                    viewDate.setTime(Date.parse(Date.getISODate(ctx.datetime.split('.')[0], ctx.datetime.split('.')[1], ctx.datetime.split('.')[2].substring(0, 4)) || Date.today.toISOString()));
                } else if (dates.format.substring(0, 3) === 'd-M' || dates.format.substring(0, 4) === 'dd-M') {
                    viewDate.setTime(Date.parse(Date.getISODate(ctx.datetime.split('-')[0], ctx.datetime.split('-')[1], ctx.datetime.split('-')[2].substring(0, 4)) || Date.today.toISOString()));
                } else if (dates.format.substring(0, 5) === 'yyyy/') {
                    var justDate = ctx.datetime.split('T')[0];
                    viewDate.setTime(Date.parse(Date.getISODate(justDate.split('/')[2], justDate.split('/')[1], justDate.split('/')[0]) || Date.today.toISOString()));
                } else {
                    viewDate.setTime(Date.parse(Date.getISODate(ctx.datetime.split('/')[1], ctx.datetime.split('/')[0], ctx.datetime.split('/')[2].substring(0, 4)) || Date.today.toISOString()));
                }
            } else{
                //Non CF
                viewDate.setTime(Date.parse(ctx.datetime || Date.today.toISOString()));
            }

            //force picker back to default mode
            showMode(viewMode);
            //update the current date display
            picker.querySelector('.date-time-header').innerText = viewDate.toDateString();
            //show the picker dropdown
            picker.show(ctx.el);
        }

        /**
         * Close the picker
         */
        function hidePicker(){
            //hide UI
            if(picker.open)picker.close();
        }

        /**
         * Toggle the display state of the picker dropdown UI
         * Remember that we can have a button trigger instead of an input
         * @param {HTMLElement} el
         * @param {Boolean} [state]
         */
        function showDropDown(el, state){
            if(state===false||(state!==true&&picker.open)){
                hidePicker();
                if (el.tagName !== 'INPUT') setActiveControl(el, false, false);
            }else{
                el.focus();
                setActiveControl(el,true,true);
            }
        }
        /**
         * load culture info
         * @param {string} [culture] the id of the new default culture
         */
        function setupCulture(culture){
            var cid = culture ? culture: System.user.cultureID;
            if(cultureID !== cid) {
                //culture changed - create culture shortcuts
                var dc = Globalization.cultures[cid];
                dates.days = dc.calendars.standard.days.namesAbbr;
                dates.daysShort = dc.calendars.standard.days.namesShort;
                dates.months = dc.calendars.standard.months.names;
                dates.monthsShort = dc.calendars.standard.months.namesAbbr;
                dates.format = dc.calendars.standard.patterns.d;
                dates.firstDay = dc.calendars.standard.firstDay;
                dates.lastDay = dates.firstDay + 6 % 7;
                dates.separator = dc.calendars.standard['/'];
                //localize template
                var tempTemplate = template.replace('[[DAYS_HEADER]]', getTemplatePart('daysHeader')).replace('[[MONTHS]]', getTemplatePart('months'));
                //reset global picket based on culture info - except what this actually was doing was repeatedly injecting the dialog.....
                if(picker){
                    //already exists - just update the content
                    picker.innerHTML = tempTemplate;
                }else{
                    //doesnt exist - so add the dialog
                    picker = DialogManager.addDialog('<dialog id="datepicker" class="datepicker popover picker" data-position="picker">' + tempTemplate + '</dialog>');
                }
            }
        }
        /**
         * Handler for internal system events
         * @param {string} type
         * @param {Object} returnValue
         * @returns {boolean}
         */
        function notifyHandler(type, returnValue) {
            switch (type) {
                case 'cm.system.globalization.loaded':
                    console.log('new globalization culture loaded');
                    setupCulture();
                    break;
            }
        }


        /**
         * 'Selects' the date and updates the ctx but doesn't commit the value yet
         * this is internal and the time element should be ignored
         * if d is undefined we will pick the value from viewDate
         * @param {string=} d
         */
        function selectDate(d){
            ctx.datetime = d === undefined ? viewDate.toISOString() : d;
            if (ctx.el.tagName === 'INPUT') {
                ctx.el.value = formatValue('d', ctx.datetime);
            }else{
                ctx.el.setAttribute('datetime', ctx.datetime);
            }
        }

        /**
         * Update the picker header display
         */
        function updateHeader(){
            var y = viewDate.getFullYear() - (viewDate.getFullYear() % 10);
            //find the days header and set the month/year title
            var header = picker.querySelector('.datepicker-header .picker-switch');
            header.setAttribute('data-days',dates.months[viewDate.getMonth()]+' '+viewDate.getFullYear());
            header.setAttribute('data-months',viewDate.getFullYear());
            header.setAttribute('data-years',y + '-' + (y + 9));
        }

        /**
         * Updates the view.
         * Create dynamic content or moves the cursor indicator
         * @param {boolean=} cursorOnly if true only moves the cursor indicator
         */
        function updateView(cursorOnly) {
            switch(viewMode){
                case 'days':
                    if(cursorOnly){
                        //update selected
                        //just a day change move the highlight
                        var key = viewDate.toISOString().substr(0,10);
                        var dc = picker.querySelector('.datepicker-days');
                        var dca = dc.querySelector('.active');
                        if(dca)dca.classList.remove('active');
                        dc.querySelector('[data-key="'+key+'"]').classList.add('active');
                    }else{
                        //setup dynamic contant
                        //start from first day of last week of previous month
                        var daysInPrevMonth = viewDate.daysInMonth('prev');
                        var daysInThisMonth = viewDate.daysInMonth();
                        var activeDate = viewDate.getDate();
                        var viewDay = viewDate.firstDayOfMonth();
                        var pm = (viewDay || 7) - 1 ;
                        var nm = daysInThisMonth + pm + 1;
                        var html = [];
                        var clsName;
                        var d, m, y;
                        var key;

                        var today = Date.today;
                        var todayM = today.getMonth()+1;
                        var todayY = today.getFullYear();
                        var todayD = today.getDate();

                        //calendar always has 42 days
                        for (var i = dates.firstDay ; i < 42 + dates.firstDay ; i++) {
                            if (i%7===0+dates.firstDay ) {//start of week
                                html.push('<tr>');
                            }
                            m = viewDate.getMonth()+1;
                            y = viewDate.getFullYear();
                            if (i <= pm) {
                                d = daysInPrevMonth  - pm + i;
                                m = m==1 ? 12 : m-1;
                                y = m==12 ? y - 1 : y;
                                clsName = ' prev';
                            } else if (i >= nm) {
                                d = i - nm + 1;
                                m = m==12 ? 1 : m+1;
                                y = m==1 ? y + 1 : y;
                                clsName = ' next';
                            }else{
                                d = i - pm;
                                clsName = (d === activeDate) ? ' active' : '';
                            }

                            if(y === todayY && m === todayM && todayD === d){
                                clsName += ' blue';
                            }

                            key = y+'-'+ String.padZeros(m,2)+'-'+String.padZeros(d,2);

                            html.push('<td class="dp day '+clsName+'" data-key="'+key+'">'+d+'</td>');
                            if (i%7===6+dates.firstDay) {
                                html.push('</tr>');
                            }
                        }
                        //find the days body and fill
                        picker.querySelector('.datepicker-days tbody').innerHTML=html.join('');
                        updateHeader();
                    }
                    break;
                case 'months':
                    //setup months panel
                    var mc = picker.querySelector('.datepicker-months');
                    var mca = mc.querySelector('.active');
                    if(mca)mca.classList.remove('active');
                    mc.querySelector('[data-key="'+(viewDate.getMonth()+1)+'"]').classList.add('active');
                    if(!cursorOnly){
                        updateHeader();
                    }
                    break;
                case 'years':
                    var yc = picker.querySelector('.datepicker-years');
                    if(cursorOnly){
                        var yca = yc.querySelector('.active');
                        if(yca)yca.classList.remove('active');
                        yc.querySelector('[data-key="'+viewDate.getFullYear()+'"]').classList.add('active');
                    }else{
                        //setup years panel
                        var html = '';
                        var year = viewDate.getFullYear();
                        var decade = parseInt(year/10, 10) * 10;
                        var yearCont = yc.querySelector('td');
                        for (var i = decade-1; i < decade+11; i++) {
                            html += '<span class="dp year'+(i === decade-1 ? ' prev' : i === decade+10 ? ' next' : '')+(i === year ? ' active' : '')+'" data-key="'+i+'">'+i+'</span>';
                        }
                        yearCont.innerHTML = html;
                        updateHeader();
                    }
                    break;
            }
        }

        /**
         * Handle picker clicks
         * @param {Event} e
         * @param view
         * @param cur
         * @param prev1
         * @param prev2
         * @param prev3
         * @returns {boolean}
         */
        function clickHandler(e, view, cur, prev1, prev2, prev3) {
            //find what was clicked
            var target = $.findFirstElementByClass(e.target,'dp');
            if (target) {
                switch(target.tagName.toLowerCase()) {
                    case 'th':
                        //clicked the header - prev/next or switch modes
                        var aid = cur;
                        switch(aid) {
                            case 'date:switch':
                                showMode(modes[viewMode].out);
                                //picker.find('.datepicker-'+ viewMode).focus();
                                break;
                            case 'date:prev':
                            case 'date:next':
                                //prev/next ... but what ... month or year .. pick up from viewmode
                                // Set the datepicker to always move to the first on the month when moving from one month to another per G.
                                viewDate['set'+modes[viewMode].navField].call(
                                    viewDate,
                                    viewDate['get'+modes[viewMode].navField].call(viewDate) +
                                    modes[viewMode].navStep * (aid === 'date:prev' ? -1 : 1),1
                                );
                                updateView();
                                break;
                        }
                        break;
                    case 'span':
                        if (viewMode !== 'days') {
                            var key = parseInt(target.getAttribute('data-key'));
                            if (viewMode === 'months') {
                                viewDate.setMonth(key-1);
                            } else {
                                viewDate.setFullYear(key);
                            }
                            showMode(modes[viewMode].inp);
                            updateView();
                        }
                        break;
                    case 'td':
                        if (target.classList.contains('day') && !target.classList.contains('disabled')){
                            var d = target.getAttribute('data-key');
                            selectDate(d+'T00:00:00');//we dont include the ms - cos all dates from server exclude the ms - so in order for us to properly detect change we must follow the server
                            //now close...
                            hidePicker();
                            commitValue();
                        }
                        break;
                }
                //put focus back to input control
                ctx.el.focus();
                return true;
            }else return false;
        }

        /**
         * Set the picker view mode and update the UI
         * @param mode
         */
        function showMode(mode) {
            viewMode = mode;
            updateView();
            picker.classList.remove('days','months','years');
            picker.classList.add(viewMode);
        }

        /**
         * Increments the current picker ViewDate
         * @param inc
         * @param mode
         */
        function incrementViewDate(inc,mode){
            var day = viewDate.getDate();
            var month = viewDate.getMonth();
            var year = viewDate.getFullYear();
            switch(mode||viewMode){
                case 'days':
                    viewDate.setDate(day+inc);
                    //month changed.... need to rebuild
                    updateView(!mode&&(viewDate.getMonth()===month));
                    break;
                case 'months':
                    viewDate.setMonth(month+inc);
                    updateView(!mode&&(viewDate.getFullYear()===year));
                    break;
                case 'years':
                    viewDate.setFullYear(year+inc);
                    updateView(!mode&&(Math.floor(viewDate.getFullYear()/10)===Math.floor(year/10)));
                    break;
            }
        }
        var cachedISODate;
        var cachedDate = new Date(0);
        /**
         * Increment the date in an input control
         * If the picker is active it sync the two
         * @param {HTMLElement} el
         * @param {number}[inc]
         */
        function incrementDateInput(el,inc){
            var iso = '';
            if(picker.open){
                incrementViewDate(inc);
                iso = viewDate.toISOString();
            }else{
                //v is ALWAYS valid or null - if null default in today...
                var v = ctx.datetime || DateTime.getDefaultDate(ctx.el, 'date');
                //now turn string into data so we can increment it then write it back
                //check if the cached date is this date if so we can skip parsing and use the cached data
                if(cachedISODate!==v){
                    cachedISODate=v;
                    cachedDate.setTime(Date.parse(v));
                }
                cachedDate.setDate(cachedDate.getDate()+inc);
                iso = cachedDate.toISOString();
            }
            ctx.datetime = iso;
            el.value = formatValue('d',iso);
        }
        var reStripNonDate = /[^dDmMyY]/
        /**
         * Commit the value if it is valid
         * Called from blur or keydown-return
         */
        function commitValue(){
            if (ctx.el.tagName === 'INPUT') {
                // check whether the date is valid and if so update the dt attrib
                // note if not required the date can be valid and NULL
                if (!ctx.el.value) {
                    //blank the date
                    selectDate(ctx.el.hasAttribute('required') ? DateTime.getDefaultDate(ctx.el, 'date') : '');
                } else {
                    //parse it
                    var v = ctx.el.value.split(dates.separator);
                    var d, m, y;
                    var fmt = String.replaceAll(dates.format, reStripNonDate, '').toLocaleLowerCase().replace('dd','d').replace('mm','m').replace('yyyy','y');
                    switch(fmt){
                        case 'mdy':
                            m = u.obn(v[0], Date.today.getMonth());
                            d = u.obn(v[1], Date.today.getDate());
                            y = u.obn(v[2], Date.today.getFullYear());
                            break;
                        case 'ymd'://yyyy-MM-dd
                            y = u.obn(v[0], Date.today.getFullYear());
                            m = u.obn(v[1], Date.today.getMonth());
                            d = u.obn(v[2], Date.today.getDate());
                            break;
                        case 'dmy':
                        default://dmy
                            d = u.obn(v[0], Date.today.getDate());
                            m = u.obn(v[1], Date.today.getMonth());
                            y = u.obn(v[2], Date.today.getFullYear());
                            break;
                    }
                    y = y > 999 ? y : y > 99 ? y * 10 : y + 2000;
                    m = m - 1 % 12 + 1;
                    d = Math.max(Math.min(Date.getDaysInMonth(y, m), d), 1);

                    selectDate(Date.getISODate(d, m, y));
                }
                //at this point datetime is empty or valid but the time part is irrelevant
                //so fix the time part and then set value into the form data
                // ctx.datetime represents the iso date format of whatever is in the control
                // it is null or a valid ISO date but the time portion should be regarded as 'undefined'
                var ndp = ctx.datetime ? ctx.datetime.substr(0, 10) : '';
                var dt = DateTime.commitInputField(ctx.el, 'date', ctx.dtm, ndp, Forms.getFormData(ctx.el.form));
                if (ctx.orig !== dt && !(!ctx.orig && !dt)) {
                    ctx.orig = dt;
                    //update the control
                    ctx.el.value = (ctx && ctx.dtm === 'datetime.CF') ? dt : formatValue('d', dt);
                    //fires the notify event and marks control and form dirty
                    Forms.commitValue(ctx.el);
                }
            }else{
                if ((ctx.orig !== ctx.datetime) || ctx.el.dataset.alwaysFireChange) {
                    ctx.orig = ctx.datetime;
                    ctx.el.setAttribute('datetime', ctx.datetime);
                    Views.fireActionEvent(ctx.el, 'datetimechange', false, false, {
                        action: 'datetimechange',
                        actionKey: ctx.datetime,
                        datetime: ctx.datetime
                    });
                }
            }
        }
        /**
         * Handler for Date Input blur event
         * @param e
         */
        function onInputBlur(e){
            if (ctx) {
                if (Focus.isFocusOut(e, ctx.el, picker)) {
                    hidePicker();
                    commitValue();
                    if (!e.force || (e.force && e.willBlur)) ctx = null;
                }
            }
        }

        /**
         * Handler for the Date Input focus event
         * @param e
         */
        function onInputFocus(e){
            //if getting focus from our own picker do nothing...
            setActiveControl(e.target, false, true);
        }
        function setActiveControl(el, show, active){
            if(active) {
                if (!ctx || ctx.el !== el) {

                    if (el.tagName === 'INPUT') {
                        //the active control....
                        ctx = Widget.getContext(el);
                        var data = Forms.getFormData(el);
                        var dt = data[el.name];
                        //this hold the full iso field value so we can track if it changed
                        ctx.orig = dt;
                        //this holds the controls adjusted view of the field value
                        //the orig iso date with the time element zero'd out
                        ctx.datetime = dt ? dt.substr(0, 10) + 'T00:00:00' : '';
                        //pickup the new viewMode
                        viewMode = ctx.viewMode;
                    } else {
                        //probably a button trigger
                        ctx = {el: el};
                        var dt = $.getAttributeValue(el, 'datetime', Date.ISOToday);
                        //this hold the full field value so we can track if it changed
                        ctx.orig = dt;
                        //this holds the controls adjusted view of the field value
                        ctx.datetime = dt ? dt.substr(0, 10) + 'T00:00:00' : '';
                        //pickup the new viewMode
                        viewMode = $.getAttributeValue(el, 'date-view-mode', 'days');
                    }
                }
                if (show)showPicker(el);
            }else{
                ctx.orig = ctx.datetime = '';
                ctx = null;
            }
        }
        /**
         * Handler for keyup
         * @param {KeyboardEvent} e
         */
        function onInputKeyup(e){
            switch(e.keyCode||e.key|| e.which){
                case 27://escape
                    if(picker.open){
                        //if open then close else undo to starting value
                        hidePicker();
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    }
                    break;
            }
        }
        /**
         * Handler for keydown
         * @param {KeyboardEvent} e
         */
        function onInputKeydown(e){
            var keyCode = e.keyCode||e.key|| e.which;
            //dropdown is active
            var block=true;
            switch(keyCode){
                case 32://SPACE
                case 13://ENTER
                    //transfer the viewDate and commit
                    if(picker.open) {
                        selectDate();
                        hidePicker();
                        commitValue();
                    }else{
                        commitValue();
                    }
                    break;
                case 38://up -7
                    if(e.altKey){
                        hidePicker();
                    }else{
                        if(picker.open) {
                            var inc = modes[viewMode].navInc;
                            incrementDateInput(e.target, -inc);
                        }else{
                            if(e.ctrlKey || e.metaKey)incrementDateInput(e.target, -1);
                            else block = false;

                        }
                    }
                    break;
                case 40://down +7
                    if(e.altKey){
                        showPicker(e.target);
                    }else{
                        if(picker.open) {
                            var inc = modes[viewMode].navInc;
                            incrementDateInput(e.target, inc);
                        }else{
                            if(e.ctrlKey || e.metaKey)incrementDateInput(e.target, 1);
                            else block = false;
                        }
                    }
                    break;
                case 37://left -1
                    if(e.ctrlKey || e.metaKey)incrementDateInput(e.target, -1);
                    else block = false;
                    break;
                case 39://right +1
                    if(e.ctrlKey || e.metaKey)incrementDateInput(e.target, 1);
                    else block = false;
                    break;
                case 68://d
                    if(picker.open)incrementViewDate(e.shiftKey ? -10 : 10,'years');
                    break;
                case 89://y
                    if(picker.open)incrementViewDate(e.shiftKey ? -1 : 1,'years');
                    break;
                case 77://m
                    if(picker.open)incrementViewDate(e.shiftKey ? -1 : 1,'months');
                    break;
                case 84://t
                    viewDate.setTime(Date.today.getTime());
                    updateView(false);
                    selectDate();
                    break;
                default:
                    block = false;
                    break;
            }
            if(block){e.preventDefault();e.stopImmediatePropagation();}
        }
        /**
         * Handler for all the callbacks from the main input controller
         * @param {Event} e
         * @returns {boolean} returns true for the Form Controller's default processing
         */
        function onInputEvent(e){
            switch(e.type){
                case 'input':return true;
                case 'change':return false;
                case 'focus':return onInputFocus(e);
                case 'blur':return onInputBlur(e);
                case 'keyup':return onInputKeyup(e);
                case 'keydown':return onInputKeydown(e);
                default:
                    return true;
            }
        }

        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         */
        function initInputElement(el, elForm) {
            var ctl = Widget.getContext(el);
            var ns = el.parentElement.nextSibling;
            if (ns && ns.classList.contains('cmtime')) {
                el.parentElement.classList.add('datetime');
            }
            //in mobile the control doesn't directly take input but becomes a button and a fake  input is shown in the dropddown
            if (Device.status.touchActive) el.type = 'button';
        }
        /**
         * Update a form controls value
         * @param {HTMLInputElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         */
        function setFieldValue(el, data){
            var fv = '';
            if (el.dataset.dtm === 'datetime.CF') { //
                // CF dates are stored in US format MM/DD/YYYY
                fv = data[el.name];
            } else {
                fv = formatValue('d', data[el.name]);
            }
            if (fv !== el.value) el.value = fv;
        }

        /**
         * topic handler for the pick list dialog
         * @param {string} topic
         * @param {Event} ae
         */
        function actionTopicHandler(topic, ae) {
            switch (ae.action) {
                case 'dialog-done':
                    dd.close();
                    ctx.showDropDown = false;
                    ctx.el.focus();
                    break;
                case 'dialog-clear':
                    //close the dropdown and let the key thru
                    if (picker.open) picker.close();
                    ctx.el.value = '';
                    break;

            }
        }

        init();
        self.onInputEvent = onInputEvent;
        self.setFieldValue=setFieldValue;
        self.initInputElement = initInputElement;

        self.showDropDown=showDropDown;
        self.actionTopicHandler = actionTopicHandler;

        self.clickHandler=clickHandler;
        return self;
    };
    /**@type DatePickerController*/
    window.DatePicker = controls.DatePickerController = new DatePickerController();
    Widget.setDefaultOptions('date',{viewMode:'days',dtm:''});
    Widget.setDefaultOptions('cmdate',{viewMode:'days',dtm:''});


})(jQuery);