(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion


    //region Lookup

    /**
     * The static singleton lookup controller
     * Note that we now always use an empty array rather than null forthe lookup list
     * @class
     * @name LookupController
     */
    var LookupController = function () {
        var self = this;
        //the context object for the active lookup
        var ctx;
        //the key search timer
        var keySearchTimer;
        //tmp cache for report options values
        var valueListCache = {};
        //local storage for cached lists
        var cache = {};
        //search timer delay
        var keySearchTimerDelay = 300;
        //the dropdown dialog element
        var dd;
        //the dropdown dialog template
        var template = '' +
            '<dialog id="lookup" class="select lookup" data-position="lookup" tabindex="-1" data-action-topic="dialog:lookup">' +
            '<header class="dialog-header">' +
            '<div class="input-control search flex">' +
            '<input type="search" data-type="search" name="searchText" class="ignore clearable" incremental data-source-target="lookup" data-search-action="lookup-search" autocomplete="off" placeholder="Search..." title="Enter some text to search for or press the search button for the entire list">' +
            '<icon class="search i-list helper" tabindex="-1"></icon>' +
            '</div>' +
            '</header>' +
            '<section class="dialog-content"></section>' +
            '<footer class="dialog-footer">' +
            '<button class="btn-dialog pad-l-m " data-action="dialog-check-none">clear</button>' +
            '<button class="btn-dialog pad-r-m  margin-l-a" data-action="dialog-done">done</button>' +
            '</footer>' +
            '</dialog>';


        /**
         * Initialize the controller
         * wheel or resize will be handled in the DM which will shut anything open....
         */
        function init() {
            //create global picker
            dd = DialogManager.addDialog(template, {owner: self});
        }
        init();
        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         * @memberOf LookupController#
         */
        function initInputElement(el, elForm) {
            //just create the lookup ctx object
            var l = Widget.getContext(el);
            //set initial value
            l.searchText = l.lastSearchText = '';
            //we are always directly linked to the labelfield
            l.labelField = el.name;
            //lookup.xxxx
            l.type = l.lookupType.split('.').pop();
            //if this has a dependent field add the filter
            if(l.dependentField)l.filterList = dependentFieldFilter;
            //if cacheable see if its been loaded and if not preload it
            if(l.cacheList)checkCacheList(l.type);

            //in mobile the control doesn't directly take input but becomes a button and a fake input is shown in the dropddown
            if (Device.status.touchActive) el.type = 'button';
        }

        /**
         * Show the lookup dropdown
         * This is only called from a lookup button click
         * It acts as a toggle
         * @param {HTMLElement} el the owning input element
         * @memberOf LookupController#
         */
        function showDropDown(el) {
            var controlCtx = Widget.getContext(el);
            //if we already the active element then focus wont fire (I think)
            if (document.activeElement === el) {
                showDropdownList(undefined, true);
            } else {//if already open - close it
                //if already closed - open it
                //let focus run
                controlCtx.showDropDown = !dd.open;
                el.focus();
            }
        }
        /**
         * Show the dropdown element
         * @param reload reload data from local cache and flush old data
         * @param {boolean} state
         */
        function showDropdownList(state, reload) {
            //open the dropdown anchor to the editor and flag the editor as active
            //if state is undefined then toggle
            if (state === undefined) state = !dd.open;
            //cleanup old state
            dd.childNodes[1].classList.remove('search-failed', 'search-error', 'too-many-rows', 'searching');
            if (state) {
                if (dd.open) {

                } else {
                    if (reload) {
                        if (Device.status.touchActive) {
                            if (ctx.useSearchFor) {
                                //but note that at this point the list is still populated with old data
                                //so even if you dont fire of a search you should flush/rebuild the display list
                                //if searchable then dont autosearch but wait for user action
                                loadFromCache(false);
                                // header is only shown on phone and tablet modes - taken care of by css
                                // show header if can incremental filter search - add class
                                // show search button if requires server search - add class
                                // hide header for cacheList true + temp
                                dd.classList.add('dd-search');
                                //dont load the search text into search box - a search starts a fresh
                                dd.querySelector('input').value = ctx.lastSearchText || '';
                                //show the dd
                                dd.show(ctx.el);
                            } else {
                                //not searchable to strip out the flag if previously set
                                dd.classList.remove('dd-search');
                                search('');
                            }
                        } else {
                            search('');
                        }
                    } else {
                        // but sometimes this is being called because the list has been filtered
                        // and the list needs repositioning (think place above and list shrinks)
                        dd.show(ctx.el);
                    }
                }
            } else {
                if (dd.open) {
                    dd.close();
                } else {

                }
            }
        }
        /**
         * Handler for list item click events
         * @param {Event} e
         * @param {View} view
         * @param {HTMLElement} el1 the list
         * @param {HTMLElement} el2 the listitem
         * @param {HTMLElement} el3
         * @param {HTMLElement} el4
         * @returns {boolean}
         * @memberOf LookupController#
         */
        function selectHandler(e, view, el1, el2, el3, el4) {
            var li = el2;
            if (li && li.dataset.index) {
                // select the target item AND commit it but not don't clear
                // in a list its ol/li
                // the click will have caused a blur event so we need to refocus the control
                selectItem(li, true);
                showDropdownList(false);
                //refocus if we need to
                if (document.activeElement !== ctx.el) {
                    ctx.showDropDown = false;
                    ctx.el.focus();
                }
                return false;
            }
        }

        /**
         * Handler for Lookup Input blur event
         * If clicking to another tab or window two blur events fire - ignore the second one
         * @param e
         */
        function onInputBlur(e) {
            if (ctx) {
                ctx.el.name = ctx.labelField;
                if (Focus.isFocusOut(e, ctx.el, dd)) {
                    if (!ctx.pendingBlur) {
                        //lookup is still valid - we don't clear the lookup here - wait for the full commit in case there is a server call
                        if (!e.force || (e.force && e.willBlur)) ctx.pendingBlur = true;
                        if (Device.status.touchActive) {
                            //in mobile mode nothing to do on blur...
                        } else {
                            // lookup is still active
                            // its poss that there is a search that hasn't run yet if so the timer will be active
                            // if there is some searchable text it could be a valid match
                            // or we could have navigated to an item
                            // or selected something
                            // 1) is the timer running if so fire the search now
                            // 2) or is the current text different from the last searched text
                            //  if empty and they tabbed out dont search its there fault
                            if (keySearchTimer) {
                                //do a search.... and select the first item that comes back
                                internalSearch();
                            }
                            //hide the dropdown if we are losing focus
                            showDropdownList(false);
                        }
                        //always must call commit even if nothing to commit as need to restore non-focus display
                        //commit will decide whether to do anyhting or not
                        commitValue(ctx);
                    }
                }
            }
        }
        /**
         * Handler for the Lookup Input focus event
         * If clicking in from another tab or window two focus events fire - ignore the second one
         * On mobile we only focus a control on a user click - and we autoopen the picker
         * formdata, row, oldValue, newValue, searchText, lastSearchText
         * @param e
         */
        function onInputFocus(e) {
            console.log('FOCUS LOOKUP')
            if (!e.target.readOnly) {
                // doesn't matter where the focus is coming from - if it is from the input, dd or helper then lookup will have been already be set
                var controlCtx = Widget.getContext(e.target);
                controlCtx.el.name = '_'+controlCtx.labelField;
                if (controlCtx !== ctx) {
                    //its a different control
                    ctx = controlCtx;
                    ctx.ds = Forms.getFormData(ctx.el.form, true);
                    ctx.formdata = ctx.ds.si;//the current row
                    ctx.row = Forms.getFormData(ctx.el);//this is usually the si but on a CF its a sub props si.CF.values
                    //382081 - change below allows you to paste in an incorrect value into an lookup field and then delete it.  Without this change you cannot delete the incorrect value
                    if (ctx.idField) ctx.el.value = ctx.row[ctx.idField] === '' ? ctx.el.value : ctx.row[ctx.idField] || '';
                    ctx.oldValue = ctx.el.value;
                    ctx.newValue = ctx.el.value;
                    ctx.searchText = ctx.searchText || '';//ctx.el.value;
                    ctx.lastSearchText = ctx.lastSearchText || '';//ctx.el.value;
                    dd.childNodes[1].innerHTML = '';
                }
                if (dd.open) {
                    if (ctx.showDropDown === false) {
                        showDropdownList(false);
                    } else {
                        //do nothing
                    }
                    ctx.showDropDown = null;
                } else {
                    //requested force close on focus
                    if (ctx.showDropDown === false) {
                        //already - closed do nothing
                    } else if (ctx.showDropDown === true) {
                        //requested force open on focus
                        showDropdownList(true, true);
                    } else {
                        //if 'task' always force open
                        //if on touch and we tabbed in so didnt trigger the click to open then also force
                        if ((Device.status.displayMode === 'desktop' && ctx.autoopen) || (Device.status.touchActive && Key.pressed.Tab)) {
                            showDropdownList(true, true);
                        }
                    }
                    ctx.showDropDown = null;
                }
            }
        }

        /**
         * Handler for an input event (cut/paste and key char events etc..)
         * @param e
         */
        function onInputInput(e) {
            //push value of fake control into the real control
            ctx.userChanged = true;
            if (Views.displayMode === 'mobile') {
                //exit if project
                if (ctx.useSearchFor) return false;
                ctx.el.value = input.value;
            }
            if (ctx.el.value) {
                if (ctx.cacheList) {
                    internalSearch();
                } else {
                    keySearchTimer = clearTimeout(keySearchTimer);
                    keySearchTimer = setTimeout(internalSearch, keySearchTimerDelay);
                }
            } else {
                //empty... per gg dont search - but deselect
                keySearchTimer = clearTimeout(keySearchTimer);
                ctx.searchText = '';
                ctx.selectedItem = null;
                ctx.selectedValue = null;
                ctx.selectedIndex = 0;
                if (dd.open) {
                    dd.close();
                }
            }
        }
        /**
         * Handler for keyup
         * These events could come from the embedded search input
         * @param {Event} e
         */
        function onInputKeyup(e) {
            switch (e.keyCode || e.which) {
                case 27://escape
                    if (dd.open) {
                        //if open then close else undo to starting value
                        showDropdownList(false);
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    }
                    break;
                case 13://return
                    // if the dropdown is open when u hit return we will close it and commit the current value
                    // otherwise we will commit the value and save
                    // we need to return false if we wish to block the form 'save'
                    if (dd.open) {
                        showDropdownList(false);
                        selectItem(dd.querySelector('.selected'), true);
                        return false;
                    } else {
                        commitValue(ctx);
                        return true;
                    }
                    break;

            }
        }

        /**
         * Keydown event handler for the input control
         * These events could come from the embedded seach input
         * @param {Event} e
         * @returns {boolean}
         */
        function onInputKeydown(e) {
            switch (e.keyCode) {
                case 38: // up
                    if (e.ctrlKey || e.metaKey) {
                        showDropdownList(false);
                        return false;
                    } else if (dd.open) {
                        moveSelect(-1);
                        return false;
                    }
                    break;
                case 40: // down
                    if (e.ctrlKey || e.metaKey) {
                        showDropDown(ctx.el);
                        return false;
                    } else if (dd.open) {
                        moveSelect(1);
                        return false;
                    }
                    break;
                case 13: // return
                    //we dont handle anything in the down we wait for the up
                    //but we kill the built in submit...
                    return false;
                    break;
                case 32://space
                    //if input is empty block space key but still trigger search
                    if (ctx.el.value && !ctx.searchText) ctx.searchText = ctx.el.value; //Issue 380698
                    if (!ctx.searchText) {
                        keySearchTimer = clearTimeout(keySearchTimer);
                        keySearchTimer = setTimeout(search, 0);
                        return false;
                    }
                    //if hit space in first position clear the search text
                    if (ctx.el.selectionStart === 0) {
                        ctx.el.value = '';
                        keySearchTimer = clearTimeout(keySearchTimer);
                        keySearchTimer = setTimeout(search, 0);
                        return false;
                    }
            }
            return true;
        }

        /**
         * Move the active selection in the dropddown list
         * @param dir
         */
        function moveSelect(dir) {
            var item = dd.querySelector('.data-row.selected');
            if (item) {
                item.classList.remove('selected');
                item = (dir === 1 ? item.nextSibling : item.previousSibling) || item;
            } else {
                item = dd.querySelector('[data-index]');
            }
            if (item) {
                if (!item.hasAttribute('data-index') || item.classList.contains('group-header')) {
                    //if this is an item without an index its not selectable so skip it
                    item.classList.add('selected');
                    $.scrollInView(item, dd.childNodes[1]);
                    if (dir === 1 && item.nextSibling === null) dir = -1;
                    if (dir === -1 && item.previousSibling === null) dir = 1;
                    moveSelect(dir);
                } else selectItem(item);
            }
        }
        /**
         * Commit the lookup value, updates the row lookup field values
         * Called from:
         * 1) A click on an item in the dd list (selectItem)
         * 2) Pressing return when the DD is open (selectItem)
         * 3) Tabbing out (Blur)
         * 4) On return from server (RDCReturnHandler) (from original blur) when lookup is no longer active as we have tabbed out
         * if the ds is destroyed there is no point in doing anything
         * if the ds exists but the form is destroyed we process without the Form.commit
         * per gg when the input box is cleared we set the selectedItem to null
         * @param {Object} l the lookup context object
         * @memberOf LookupController#
         */
        function commitValue(l) {
            var changed = true;
            if(l.ds.options.state==='destroyed'){
                //no DS so just clear and bale out
                l.pendingBlur = false;
                l.activeList = null;
                l.selectedIndex = null;
                l.selectedItem = null;
                l.searchText = '';
                l.oldValue = null;
                l.row = null;
                l.formdata = null;
                l.ds = null;
                if (ctx === l) ctx = null;
                changed=false;
            }else {
                if (l.pendingResponse) {
                    //we mark the pending as a form of error - this will prevent the row from being saved until the server call completes
                    window.ValidationManager.setDataError('validating', l.labelField, l.formdata, '', l.ds);
                    /**
                     * QMD 10/30/15
                     * Set the validating flag because on a tab out nothing has changed and the validating
                     * error doesn't get removed. On the next commit if validating remove the error and move along
                     * as normal. To recreate this you'll need to throttle your network speed and then tab out when
                     * the wmj guy is spinning.  See a couple lines downs for additional code where validating is true
                     */
                    l.validating = true;
                } else {
                    // QMD from 10/30/15: part of issue description above
                    if (l.validating) {
                        window.ValidationManager.setDataError('', l.labelField, l.formdata, 'lookup', l.ds);
                        l.validating = false;
                    }
                    //old and new value could be empty but searched and soft selected and tabbed out
                    if (l.el.value === l.oldValue && (!l.selectedValue || l.el.value === l.selectedValue)) {
                        //if value never changed then skip the commit
                        changed = false;
                    } else {
                        //dropdowns as ever screw up the lookup code and vice versa....... so we have to try and untangle things here.....
                        if (l.keyField) {
                            if (l.selectedItem && l.selectedItem.disabled !== 1) {
                                //we have selected something so it's valid - update the data row
                                if (l.row[l.keyField] != l.selectedItem.k) {
                                    l.row[l.keyField] = l.selectedItem.k;
                                    if (l.idField) l.row[l.idField] = l.selectedItem.i;
                                    l.row[l.labelField] = l.selectedItem.l;
                                    if (l.updateValue) l.updateValue(l);
                                    //we need to remove any previous error
                                    window.ValidationManager.setDataError('', l.labelField, l.formdata, 'lookup', l.ds);
                                } else changed = false;
                            } else {
                                //nothing selected either - remember this is a committable action!!!
                                // 1) what was entered wasn't valid or
                                // 2) wasnt a match but non restricted or
                                // 3) it was blanked out
                                var newValue = l.el.value;
                                if (newValue !== l.oldValue) {
                                    //invalid input or at least nothing found on server - set error message below
                                    l.row[l.keyField] = l.reportMode ? '' : 0;
                                    if (l.idField) l.row[l.idField] = '';
                                    l.row[l.labelField] = newValue;
                                    if (l.updateValue) l.updateValue(l);
                                    //we need to add/remove any error
                                    window.ValidationManager.setDataError((l.restrict && newValue) ? l.type + '.invalid' : '', l.labelField, l.formdata, 'lookup', l.ds);
                                } else changed = false;
                            }
                        } else {
                            if (l.selectedItem) {
                                //we have selected something so it's valid - update the data row
                                if (l.row[l.labelField] != l.selectedItem.l) {
                                    l.row[l.labelField] = l.selectedItem.l;
                                    if (l.updateValue) l.updateValue(l);
                                    //we need to remove any previous error
                                    window.ValidationManager.setDataError('', l.labelField, l.formdata, 'lookup', l.ds);
                                } else changed = false;
                            } else {
                                //nothing selected either what was entered wasn't valid or it was blanked out
                                var newValue = l.el.value;
                                if (newValue !== l.oldValue) {
                                    //invalid input or at least nothing found on server - set error message below
                                    l.row[l.labelField] = newValue;
                                    if (l.updateValue) l.updateValue(l);
                                    //we need to add/remove any error
                                    window.ValidationManager.setDataError((l.restrict && newValue) ? l.type + '.invalid' : '', l.labelField, l.formdata, 'lookup', l.ds);
                                } else changed = false;
                            }
                        }
                    }
                    //now we need to check if we need to update the display value back to its non-editing version.....
                    //1) we are still the active lookup but are losing focus so update the display
                    //2) the active lookup has changed but we are different element so update our old display
                    //3) the active lookup changed but we are the same element so ignore
                    //even if not changed we might need to switch from focused to non focused display
                    if (ctx && ctx.el === l.el && !l.pendingBlur) {
                        //still focused
                        l.el.value = l.oldValue = l.row[l.idField] || l.row[l.labelField] || '';
                    } else {
                        //not focused
                        l.el.value = l.oldValue = formatValue('lookup', l.row[l.idField], l.row[l.labelField]);
                    }

                    //this will fire the change event - and only for the active control - ie not if we move off the row
                    if (changed) {
                        // This clears out a Custom Field selection that is dependant 
                        // on a 'Parent Field' when the Parent selection is changed.
                        var item = l.el.form.querySelectorAll('input[data-dependent-field=' + l.labelField + ']')
                        if (u.obs(item)) {
                            for (var i = 0, len = item.length; i < len; i++) {
                                l.row[item[i].name] = '';
                            }
                        }

                        if(l.el.form.ctx){
                            Forms.commitValue(l.el);
                        }else{
                            //need to commit locally because the form no longer exists
                            l.ds.itemUpdated('U', l.row);
                            l.ds.fire('field', l.labelField);
                            // so if we move of the grid edit row before the lookup returned the non edit row may now be
                            // displaying the old data and won't update until the next change
                            // so we need the grid to rerender but not if it was already going to do so...
                            l.ds.refresh();
                            //but what of a pending save... no instance of control or form so if ds is not in batch mode force a save
                            if(!l.ds.options.batchMode){
                                l.ds.getModifiedItems(true);
                            }
                        }
                    }else{
                        //no change in the value so ntg to do BUT if the row was pending a save we need to force it...
                        if(!l.el.form.ctx){
                            // so if we move of the grid edit row before the lookup returned the non edit row may now be
                            // displaying the old data and won't update until the next change
                            // so we need the grid to rerender but not if it was already going to do so...
                            l.ds.refresh();
                            //but what of a pending save... no instance of control or form so if ds is not in batch mode force a save
                            if((l.row.rowstate.contains('U') || l.row.rowstate.contains('I')) && !l.ds.options.batchMode){
                                l.ds.getModifiedItems(true);
                            }
                        }
                    }
                    if (l.pendingBlur) {
                        //exiting input so clear everything...
                        l.pendingBlur = false;
                        l.activeList = null;
                        l.selectedIndex = -1;
                        l.selectedItem = null;
                        l.selectedValue = '';
                        l.searchText = '';
                        l.oldValue = '';
                        l.row = null;
                        l.formdata = null;
                        l.ds = null;
                        l.userChanged = false;

                        if (ctx === l) ctx = null;
                    }
                }
            }
            return changed;
        }
        /**
         * Mark the list item as selected and set the selectIndex and selectItem properties
         * @param {HTMLLIElement} li the element to be selected
         * @param {boolean} commit if true the controls value and searchText are also updated
         */
        function selectItem(li, commit) {
            var changed = false;
            //deselect any exisiting item
            var item = dd.querySelector('LI.data-row.selected');
            if (item) {
                item.classList.remove('selected');
                ///if (li === item) li = null;
            }
            // when we load we auto select the first item or the one with a matching id
            ctx.selectedIndex = li ? +li.dataset.index : -1;
            ctx.selectedItem = (ctx.selectedIndex === -1) ? null : ctx.activeList[ctx.selectedIndex];
            ctx.selectedValue = (ctx.selectedItem) ? ctx.selectedItem.i || ctx.selectedItem.l : '';

            if (li && ctx.selectedIndex !== -1) {
                li.classList.add('selected');
                $.scrollInView(li, dd.childNodes[1]);
            }
            if (commit) {
                //since we will flatten the search text no need to search
                keySearchTimer = clearTimeout(keySearchTimer);
                //if we clicked on an item or hit return we update the control display and the search value - this will stop searches
                ctx.el.value = ctx.searchText = ctx.selectedValue;
                //write values to field but doesn't trigger the change event
                changed = commitValue(ctx);
            }
            return changed;
        }

        /**
         * internal handler for normal search
         * only happens in desktop mode - search based on value in input box
         */
        function internalSearch() {
            keySearchTimer = clearTimeout(keySearchTimer);
            //adding protection for forms closing during the search or between the keysearchtimer
            if (!ctx || !ctx.row) return;

            //get current control text
            var v = ctx.el.value;
            //if its different from current searched text then run search again
            if (v !== ctx.searchText) {
                search(v);
            }
        }

        /**
         * the actual search processors
         * @param {string} value
         */
        function search(value) {
            keySearchTimer = clearTimeout(keySearchTimer);

            //adding protection for forms closing during the search or between the keysearchtimer
            if (!ctx || !ctx.row) return;

            //update the current searched text
            ctx.searchText = value || '';

            if (ctx.cacheList) {
                //load from local cache
                console.log('load from cache');
                loadFromCache(true);
            } else {
                console.log('load from server');
                //not cached locally need to load from server on every input
                loadData();
            }
        }

        /**
         * All data exists in the cache
         * some are replaced on every search, others remain in the cache permanently and are just filtered
         * called from search for a cacheable list - if no data (pre the filter step) then try and load from server
         * called from loadData for local sourced lists - if no data don't call load again
         * called from rdc return - if no data don't call load again
         * @param {boolean} canReload set to true if we can call the server to reload data
         */
        function loadFromCache(canReload) {
            //if not in cache then load it...
            //if in cache we need to filter it...
            //if old filter is a subset of current filter we can filter the subset...
            var data = getCachedList(ctx);
            if (data.length) {
                //got the list so filter it(if its not cacheable its already filtered) and pass it on to the display routine...
                data = filterList(ctx, data);
                //find the dropdown-find the list-search list
                if (data.length) {
                    dd.childNodes[1].innerHTML = getDataMarkup(data);
                    //check for a match
                    //only show if have data - call this first because selectItem-scrollInView depends on the dd being sized/positioned
                    showDropdownList(true);
                    //if we did open all and we had a value in the input we should find and select that
                    //select the first exact matching id
                    selectItem(dd.querySelector('LI.data-row[data-index="' + findMatch(ctx.el.value, data, ctx) + '"]'));
                } else {
                    dd.childNodes[1].innerHTML = '';
                    selectItem();//no data so deselect everything...
                }
                DialogManager.positionDropdown(dd, ctx.el, dd.info);
            } else {
                //list is not available locally so we need to load it from server
                dd.childNodes[1].innerHTML = '';
                DialogManager.positionDropdown(dd, ctx.el, dd.info);
                if (canReload) loadData();
                if (ctx.restrict === false) selectItem(); //no data so deselect everything...
            }
        }

        /**
         * Add a temp cached list
         * @param {string} values delimited list of values uses '|' as delimiter
         * @param {boolean} noKeys true if the values do not contain keys (custom fields)
         * @param {string} [delim] the list delimeter
         * @param {string} [listID] the list cache id
         * @memberOf LookupController#
         */
        function addValueList(values, noKeys, delim, listID) {
            //check if list already exists
            for (var o in valueListCache) {
                if (valueListCache[o].values === values) {
                    return o;
                }
            }
            //not found so create a new one
            var list = values.split(delim || '|');
            if (values.indexOf('::') !== -1) {
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i].split('::');
                    if(v.length===2){
                        list[i] = {rid:v[0], i: v[1], k: $.stringToValue(v[1]), l: v[1] };
                    }else{
                        list[i] = {rid:"", i: v[0], k: $.stringToValue(v[0]), l: v[0] };
                    }
                    
                }
            }else if (values.indexOf(':') === -1 || noKeys) {
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i];
                    list[i] = { i: v, k: v, l: v };
                }
            } else {
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i].split(':');
                    list[i] = { i: v[0], k: $.stringToValue(v[1]), l: v[0] };
                }
            }

            listID = '_vl' + (listID || ($.guid++));
            valueListCache[listID] = { list: list, values: values };
            return listID;
        }

        /**
         * filter a lookup list depending on the value in a dependent field
         * 'this' is the lookup ctx object
         * find all rows where the rid property mathces the value in the dependent field
         * @param {Object} item
         * @param {number} idx
         * @param {Array} src
         * @return {boolean}
         */
        function dependentFieldFilter(item, idx, src) {
            return !item.rid || item.rid.toLowerCase() === u.obs(this.row[this.dependentField]).toLowerCase();
        }

        /**
         *
         * @param type
         */
        function checkCacheList(type){
            if(!cache[type]) cache[type] = DBManager.LookupCache.getItem(type).done(function(list){cache[type] = list;});
        }
        /**
         * Load the cached list
         * @param {Object} l
         * @returns {Array}
         */
        function getCachedList(l) {
            return (l.cacheList === true ? cache[l.type] : l.cache)||[];
        }

        /**
         * Cache a list
         * @param {Object} l
         * @param {Array} data
         */
        function setCachedList(l, data) {
            if (l.cacheList === true) {
                DBManager.LookupCache.setItem(l.type, (cache[l.type] = data || []));
            } else l.cache = data||[];
        }
        /**
         * Finds the item with a matching id
         * Per the rules the match is case insensitive, EXACT, and on the 'i' field only
         * used when trying to select the 'current' item on returning data from server or after loading data
         * @param {string }searchText
         * @param {Array} data
         * @param {Object} l the lookup ????
         * @returns {number}
         */
        function findMatch(searchText, data, l) {
            //check we have something to search and that object supports the 'i' field
            if (searchText) {
                if (data.length) {
                    var searchLabel = u.obb(l.restrictSearchToLabelField);
                    //are we working on the 'i' or 'l' field and take account for restrict option
                    var f = searchLabel ? 'l' : l.idField ? 'i' : 'l';
                    //if searching in 'i' fallback to 'l' if no match on the current row - so cant have an 'i' value that duplicates an 'l' value!
                    searchText = '' + searchText.toLowerCase();
                    for (var i = 0, len = data.length; i < len; i++) {
                        if (l.initialSelectByKey && !l.userChanged) {
                            // since there is no key field if restrict to search field to label field is true
                            if (searchLabel) {
                                if (('' + data[i][f]).toLowerCase() === searchText) {
                                    return i;
                                }
                            } else {
                                if (data[i].k === l.row[l.keyField]) {
                                    return i;
                                }
                            }
                        } else if ((('' + data[i][f]).toLowerCase() === searchText) || (f === 'i' && (data[i].l.toLowerCase() === searchText))) {
                            if (!data[i].disabled) {
                                return i;
                            }
                        }
                    }
                    //if there is a partial match then return first otherwise return nothing
                    if (data[0].disabled){
                        for (var j = 1, len = data.length; j < len; j++) {
                            if (!data[j].disabled) {
                                return ((('' + data[j][f]).toLowerCase().includes(searchText)) || (f === 'i' && (data[j].l.toLowerCase().includes(searchText)))) ? j : -1;
                                break;
                            }
                        }
                    }
                    return ((('' + data[0][f]).toLowerCase().includes(searchText)) || (f === 'i' && (data[0].l.toLowerCase().includes(searchText)))) ? 0 : -1;
                } else {
                    return -1;
                }
            } else return -1;
        }
        /**
         * Filter a cacheable list - cacheable list contain a 'searchable' field which can be searched
         * @param {Object} l lookup context
         * @param {Array} list
         * @returns {Array} filtered list also set as l.activeList
         */
        function filterList(l, list) {
            //we filter all cached lists (for search text) and also anything that has an explicit custom filter (for constraint)
            return l.activeList = l.cacheList || l.filterList ?  list.filter(l.filterList||listFilter, l) : list;
        }

        /**
         * The default filter function - only called against cacheable list using incremental filtering
         * @param item
         * @param index
         * @param array
         * @returns {boolean}
         */
        function listFilter(item, index, array) {
        	//either contains or subset
        	if (item.i) {
        		return !this.searchText || item.l.toLowerCase().indexOf(ctx.searchText.toLowerCase()) !== -1 || item.i.toLowerCase().indexOf(ctx.searchText.toLowerCase()) !== -1;
        	} else {
        		return !this.searchText || item.l.toLowerCase().indexOf(ctx.searchText.toLowerCase()) !== -1;
        	}
            
        }

        /**
         * Get the params for sending to server
         * This can only be called against the active lookup so the global 'lookup' is valid
         * @returns {*}
         */
        function getParams() {
            var parms = ctx.getParams ? ctx.getParams(ctx) : {};
            //type is used server side to get the type of lookup data wanted...
            parms.type = ctx.type;
            parms.value = ctx.searchText;
            parms.reportMode = ctx.reportMode || false;
            return parms;
        }

        /**
         * Only called from an active element, global 'lookup' is therefore valid
         */
        function loadData() {
            if (ctx.source) {
                var data = (ctx.source.indexOf('_vl') === 0) ? valueListCache[ctx.source].list : Views.findView(ctx.el).view.module.model[ctx.source];
                setCachedList(ctx, data);
                loadFromCache();
            } else {
                var params = getParams();
                var hashCode = String.hashCode(JSON.stringify(params));
                if (hashCode === ctx.lastSearchParams) {
                    loadFromCache();
                } else {
                    ctx.pendingResponse = true;
                    ctx.lastSearchText = ctx.searchText;
                    ctx.lastSearchParams = params.hashCode = hashCode;
                    dd.childNodes[1].classList.remove('search-failed', 'search-error', 'too-many-rows', 'searching');
                    dd.childNodes[1].classList.toggle('searching', true);
                    if (!dd.open) dd.show(ctx.el);
                    Remote.remoteDataCall('Framework.Lookup', 'loadLookup', params, ctx).done(remoteDataCallReturnHandler);
                }
            }
        }

        /**
         * The return handler for remote call to load data
         * @param {Object} results
         * @param {Object} params
         * @param {string} method
         */
        function remoteDataCallReturnHandler(results, params, method) {
            //so the params is the old lookup
            var l = params;
            //is this response the last requested for this lookup....
            if(l.lastSearchParams === this.parameters.hashCode) {
                //we could be in a form that has closed or a view that has closed (we will handle or just not the active lookup.....
                //turn pending off
                l.pendingResponse = false;
                //and kill the message
                l.el.nextSibling.classList.toggle('loading', l.pendingResponse);
                //check whether the view still exists...if it doesn't then ignore
                if (l === ctx) {
                    dd.childNodes[1].classList.toggle('searching', false);
                }
                //set the data into cache - if there is an error server side 'list' might not exist
                setCachedList(l, (results.data && results.data.list) || []);
                //if pending a blur commit
                if (l.pendingBlur) {
                    //lookup may be no longer active but still need to commit it using the params 'lookup'
                    //since we just returned no selectedItem or active list at the moment
                    //filters and sets as active list - but if no filter, no cache and null returned from server...
                    //how can we get null from server, ss err, net err and rdc abort or search term didn't match anything
                    filterList(l, getCachedList(l));
                    var idx = findMatch(l.searchText, l.activeList, l);
                    l.selectedIndex = idx;
                    l.selectedItem = l.activeList[l.selectedIndex];
                    if (l.selectedItem) {
                        //if you tab out fast you get back the entire list, but find text gets set to 0 because it does not find the text.
                        //need to check if the item is disabled and move the pointer.
                        while (l.selectedItem.disabled && l.selectedIndex < l.activeList.length) {
                            l.selectedItem = l.activeList[l.selectedIndex];
                            l.selectedIndex++;
                        }
                        //if you go through the whole list and don't find anything not disabled, then reset selected item to null
                        if (l.selectedIndex == l.activeList.length) {
                            l.selectedIndex = -1;
                            l.selectedItem = null;
                        }
                    }
                    commitValue(l);
                } else {
                    //called from the active element so the global 'lookup' is still valid
                    //need to feed this back into load from cache
                    //so how can lookup be null and no pendingBlur???
                    if (ctx && ctx===l) {
                        //sometimes the server call succeeds, there are no result.errors but result.data is an empty object
                        //previously we called that a network error but now we call it search failed
                        //no idea what causes it but its a server side issue.....
                        var searchSuccess = !!results.data && results.data.list && results.data.list.length;
                        var searchFailed  = (!!results.data && results.data.list && results.data.list.length===0) || (!!results.data && !results.data.list);

                        var errorsTooManyRows = !!results.errors && results.errors[0].message === 'TooManyRows';
                        //some other explicit server side error
                        var errorsSearch = !!results.errors && !errorsTooManyRows;
                        //some other unknown error (server down, blew up on server, network failure
                        var errorsNetwork = !searchSuccess && !searchFailed && !errorsTooManyRows&& !errorsSearch;

                        dd.childNodes[1].classList.toggle('search-failed', searchFailed);
                        dd.childNodes[1].classList.toggle('search-error', errorsSearch || errorsNetwork);
                        dd.childNodes[1].classList.toggle('too-many-rows', errorsTooManyRows);
                        loadFromCache();
                    }
                }
            }
        }

        /**
         * Build the dropdown list inner markup
         * @param {Array} data
         * @returns {string} the result html markup
         */
        function getDataMarkup(data) {
            // limited results to a max number
            var len = ((ctx.maxItemsToShow > 0) && (ctx.maxItemsToShow < data.length)) ? ctx.maxItemsToShow : data.length;
            var html = '';
            var listviewClass = ctx.listClass || 'listview-lookup';
            var headerRowClass = ctx.headerRowClass || 'header-row';
            var dataRowClass = ctx.dataRowClass || 'data-row';
            //data-type added to aid in click event filtering
            html += '<ol class="' + listviewClass + '" data-type="listview-lookup">';

            if (ctx.formatHeader) {
                html += '<li class="' + headerRowClass + '" data-index="' + i + '">' + ctx.formatHeader() + '</li>';
            }

            //if a custom formatter returns an empty string we will hide the row via css
            var node = '';
            for (var i = 0; i < len; i++) {
                node = (ctx.formatItem ? ctx.formatItem(data[i], i, len) : '<div class="hbox"><div class="margin-r-a">' + data[i].l + '</div></div>');
                html += '<li class="' + dataRowClass + (data[i].isGroupRow ? ' group-header' : '') + ' nofocus" data-index="' + i + '">' + node + '</li>';
            }

            html += '</ol>';

            return System.localize(html);
        }

        /**
         * Add a lookup type config
         * @param type
         * @param config
         * @returns {*}
         * @memberOf LookupController#
         */
        function addType(type, config) {
            return Widget.setDefaultOptions(type, config);
        }
        /**
         * Update a form controls value
         * @param {HTMLInputElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         * @memberOf LookupController#
         */
        function setFieldValue(el, data) {
            //do we need to update?
            if (ctx && ctx.el === el) {
                // if there is an active lookup and this element is the one we are currently targeting then ignore the value change
                // as it derived from the control itself
                // except if this is also the focused 'activeElement' and the row object changed then we need to resetup the control
                if (el === document.activeElement && ctx.row !== data) {
                    //force reinit
                    //its a different row object
                    ctx.ds = Forms.getFormData(ctx.el.form, true);
                    ctx.formdata = ctx.ds.si;//the current row
                    ctx.row = Forms.getFormData(ctx.el);//this is usually the si but on a CF its a sub props si.CF.values
                    ctx.el.value = ctx.idField ? ctx.row[ctx.idField] || '' :  ctx.row[ctx.labelField] || '';
                    ctx.oldValue = ctx.el.value;
                    ctx.newValue = ctx.el.value;
                    ctx.searchText = ctx.searchText || '';
                    ctx.lastSearchText = ctx.lastSearchText || '';
                    ctx.selectedIndex = -1;
                    ctx.selectedItem = null;
                    ctx.selectedValue = '';
                    ctx.pendingBlur = false;
                    ctx.activeList = null;
                    ctx.userChanged = false;
                    dd.childNodes[1].innerHTML = '';
                }
            } else {
                //otherwise update it
                var l = Widget.getContext(el);
                var fv = formatValue('lookup', l.idField ? data[l.idField] || '' : '', data[l.labelField]);
                if (el.value !== fv) el.value = fv;
            }
        }
        /**
         * Handler for all the callbacks from the main input controller
         * We ignore change events as we write values back in commmitValue on blur
         * @param {Event} e
         * @returns {boolean} returns true for the Forms Controller's default processing
         * @memberOf LookupController#
         */
        function onInputEvent(e) {
            //keyup/keydown events could come from the embedded search field
            switch (e.type) {
                case 'input': return onInputInput(e);
                case 'change': return false;
                case 'focus': return onInputFocus(e);
                case 'blur': return onInputBlur(e);
                case 'keyup': return onInputKeyup(e);
                case 'keydown': return onInputKeydown(e);
                default:
                    return true;
            }
        }

        /**
         * topic handler for the pick list dialog
         * @param {string} topic
         * @param {Event} ae
         * @memberOf LookupController#
         */
        function actionTopicHandler(topic, ae) {
            switch (ae.action) {
                case 'dialog-check-none':
                    selectItem(null, true);
                    showDropdownList(false);
                    ctx.showDropDown = false;
                    ctx.el.focus();
                    break;
                case 'dialog-done':
                    showDropdownList(false);
                    ctx.showDropDown = false;
                    ctx.el.focus();
                    break;
                case 'lookup-search':
                    // pure search text
                    // unless restrict is false in which case its also potential value text
                    search(ae.actionKey);
                    break;

            }
        }
        self.onInputEvent = onInputEvent;
        self.setFieldValue = setFieldValue;
        self.initInputElement = initInputElement;

        self.showDropDown = showDropDown;
        self.selectHandler = selectHandler;

        self.addValueList = addValueList;
        self.addType = addType;
        self.commitValue = commitValue;
        self.actionTopicHandler = actionTopicHandler;
        return self;
    };
    //shortcut
    /** @alias LookupController*/
    window.Lookup = controls.LookupController = new LookupController();
    //endregion


})(jQuery);