(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion

    //region SelectList
    /**
     * Data loaded from server and cached in memory or as a valueList added explicitly to the controller or referenced by ID as a DS in the Model
     * select list differ slightly from lookups in that the true input is never focused instead an editor proxy is focused
     * labelField,idField,keyField,name output fields
     * l,k,i source fields
     * So is it selectList or selectlist or select-list or select.list or jumplist.selectList or select -
     * well all of those are in use and none of them apply to the selectlist control!!
     * It was supposed to be TagList
     * In 'tag' mode dropdown shows items whether checked or not, max 4 items
     * In 'txt' mode all items are show and checked are flagged
     * In general labels are '|' seperated string abnd keys are ',' seperated
     * labels are stored in the labelField and keys in the keyField as strings
     * labels are required but keys are optional
     * SO.. a control might have labelField and read that on focus but not have keys another control might have keys as well as label
     * if we dont have a labelfield then how do we load - we could parse the keys and do a lookup to get the labels
     * BUT we might not have the lookup datasource et we might have to fetch it....
     * BUT really we ALWAYS must have the labels as they are need to display in the grid....
     * so new behavior prompting a rewrite...
     * both seperators can be anything they can imagine not even a single character but can be a whole format with the value being embedded somewhere unknown in the string
     * it can not be specified so the dev has to have total control on the fly
     * same with the display of the tags they can be anything and cant be asked to specify it must be total flexibile
     * that accounts for why there are all the calls needed the type to do anything below
     * @returns {SelectListController}
     * @constructor
     */
    var SelectListController = function () {
        var self = this;
        //the dropdown dialog element
        var dd;
        //the context object for the active selectlist
        var ctx;
        //tmp cache for report options values
        var valueListCache = {};
        //locally cached list
        var cache = {};
        //auto commmit on list click or spacebar or return
        var autocommit = false;

        var template =
            '<dialog id="selectlist" class="select selectlist dd-search"  data-position="lookup" tabindex="-1" data-action-topic="dialog:selectlist">' +
            '<header class="dialog-header">' +
            '<div class="input-control search flex">' +
            '<input type="search" data-type="search" name="searchText" class="ignore clearable" incremental data-source-target="selectlist" data-search-action="lookup-search" autocomplete="off" placeholder="Search..." title="Enter some text to search for">' +
            '<icon class="search i-list helper" tabindex="-1"></icon>' +
            '</div>' +
            '</header>' +
            '<section class="dialog-content"></section>' +
            '<footer class="dialog-footer">' +
            '<button class="btn-dialog pad-l-m " data-action="dialog-check-all">select all</button>' +
            '<button class="btn-dialog pad-l-m " data-action="dialog-check-none">select none</button>' +
            '<button class="btn-dialog pad-r-m" data-action="dialog-done">done</button>' +
            '</footer>';

        /**
         * Initialize the controller
         */
        function init() {
            //create global picker - when we add a dialog we can set an 'owner' a static controller that can handle dialog actions
            dd = DialogManager.addDialog(template, {owner: self});
        }
        init();

        /**
         * Add a select list lookup type config
         * @param type
         * @param config
         * @returns {*}
         */
        function addType(type, config) {
            return Widget.setDefaultOptions(type, config);
        }

        /**
         * Init a forms input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         */
        function initInputElement(el, elForm) {
            //just create the selectlist object
            var controlCtx = Widget.getContext(el);
            //adjust the type - selectlist.xxxx
            controlCtx.searchText = ''; //init the searchText to a string
            controlCtx.type = controlCtx.selectlistType.split('.').pop();
            //add default processors
            controlCtx.joinTags = controlCtx.joinTags || joinTags;
            controlCtx.splitTags = controlCtx.splitTags || splitTags;
            controlCtx.formatTag = controlCtx.formatTag || formatTag;
            //if this has a dependent field add the filter
            if(controlCtx.dependentField)controlCtx.filterList = dependentFieldFilter;
            //if the list is cached then pull it from cache
            if(controlCtx.cacheList)loadCacheList(controlCtx.type);

            //in mobile the control doesn't directly take input but becomes a button and a fake  input is shown in the dropddown
            if (Device.status.touchActive) el.type = 'button';
            // move input inside the tag container
            var parent = el.parentElement;
            //add the mode class
            el.parentElement.classList.add(controlCtx.mode+'list');
            var wrapper = document.createElement('div');
            wrapper.classList.add('tag-container', 'nofocus');
            wrapper.appendChild(el);
            parent.appendChild(wrapper);
        }
        /**
         * Show the dropdown UI
         * This is only called from a selectlist button click
         * @param {HTMLElement} el the owning input element
         */
        function showDropDown(el) {
            var controlCtx = Widget.getContext(el);
            //if we are the active control show the dd else activate the ctl
            if (ctx === controlCtx) {
                //assume either kbd cmd or prg call or someone clicked the button
                showDropdownList(undefined, true);
            }else{
                // simply put focus back on the input control
                // the focus event fires immediately and marks the control as active in chrome - other browsers defer this to the next frame....
                controlCtx.showDropDown = !dd.open;
                el.focus();
            }
        }
        /**
         * Show the dropdown element
         * @param {boolean} state the requested state
         * @param {boolean} reload do we need to rebuild the html list
         */
        function showDropdownList(state, reload) {
            //open the dropdown anchor to the editor and flag the editor as active
            //if state is undefined then toggle
            if (state === undefined) state = !dd.open;
            if (state) {
                if (dd.open) {
                    //we are already - open do nothing
                } else {
                    //rebuild the list markup
                    loadFromCache();
                    //and remove any current filter
                    var ctl = dd.querySelector('input');
                    ctl.value = '';
                    ctl.classList.remove('hascontent');
                    dd.show(ctx.el);
                }
            } else {
                if (dd.open) {
                    dd.close();
                } else {
                    //already closed - nothing to do
                }
            }
        }
        /**
         * Handler for dropdown list item click events
         * @param {Event} e
         * @param {View } view
         * @param {HTMLOListElement} ol
         * @param {HTMLLIElement} li
         * @returns {boolean}
         */
        function selectHandler(e, view, ol, li) {
            if (li) {
                // select the target item, check it, and refocus the proxy input control
                // the click will have caused a blur event so we need to refocus the control
                setSelectedItem(li, e);
                toggleCheckedState(li);
                if (autocommit) showDropdownList(false);
            }
            return false;
        }
        /**
         * Handler for editor blur event
         * If clicking to another tab or window two blur events fire - ignore the second one
         * @param e
         */
        function onInputBlur(e) {
            if (ctx) {
                if (Focus.isFocusOut(e, ctx.el, dd)) {
                    if (!ctx.pendingBlur) {
                        //we are really losing focus from the control
                        if (!e.force || (e.force && e.willBlur)) ctx.pendingBlur = true;
                        if (Device.status.touchActive) {
                            //in mobile mode nothing to do on blur...
                        } else {
                            //hide the dropdown if we are losing focus
                            showDropdownList(false);
                        }
                        commitValue();
                    }
                }
            }
        }
        /**
         * Handler for the focus event - the target should be the editor or the helper button
         * If clicking in from another tab or window two focus events fire - ignore the second one
         * TODO:Need to check on idField handling - previous code didnt make sense so I removed it
         * @param {Event} e
         */
        function onInputFocus(e) {
            if (!e.target.readOnly) {
                var controlCtx = Widget.getContext(e.target);
                if (controlCtx !== ctx) {
                    //only if we are really becoming active
                    ctx = controlCtx;
                    // get the row so we can update from it
                    // this is usually the ds si but on a CF its a sub prop si.CF.values
                    ctx.row = Forms.getFormData(ctx.el);
                    //pull the inital values from the data - note at this point the display is already correct and upto date
                    ctx.labels = ctx.row[ctx.labelField] || '';
                    ctx.keys = ctx.row[ctx.keyField] || '';
                    //bind on.......
                    Key.bindKey('selectlist','ctrl+a', shortcutHandler, 'select all');
                    Key.bindKey('selectlist','ctrl+n', shortcutHandler, 'deselect all');
                    if (ctx.cache){if(!ctx.cacheList){ctx.cache = null;}}
                    //loads the lookup display
                    loadFromCache();
                }
                if (dd.open) {
                    if (ctx.showDropDown === false) {
                        showDropdownList(false);
                    } else {
                        //do nothing
                    }
                    ctx.showDropDown = null;
                } else {
                    //requested force close on focus
                    if (ctx.showDropDown === false) {
                        //already - closed do nothing
                    } else if (ctx.showDropDown === true) {
                        //requested force open on focus
                        showDropdownList(true, true);
                    } else {
                        //if 'task' always force open
                        //if on touch and we tabbed in so didnt trigger the click to open then also force
                        if (ctx.autoopen || (Device.status.touchActive && Key.pressed.Tab)) {
                            showDropdownList(true, true);
                        }
                    }
                    ctx.showDropDown = null;
                }
            }
        }

        /**
         * Get the params for sending to server
         * This can only be called against the active selectlist
         * @returns {*}
         */
        function getParams() {
            var parms = ctx.getParams ? ctx.getParams(ctx) : {};
            parms.type = ctx.type;
            return parms;
        }

        /**
         * Get the lookup data
         * a datasource that starts _vl means an internal  valuelist
         * if source is set its either a valueList or a list in the model
         * if source is not set then load from in memory cache/server
         */
        function loadData() {
            if (ctx.source) {
                //we have a local source
                var data = (ctx.source.indexOf('_vl') === 0) ? valueListCache[ctx.source].list : Views.findView(ctx.el).view.module.model[ctx.source];             
                //allow dev to append additional data - used mainly for the cross preds. If we need an additional mode create a custom function like merge data in LookupTypes
                //to allow the dev handle the adding of data.
                if (ctx.row.additionalData) data = data.flatIndex.concat(ctx.row.additionalData);
                //push into the cache
                setCachedList(ctx, data);
                //and reload
                loadFromCache();
            } else {
                //we can load from server and cache for the lifetime of the parent view
                var params = getParams();
                var searchParams = JSON.stringify(params);
                var hashCode = String.hashCode(searchParams);
                var view = Views.findView(ctx.el).view;
                var lookupCache = view.lookupCache || (view.lookupCache = {});
                var data = lookupCache[hashCode];
                if (data) {
                    //allow dev to append additional data - used mainly for the cross preds. If we need an additional mode create a custom function like merge data in LookupTypes
                    //to allow the dev handle the adding of data.
                    if (ctx.row.additionalData) data = data.flatIndex.concat(ctx.row.additionalData);
                    //push into the cache
                    setCachedList(ctx, data);
                    //and reload
                    loadFromCache();
                } else {
                    //need to call the server
                    ctx.pendingResponse = true;
                    //make the helper show as a loading icon
                    ctx.el.closest('.input-control').lastChild.classList.toggle('loading', ctx.pendingResponse);
                    //make the call
                    Remote.remoteDataCall('Framework.Lookup', 'loadLookup', params, ctx).done(remoteDataCallReturnHandler);
                }
            }
        }
        /**
         * The return handler for remote call to load data
         * @param {Object} results
         * @param {Object} controlCtx
         * @param {string} method
         */
        function remoteDataCallReturnHandler(results, controlCtx, method) {
            //forget it if there was some error
            if (results.data.ErrorCode||results.errors) {
                //notify error
                Notify.alert(results.data.ErrorString || results.data.ErrorMessage || results.errors[0].message);
                //turn pending off
                controlCtx.pendingResponse = false;
            } else {
                // rem we could be in a form that has closed or a view that has closed
                // we will handle or just not the active lookup.....
                // turn pending off
                controlCtx.pendingResponse = false;
                //and kill the loading message
                controlCtx.el.closest('.input-control').lastChild.classList.toggle('loading', controlCtx.pendingResponse);
                //set into cache
                setCachedList(controlCtx, results.data.list);
                //and reload but only if we are the active selectlist
                if (ctx === controlCtx) loadFromCache();
            }
        }

        /**
         * load a list from storage - only called on control init
         * @param {string} type
         */
        function loadCacheList(type){
            if(!cache[type]) cache[type] = DBManager.LookupCache.getItem(type).done(function(list){cache[type] = list;});
        }

        /**
         * get the cached list
         * @param {Object} controlCtx
         * @returns {Array}
         */
        function getCachedList(controlCtx) {
            //its either in the controller cache or is on the context cache or not yet loaded from server
            return controlCtx.cacheList === true ? cache[controlCtx.type] : controlCtx.cache;
        }

        /**
         * Add a list to the cache - if cachelist is true then persist to storage
         * @param {Object} controlCtx
         * @param {Array} data
         */
        function setCachedList(controlCtx, data) {
            if (controlCtx.cacheList === true) {
                DBManager.LookupCache.setItem(controlCtx.type, data);
            } else controlCtx.cache = data;
        }

        /**
         * All data exists in the cache eventually but might have to fetch from server initially
         * If we get it then we build the pick list and force the picker to display
         */
        function loadFromCache() {
            //if not in cache then load it from server...
            var data = getCachedList(ctx);
            if (data) {
                //got the list so filter it(if its not cacheable its already filtered) and pass it on to the display routine...
                data = filterList(ctx, data);
                //generate the lookup list markup and check the current items
                loadListItems(data);
            } else {
                //the list is not available locally so we need to load it from server - need to wait
                loadData();
            }
        }
        /**
         * Filter a cacheable list - cacheable list contain a 'searchable' field which can be searched
         * @param {Object} l lookup context
         * @param {Array} list
         * @returns {Array} filtered list also set as l.activeList
         */
        function filterList(l, list) {
            //we filter all cached lists (for search text) and also anything that has an explicit custom filter (for constraint)
            return l.filterList ?  list.filter(l.filterList, l) : list;
        }
        /**
         * Handler for the input event
         * @param {Event} e
         */
        function onInputInput(e) {
            //if first time or not open this will load and build the dropdown and reset
            showDropdownList(true, true);
            //so then we figure out the search string which will fitler the display
            search(ctx.el.value.trim());
            DialogManager.positionDropdown(dd, ctx.el, dd.info, true);
        }

        /**
         * Handler for the keyup event
         * @param {Event} e
         */
        function onInputKeyup(e) {
            var killkey = false;
            // a key event..... we only trap key events to look for the esc key otherwise we use the search/change event
            var keyCode = e.keyCode|| e.key || e.which;
            switch (keyCode) {
                case 13://ENTER
                case 27://ESC
                    killkey = true;
                    break;
            }
            if(killkey) {
                e.preventDefault();
                e.stopPropagation();
            }
        }
        /**
         * Keydown event handler for the input control
         * @param {Event} e
         */
        function onInputKeydown(e) {
            if(ctx) {
                var killkey = false;
                var keyCode = e.keyCode || e.key || e.which;
                switch (keyCode) {
                    case 9://tab
                        //close
                        showDropdownList(false);
                        break;
                    case 38: // up
                        if (Key.ctrlKey) {
                            //toggle close
                            showDropdownList(false);
                        } else setSelectedItem('prev');
                        killkey = true;
                        break;
                    case 40: // down
                        if (Key.ctrlKey) {
                            //toggle open
                            showDropdownList(true, true);
                        } else setSelectedItem('next');
                        killkey = true;
                        break;
                    case 13:// return
                        //if ntg in editor then commit else add if allowed or search if not
                        if (!ctx.el.value.trim()) {
                            //commit and close and save
                            if(dd.open){
                                //close
                                showDropdownList(false);
                            }else commitValue();
                        } else {
                            //has a value in the edit box
                            if (ctx.restrict) {
                                //if the dd is open and there is an item then select it and close
                                //toggle select state
                                toggleCheckedState();
                            }else{
                                //if there is content then add it and clear the editor
                                var label = ctx.el.value.trim();
                                var data = getCachedList(ctx);
                                var index = findInCacheList(data, label);
                                if (index === -1) {
                                    var newItem = {};
                                    newItem.i = newItem[ctx.sourceKeyField] = newItem[ctx.sourceLabelField] = label;
                                    data.push(newitem);
                                    loadListItems(data);
                                    addTag(label);
                                    refreshDisplayList(ctx, ctx.labels, ctx.keys);
                                }
                            }
                            killkey = true;
                        }
                        break;
                    case 32:// space
                        //if ntg in the editor and the dropddown is open then toggle state
                        if (!ctx.el.value.trim()){
                            if(dd.open) {
                                //toggle select state
                                toggleCheckedState();
                            }else{
                                onInputInput(e);
                            }
                            killkey = true;
                        }
                        break;
                    case 8://backspace
                        //if ntg in editor then bs is delete last element...
                        if(!ctx.el.value.trim()) {
                            deleteTag('last');
                            killkey = true;
                        }
                        break;
                    case 27://escape
                        //if open then close else undo to starting value
                        showDropdownList(false);
                        killkey = true;
                        break;

                }
                if (killkey) {
                    e.preventDefault();
                    e.stopPropagation();
                }
            }
        }

        /**
         * Check whether a label already exists in an array
         * @param {Array} arr
         * @param {string} label
         * @returns {number} index of item or -1 if not found
         */
        function findInCacheList(arr, label) {
            var i;
            var found = arr.some(function (item, index) {
                i = index;
                return label === item[ctx.sourceLabelField];
            });
            return found ? i : -1;
        }
        /**
         * Check/UnCheck all the list items
         * @param {boolean} all - state all true/false
         */
        function setAllChecked(all) {
            var list = dd.querySelector('ol');
            var items = list.querySelectorAll('li:not(.hidden)');
            for (var i = 0, len = items.length; i < len; i++) {
                items[i].classList.toggle('checked', !!all);
            }
        }

        /**
         *
         * @param {HTMLElement} el
         * @param {string} dir prev/next
         * @returns {Element}
         */
        function getNextAvailable(el, dir){
            if(el) {
                var next = (dir === 'prev') ? el.previousElementSibling : el.nextElementSibling;
                return ((next && next.classList.contains('hidden')) ? getNextAvailable(next, dir) : next) || el;
            }else{
                var list = dd.childNodes[1].firstChild;
                return dir === 'prev' ? list.querySelector('LI:not(.hidden)') : list.querySelector('LI:not(.hidden)');
            }
        }
        /**
         * Set the selected index and update the UI to reflect
         * note selected is NOT checked
         * @param {string|number|HTMLLIElement} item keyword prev, next or index value
         */
        function setSelectedItem(item, e) {
            //and flag as selected - note selected is not the same as checked
            //note this does not commit or check a value, simply moves the 'selected' marker
            //the current item
            var el = dd.childNodes[1].querySelector('LI.selected');
            //unselect the current selected item
            if (el) el.classList.remove('selected');
            
            //the new item - default first item, prev/next based on current or index value
            //dd.childNodes[1].querySelector('LI:not(.hidden)')
            el = (item === undefined) ? null: item === 'prev' ? getNextAvailable(el, item) : item === 'next' ? getNextAvailable(el, item) : item instanceof HTMLElement ? item : dd.childNodes[1].querySelector('LI[data-index="' + item + '"]');
            //select the new item
            if (el){
                if (e){
                    //in order to keep the tabbing functionality in the grid we need to identify if it's a mouse click 
                    if (e.e.type === 'mousedown'){
                        if (!item.classList.contains('checked')){
                            el.classList.add('selected');
                        }
                    }
                }else{
                    el.classList.add('selected');
                }                    
            }
            
            //force the UI to scroll in to view the selected item
            if(el)$.scrollInView(el, el);
        }

        /**
         * Flip the checked state of an item - called from select click and return with an empty editor
         * It only flip the list stateit doesnt add the item or commit
         * @param {HTMLLIElement} li
         */
        function toggleCheckedState(li) {
            //either passed the item or the current selected or the first item
            li = li || dd.childNodes[1].querySelector('LI.selected') || dd.childNodes[1].querySelector('LI:not(.hidden)');
            //flip the checked state of an item in the UI
            li.classList.toggle('checked');
        }

        /**
         * Update the internal properties from the dropdown values
         */
        function updateFromDropDown(){
            //update the value display and internal props from the dropdown.....
            var list = dd.querySelector('ol');
            //somehow we can get here and the list has been cleared need to fix that
            if(list) {
                var items = list.querySelectorAll('li.checked');
                var keys = [];
                var labels = [];
                //get the lookup datasource..
                //if a datasource and in tree mode then we need the flat index
                var data = filterList(ctx, getCachedList(ctx));
                data = data.mode==='tree' ? data.flatIndex :data;
                var row;
                for (var i = 0, len = items.length; i < len; i++) {
                    row = data[+items[i].dataset.index];                    
                    $.ArrayPush(keys, row[ctx.sourceKeyField]);
                    labels.push(row[ctx.sourceLabelField]);
                }
                //tag are added incrementally and cleared by clicking                  
                ctx.labels = ctx.joinTags('labels', labels);                
                ctx.keys = ctx.joinTags('keys', keys);
            }
        }

        /**
         * a ctx default function
         * join an array of string into a string
         * @param {string} type 'labels' or 'keys'
         * @param {Array} tags
         * @returns {string}
         */
        function joinTags(type, tags){
            var sep = type === 'labels' ? '|' : ',';
            return tags.join(sep);
        }

        /**
         * a ctx default function
         * split a string into an array of strings
         * @param {string} type 'labels' or 'keys'
         * @param {string} tags
         * @returns {Array}
         */
        function splitTags(type, tags){
            var sep = type === 'labels' ? '|' : ',';
            tags = tags.toString(); // in case we have a single number i.e. tags = 23, split is undefined
            return tags ? tags.split(sep) : [];
        }

        /**
         * a ctx default function
         * before a tag is added to the display this function is called to optionally modify it
         * @param tag
         * @param labels
         * @param keys
         * @returns {*}
         */
        function formatTag(tag,labels,keys){
            return tag;
        }
        /**
         * Refresh the active controls csv selected tag display list
         * Called from commit or pressing esc key or from toggle (shouldnt that be a commit anyway?)
         * polls the dropdown list for checked items and builds the internal labels and keys strings then sets the input value
         * @param {Object} ctx
         * @param {string} labels
         * @param {string} keys
         */
        function refreshDisplayList(ctx, labels, keys) {
            //note that MAS pulled value only from labelField - so assume 'name' should be same as labelField ??
            //add the tag elements - remove existing then add
            var parent = ctx.el.parentElement;
            var children = $.NodeListToArray(parent.children);
            for (var i = 0, len = children.length; i < len; i++) {
                var item = children[i];
                if(item.tagName==='ICON')item.remove();
            }
            if(labels) {
                var alabels = ctx.splitTags('labels', labels);
                var akeys = ctx.splitTags('keys', keys);
                for (var i = 0, len = alabels.length; i < len; i++) {
                    var tag = document.createElement('icon');
                    tag.classList.add('tag', 'nofocus');
                    tag.dataset.action = 'tag-click';
                    tag.dataset.actionKey = akeys[i] || alabels[i];
                    tag.innerText = alabels[i];
                    if(ctx.formatTag) ctx.formatTag(tag, alabels[i], akeys[i] || alabels[i]);
                    parent.insertBefore(tag, ctx.el);
                }
                ctx.el.placeholder = '';
            }
            if (!labels) { if (!ctx.el.dataplaceholder) {ctx.el.dataplaceholder = ctx.el.placeholder}; ctx.el.placeholder = ctx.el.dataplaceholder; }
            //NOT NICE solution but this fixes the issue with a required select list on a custom fields.  The el value is set to ' ' so that the validity property 
            //is has a boolean of valueMissing set to false.
            ctx.el.value = labels?' ':''; 
        }
        /**
         * Commit the current selected items update the input control value and related properties
         * Unlike Lookup we never commit on a non-active list (and only ever on a blur?)
         */
        function commitValue() {
            if (ctx.row[ctx.labelField] !== ctx.labels) {
                //update the row and the control
                ctx.row[ctx.labelField] = ctx.labels;
                //only write if the key field is defined
                if(ctx.keyField) ctx.row[ctx.keyField] = ctx.keys;
                //this will fire the change event - only for the active control
                Forms.commitValue(ctx.el);
            }
            if(ctx.pendingBlur) {
                ctx.pendingBlur = false;
                unloadList();
                ctx.row = null;
                ctx.labels = null;
                ctx.keys = null;
                ctx = null;
                Key.bindKey('selectlist','ctrl+a');
                Key.bindKey('selectlist','ctrl+n');
            }
        }

        /**
         * Build the pick list markup and flag the checked and selected items
         * @param {Array} data the array of lookup values
         */
        function loadListItems(data) {
            if (data) {
                //set the pick list
                dd.childNodes[1].innerHTML = getDataMarkup(data);
                setSelectedItem();
            }
        }

        /**
         * Builds the select list markup string
         * @param {Array} data
         * @returns {string}
         */
        function getDataMarkup(data) {
            var row, checked, rowHTML;
            var html = '';
            var shtml = '';
            //if a datasource and in tree mode then we need the flat index
            data = data.mode==='tree' ? data.flatIndex :data;
            // find items that should be checked
            // there can be either a key and label fields or just a label field
            // keys are ',' deliminated and labels are '|' deliminated
            // would have been so much better if that was reversed...
            // split into array of string 'keys' - could be based on the key or label
            var field = ctx.keyField ? ctx.sourceKeyField : ctx.sourceLabelField;
            var keys = ctx.keyField ? ctx.keys.toString().split(',') : ctx.labels.replace(/[\n\r]+/g, '').split('|');

            for (var i = 0; i < data.length; i++) {
                // match lookups - use data-row identifier - lookups can be grids...
                row = data[i];
                // check for a match against the k or l field value
                // if we can find a key then tag it as checked
                checked = (keys.indexOf('' + row[field]) !== -1) ? 'checked' : '';

                // as there may be no key we have to use an 'index' as the row identifier
                rowHTML = '';
                rowHTML += '<li class="data-row nofocus ' + checked + '" data-index="' + i + '">';
                rowHTML += (ctx.formatItem ? ctx.formatItem(data[i], i, data) : '<div class="hbox"><div class="margin-r-a">' + System.localize(row[ctx.sourceLabelField].toString()) + '</div></div>');
                rowHTML += '</li>';

                html += rowHTML;

            }
            return '<ol class="listview-selectlist" data-type="listview-selectlist">' + html + '</ol>';
        }

        /**
         * Clears out the markup in the selectlist list
         */
        function unloadList() {
            dd.childNodes[1].innerHTML = '';
        }

        /**
         * Update the value of a form control
         * @param {HTMLInputElement} el the true input control rather than the proxy
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data - this might be a CF 'values' object
         */
        function setFieldValue(el, data) {
            //do we need to update?
            if (ctx && ctx.el === el) {
                // if there is an active selectlist and the element is the one we are targeting then ignore it
                // we never update active controls here
                ctx.row = data;
                //pull the inital values from the data - note at this point the display is already correct and upto date
                ctx.labels = ctx.row[ctx.labelField] || '';
                ctx.keys = ctx.row[ctx.keyField] || '';
                refreshDisplayList(ctx, data[ctx.labelField] || '', data[ctx.keyField] || '');        
            } else {
                //none-active so update it
                var ctxNA = Widget.getContext(el);
                refreshDisplayList(ctxNA, data[ctxNA.labelField] || '', data[ctxNA.keyField] || '');
            }
        }
        /**
         * Input to a search filter - we leave the list content unchanged but toggle show/hide
         * valuelist Array, model DataSource , memory Array
         * @param {string} value the search value
         */
        function search(value) {
            //if changed and either value is empty or none-empty do a filter
            var filterValue = value.toLowerCase();
            var rootNode = dd.childNodes[1];
            if(ctx.searchText!==filterValue) {
                if(!ctx.el.hasAttribute('incremental')){
                    ctx.searchText = filterValue;
                    var parentNode = rootNode.firstChild;
                    removeHighlight(parentNode);
                    var nl = parentNode.childNodes;
                    var found;
                    for (var i = 0, len = nl.length; i < len; i++) {
                        if(filterValue) {
                            found = nl[i].innerText.toLowerCase().includes(filterValue);
                            if (found) {
                                nl[i].classList.remove('hidden');
                                innerHighlight(nl[i], filterValue);
                            }else nl[i].classList.add('hidden');
                        }else{
                            nl[i].classList.remove('hidden');
                        }
                    }
                }
            }
            //now we have it all filtered select the first available item
            setSelectedItem(ctx.searchText ? rootNode.querySelector('LI:not(.hidden)') : null);
        }

        /**
         * Highligh matching text in nodes
         * @param nodes
         * @param searchText
         */
        function highlight(nodes, searchText){
            var pat = searchText.pat.toUpperCase();
            for (var i = 0, len = nodes.length; i < len; i++) {
                innerHighlight(nodes[i], pat);
            }
        }

        /**
         * Highlight matching text in node
         * @param node
         * @param pat
         * @returns {number}
         */
        function innerHighlight(node, pat) {
            var skip = 0;
            if (node.nodeType === 3) {
                //text node
                var pos = node.data.toLowerCase().indexOf(pat);
                pos -= (node.data.substr(0, pos).toLowerCase().length - node.data.substr(0, pos).length);
                if (pos >= 0) {
                    var spannode = document.createElement('b');
                    ///spannode.className = 'highlight';
                    var middlebit = node.splitText(pos);
                    var endbit = middlebit.splitText(pat.length);
                    var middleclone = middlebit.cloneNode(true);
                    spannode.appendChild(middleclone);
                    middlebit.parentNode.replaceChild(spannode, middlebit);
                    skip = 1;
                }
            }else if (node.nodeType === 1 && node.childNodes) {
                //element node
                for (var i = 0; i < node.childNodes.length; ++i) {
                    i += innerHighlight(node.childNodes[i], pat);
                }
            }
            return skip;
        }

        /**
         * Remove highlighting from nodes
         * @param rootNode
         */
        function removeHighlight(rootNode) {
            var els = rootNode.querySelectorAll('b');
            for (var i = 0, len = els.length; i < len; i++) {
                var el = els[i];
                var parentNode = el.parentNode;
                parentNode.replaceChild(el.firstChild, el);
                parentNode.normalize();
            }
        }
        /**
         * Handler for all the callbacks from the main input controller and redirected events from the proxy
         * @param {Event} e
         * @returns {boolean} returns true for the Forms Controller's default processing
         */
        function onInputEvent(e) {
            if (e.target.classList.contains('ignore')) {
                //from the input proxy
                if (Device.status.touchActive) {
                    //in touch mode ignore proxy events
                } else {
                    //proxy in mouse mode
                    switch (e.type) {
                        case 'keyup':
                            return onInputKeyup(e);
                        case 'keydown':
                            return onInputKeydown(e);
                        default:
                            return true;
                    }
                }

            } else {
                switch (e.type) {
                    case 'input': return onInputInput(e);
                    case 'change': return false;
                    case 'focus': return onInputFocus(e);
                    case 'blur': return onInputBlur(e);
                    case 'keyup': return onInputKeyup(e);
                    case 'keydown': return onInputKeydown(e);
                    default:
                        return true;
                }
            }
        }
        /**
         * Add a temp cached list
         * @param {string} values delimited list of values uses '|' as delimiter
         */
        function addValueList(values) {
            //check if list already exists
            for (var o in valueListCache) {
                if (valueListCache[o].values === values) {
                    return o;
                }
            }
            //not found so create a new one
            var list = values.split('|');
            if (values.indexOf('::') !== -1) {
                //each item is just the k
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i].split('::');
                    if(v.length===2){
                        list[i] = {rid:v[0], i: v[1], k: $.stringToValue(v[1]), l: v[1] };
                    }else{
                        list[i] = {rid:"", i: v[0], k: $.stringToValue(v[0]), l: v[0] };
                    }
                }
            }else if (values.indexOf(':') === -1) {
                //each item is just the k
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i];
                    list[i] = { i: v, k: v, l: v };
                }
            } else {
                //each item is a k/v pair
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i].split(':');
                    list[i] = { i: v[0], k: v[1], l: v[0] };
                }
            }
            //give the list a uid and cache it
            var listID = '_vl' + ($.guid++);
            valueListCache[listID] = { list: list, values: values };
            return listID;
        }
        /**
         * filter a lookup list depending on the value in a dependent field
         * 'this' is the lookup ctx object
         * find all rows where the rid property mathces the value in the dependent field
         * @param {Object} item
         * @param {number} idx
         * @param {Array} src
         * @return {boolean}
         */
        function dependentFieldFilter(item, idx, src) {
            return !item.rid || item.rid.toLowerCase() === u.obs(this.row[this.dependentField]).toLowerCase();
        }

        /**
         * delete a current tag
         * @param {HTMLElement|string|number} tag tag or prev/first or label or index
         */
        function deleteTag(tag){
            if(tag) {
                var labels = ctx.splitTags('labels', ctx.labels);
                var keys = ctx.splitTags('keys', ctx.keys);
                var index = -1;
                //could be an item - or first - or last - or an index - or a label
                if (tag instanceof HTMLElement) {
                    index = labels.indexOf(tag.innerText);
                } else if (tag === 'last') {
                    index = labels.length - 1;
                } else if (tag === 'first') {
                    index = 0;
                } else if (typeof(tag) === 'string') {
                    index = labels.indexOf(tag);
                } else {
                    index = +tag;
                }
                if (index !== -1) {
                    labels.splice(index, 1);
                    ctx.labels = ctx.joinTags('labels', labels);
                    keys.splice(index, 1);
                    ctx.keys = ctx.joinTags('keys', keys);
                    refreshDisplayList(ctx, ctx.labels, ctx.keys);
                }
            }
        }

        /**
         * Adds a tag to labels/keys and updates the display
         * @param label
         * @param key
         */
        function addTag(label, key){
            if(label){
                ctx.labels = ctx.joinTags('labels', ctx.splitTags('labels',ctx.labels).concat([label]));
                ctx.keys = ctx.joinTags('keys', ctx.splitTags('keys',ctx.keys).concat([key||label]));
                refreshDisplayList(ctx, ctx.labels, ctx.keys);
            }
        }
        /**
         * topic handler for the pick list dialog
         * @param {string} topic
         * @param {Event} ae
         */
        function actionTopicHandler(topic ,ae){
            switch (ae.action){
                case 'tag-click':
                    //it is possible with the tags to clik con a tag without first having focused the control and hence not built the dd list or setup the ctx...
                    //it is also possible to have one ctx active and then click on another...
                    var target = ae.target;
                    if(!ctx || !ctx.el.parentElement.contains(target)){
                        //need to focus the control first
                        var input = target.closest('.input-control').querySelector('input');
                        input.focus();
                        onInputFocus({target: input}); //this needs to be called to set the ctx object. Otherwise the code below will throw an error
                    }
                    var action = ((target.offsetWidth - ae.e.offsetX)) > 22 ? ctx.el.dataset.action || 'delete' : 'delete';
                    if(action==='delete'){
                        var labels = ctx.splitTags('labels', ctx.labels);
                        var keys = ctx.splitTags('keys', ctx.keys);
                        var label = target.innerText;
                        var index = labels.indexOf(label);
                        labels.splice(index,1);
                        ctx.labels = ctx.joinTags('labels', labels);
                        keys.splice(index,1);
                        ctx.keys = ctx.joinTags('keys', keys);
                        refreshDisplayList(ctx,ctx.labels,ctx.keys);
                        commitValue(); //need to call commitValue so that it fires the change event to the module. Otherwise the user clicks the "X", but doesn't see a change until leaving the field
                    }else{
                        //fire action back to config let it handle this......
                        if(ctx.actionHandler){
                            ctx.actionHandler(ctx, action, target.dataset.actionKey);
                        }else{
                            Views.fireActionEvent(ctx.el, action, false, false, {action:action, key:target.dataset.actionKey});
                        }
                    }
                    break;
                case 'lookup-search':
                    search(ae.actionKey);
                    break;
                case 'dialog-check-all':
                    setAllChecked(true);
                    break;
                case 'dialog-check-none':
                    setAllChecked(false);
                    break;
                case 'dialog-done':
                    showDropdownList(false);
                    ctx.el.focus();                    
                    break;
                case 'selectlist.dialog-close':
                    //commit but not if we are already in the middle of a blur
                    if(true || !ctx.pendingBlur){
                        showDropdownList(false);
                        //per gg force the state of the current 'selected' item to be 'checked' on closing the dropdown
                        var li = dd.querySelector('li.selected:not(.checked)');
                        if (li)toggleCheckedState(li);
                        //now update the internal fields based on the dropdown display
                        updateFromDropDown();
                        //force tag container to redrw
                        refreshDisplayList(ctx, ctx.labels, ctx.keys);
                        if(!ctx.pendingBlur) commitValue();
                    }
                    break;
            }
        }

        /**
         * Callback for the shortcut handler
         * @param {Object} sc
         */
        function shortcutHandler(sc){
            switch(sc.key){
                case 'ctrl+a':
                    setAllChecked(true);
                    break;
                case 'ctrl+n':
                    setAllChecked(false);
                    break;
            }
        }
        self.addValueList = addValueList;
        self.onInputEvent = onInputEvent;
        self.setFieldValue = setFieldValue;
        self.initInputElement = initInputElement;

        self.showDropDown = showDropDown;
        self.selectHandler = selectHandler;
        self.actionTopicHandler = actionTopicHandler;

        self.addType = addType;
        return self;
    };
    //shortcut
    /**@alias SelectListController*/
    window.SelectList = controls.SelectListController = new SelectListController();

    //register defaults
    Widget.setDefaultOptions("selectlist", { source: '', labelField:'', keyField:'', idField:'', mode:'taglist', cfg:''});
    //endregion


})(jQuery);