(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion


    //region TextEditor
    /**
     * Controller for the Text Editor input control
     * @returns {TextEditorController}
     * @constructor
     */
    var TextEditorController = function() {
        var self = this;
        var domCache;
        var dlgFont;
        var blankLine;

        function init() {
            dlgFont = DialogManager.addDialog(getBuiltHTML('fontname'));
            domCache = document.createDocumentFragment();
            var div = document.createElement('menu');
            div.setAttribute('type', 'toolbar');
            div.classList.add('text-editor-toolbar', 'valign-center', 'toolbar');
            div.innerHTML = getBuiltHTML('toolbar');
            domCache.appendChild(div);
            //change the default paragrpah sep to P to make it consistent xbrowser
            document.execCommand('defaultParagraphSeparator', false, 'p');
            blankLine = document.createElement('P');
            blankLine.appendChild(document.createElement('BR'));
        }
        function getBuiltHTML(type) {
            var html = '';
            switch(type) {
                case 'editor':
                    html += '<article class="text-editor contenteditable usr" contenteditable="true" spellcheck="true" tabindex="0"></article>';
                    break;
                case 'fontname':
                    html += '<dialog id="fontname" class="tool popup" data-position="lookup" modal>';
                    html += '<header class="dialog-header" hidden></header>';
                    html += '<section class="dialog-content">';
                    html += '<menu type="popup" data-type="popup" class="popup dropdown">';
                    html += '<menuitem command="fontName" face="sans-serif" class="face sans-serif">Sans Serif</menuitem>';
                    html += '<menuitem command="fontName" face="serif" class="face serif">Serif</menuitem>';
                    html += '<menuitem command="fontName" face="arial black" class="face arial-black">Wide</menuitem>';
                    html += '<menuitem command="fontName" face="arial narrow" class="face arial-narrow">Narrow</menuitem>';
                    html += '<menuitem command="fontName" face="comic sans ms" class="face comic-sans-ms">Comic Sans</menuitem>';
                    html += '<menuitem command="fontName" face="courier new" class="face courier-new">Courier New</menuitem>';
                    html += '<menuitem command="fontName" face="garamond" class="face garamond">Garamond</menuitem>';
                    html += '<menuitem command="fontName" face="georgia" class="face georgia">Georgia</menuitem>';
                    html += '<menuitem command="fontName" face="tahoma" class="face tahoma">Tahoma</menuitem>';
                    html += '<menuitem command="fontName" face="trebuchet ms" class="face trebuchet-ms">Trebuchet MS</menuitem>';
                    html += '<menuitem command="fontName" face="verdana" class="face verdana">Verdana</menuitem>';
                    html += '</menu>';
                    html += '</section>';
                    html += '<footer class="dialog-footer"><button class="btn-dialog" data-action="dialog-close">Done</button></footer>';
                    html += '</dialog>';
                    break;
                case 'toolbar':
                    var ctrlKey = Key.HelpMap.ctrl;
                    html += '<menu type="toolbar" class="toolbar">';
                    html += '<button data-action="cmd" data-cmd="removeFormat" title="Remove Formatting" tabindex="-1" class="no-text icon i-clearformatting nofocus"></button>';

                    html += '<menu type="toolbar" class="toolbar expandable items5">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Format" tabindex="-1" class="notext icon i-font-format toolbar-button nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="bold" title="Bold (' + ctrlKey + '+B)" tabindex="-1" class="no-text icon i-bold nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="italic" title="Italic (' + ctrlKey + '+I)" tabindex="-1" class="no-text icon i-italic nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="underline" title="Underline (' + ctrlKey + '+U)" tabindex="-1" class="no-text icon i-underline nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="strikethrough" title="Strikethrough" tabindex="-1" class="no-text icon i-strikethrough nofocus"></button>';
                    html += '</menu>';

                    html += '<menu type="toolbar" class="toolbar expandable items5">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Size" class="no-text icon i-font-size toolbar-button nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="fontSize" value="1" title="Small" tabindex="-1" class="nofocus"><font size="1">A</font></button>';
                    html += '<button data-action="cmd" data-cmd="fontSize" value="2" title="Normal" tabindex="-1" class="nofocus"><font size="2">A</font></button>';
                    html += '<button data-action="cmd" data-cmd="fontSize" value="4" title="Big" tabindex="-1" class="nofocus"><font size="4">A</font></button>';
                    html += '<button data-action="cmd" data-cmd="fontSize" value="5" title="Huge" tabindex="-1" class="nofocus"><font size="5">A</font></button>';
                    html += '</menu>';

                    html += '<menu type="toolbar" class="toolbar expandable items6">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Color" class="no-text icon toolbar-button nofocus"><span class="color-swatch bg-red"></button>';
                    html += '<button data-action="cmd" data-cmd="foreColor" value="#d0333e" tabindex="-1" class="nofocus"><span class="color-swatch bg-red"></span></button>';
                    html += '<button data-action="cmd" data-cmd="foreColor" value="#0997d5" tabindex="-1" class="nofocus"><span class="color-swatch bg-blue"></span></button>';
                    html += '<button data-action="cmd" data-cmd="foreColor" value="#7bb724" tabindex="-1" class="nofocus"><span class="color-swatch bg-green"></span></button>';
                    html += '<button data-action="cmd" data-cmd="foreColor" value="#f9c301" tabindex="-1" class="nofocus"><span class="color-swatch bg-yellow"></span></button>';
                    html += '<button data-action="cmd" data-cmd="foreColor" value="#333333" tabindex="-1" class="nofocus"><span class="color-swatch bg-grayDark"></span></button>';
                    html += '</menu>';

                    html += '<button data-action="dialog" data-target="fontname" anchor title="Font" class="keepcase no-margin nofocus">Sans Serif</button>';

                    html += '<menu type="toolbar" class="toolbar expandable items3">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Lists" tabindex="-1"class="no-text icon i-list-bullet toolbar-button nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="insertunorderedlist" title="Bullet list" tabindex="-1"class="no-text icon i-list-bullet nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="insertorderedlist" title="Number list" tabindex="-1" class="no-text icon i-list-numbered nofocus"></button>';
                    html += '</menu>';

                    html += '<menu type="toolbar" class="toolbar expandable items3">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Indent" tabindex="-1" class="no-text icon i-indent-increase toolbar-button nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="outdent" title="Reduce indent (Shift+Tab)" tabindex="-1" class="no-text icon i-indent-decrease nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="indent" title="Indent (Tab)" tabindex="-1" class="no-text icon i-indent-increase nofocus"></button>';
                    html += '</menu>';

                    html += '<menu type="toolbar" class="toolbar expandable items5">';
                    html += '<button data-action="expandable" value="close" data-target="parent" title="Align" tabindex="-1" class="no-text icon i-paragraph-justify toolbar-button nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="justifyleft" title="Align Left (' + ctrlKey + '+L)" tabindex="-1" class="no-text icon i-paragraph-left nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="justifycenter" title="Center (' + ctrlKey + '+E)" tabindex="-1" class="no-text icon i-paragraph-center nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="justifyright" title="Align Right (' + ctrlKey + '+R)" tabindex="-1" class="no-text icon i-paragraph-right nofocus"></button>';
                    html += '<button data-action="cmd" data-cmd="justifyfull" title="Justify (' + ctrlKey + '+J)" tabindex="-1" class="no-text icon i-paragraph-justify nofocus"></button>';
                    html += '</menu>';
                    html += '</menu>';
                    html += '<button class="btn-action secondary icon-c no-text i-formattext margin-l-a active-on-toggle nofocus" data-action="toggle-class" data-target="grandparent" value="toolbar-active" tabindex="-1"></button>';
                    break;
            }
            return html;
        }
//region Paste
        var matchTypeImage = /^image.*/;
        var matchTypeText = /^text.*/;
        var matchTypeGoogle = /google.*/;
        /**
         * Handler for paste events in the TextEditor
         * @param {Event} e
         */
        function pasteHandler(e){
            //this is before pasted content has been inserted
            var clipboardData = e.clipboardData || window.clipboardData;//<<IE11
            var target = e.target;
            var editor = e.currentTarget;
            var isPreformatted = target.closest('pre');
            //take a snapshot
            //undoManager.get(editor).pushHistory(true);

            if($.support.browser.isIE11){
                setTimeout(afterPasteHandler,0, target, editor, clipboardData.files.length);
            }else {
                //first determine type of content text,html,image,rtf anything else will be ignored
                //the clipboard holds multiple types, we loop over all types to see what is available
                //RTE doesnt have capability to upload images as attachments so all images are converted to base64 and inlined
                //html and text has to be cleaned
                //NB this is for input to an HTML field all content is considered HTML after processing it CANT be used for text fields!!!!
                var isHTML, isPlainText, isImage, isText, isGoogle, isRTFText;
                Array.from(clipboardData.items).forEach(function (item, i) {
                    if (item.type.match(matchTypeImage)) {
                        isImage = item;
                    } else if (item.type === 'text/html') {
                        isHTML = item;
                    } else if (item.type === 'text/plain') {
                        isPlainText = item;
                    } else if (item.type === 'text/rtf') {
                        isRTFText = item;
                    } else if (item.type.match(matchTypeText)) {
                        isText = item;
                    }
                });
                if (isPreformatted) {
                    isPlainText.getAsString(pasteTextHandler);
                } else if (isHTML) {
                    var html = clipboardData.getData('text/html');
                    var rtf = isRTFText ? clipboardData.getData('text/rtf') : '';
                    pasteHTMLHandler(html, rtf);
                } else if (isImage) {
                    pasteImageHandler(isImage.getAsFile());
                } else if (isPlainText) {
                    isPlainText.getAsString(pasteTextHandler);
                } else if (isText) {
                    isText.getAsString(pasteTextHandler);
                }
                e.preventDefault();
            }
        }
        /**
         * Handler for IE11 paste events in the TextEditor
         * @param {HTMLElement} target
         * @param {HTMLElement} editor
         * @param {boolean} isImg
         */
        function afterPasteHandler(target, editor, isImg) {
            if(isImg) {
                var images = editor.querySelectorAll('img:not(.usr)');
                Array.from(images).forEach(function (img) {
                    img.classList.add('image-preview', 'usr');
                })
            }
            //HTMLSanitizer.clean(editor);
            //undoManager.get(editor).pushHistory(true);
        }

        /**
         * Handler for pasted html
         * @param {string} html
         * @param {string} rtf
         */
        function pasteHTMLHandler(html,rtf){
            //var editor = document.activeElement;
            var range = window.getSelection().getRangeAt(0);
            if(!range.collapsed)range.deleteContents();
            //clean up the markup and return the frag
            range.insertNode(HTMLSanitizer.cleanPastedHTML(html, rtf, true));
            range.collapse();
            var sel = window.getSelection();
            sel.removeAllRanges();
            sel.addRange(range);
            //we took a snapshot before pasting so do another one now
            //undoManager.get(editor).pushHistory(true);
        }
        /**
         * Handler for pasted text
         * @param {string} text
         */
        function pasteTextHandler(text){
            //var editor = document.activeElement;
            var range = window.getSelection().getRangeAt(0);
            if(!range.collapsed)range.deleteContents();
            var parentNode = range.commonAncestorContainer;
            //if its in a pre block paste as is otherwise parse
            var preformatted = !!(parentNode.nodeType === Node.ELEMENT_NODE ? parentNode : parentNode.parentNode).closest('pre');
            //if its preformatted escape it but dont touch it otherwise lets process it
            if(preformatted){
                //we are injecting into a pre block cTN will handle escaping if there was any markup
                var node = document.createTextNode(text);
                range.insertNode(node);
            }else{
                //inserting text into an html block... escape and convert crlf
                var frag = range.createContextualFragment(HTMLSanitizer.cleanPastedText(text));
                range.insertNode(frag);
            }
            range.collapse();
            var sel = window.getSelection();
            sel.removeAllRanges();
            sel.addRange(range);
            //we dont do any other cleanup we let smart quotes stand...
            //undoManager.get(editor).pushHistory(true);
        }

        /**
         * Handler for pasted image file - convert to Base64
         * @param {File} file
         */
        function pasteImageHandler(file){
            FileUpload.asDataURL(file).then(function(result){
                var img = document.createElement('img');
                img.classList.add('image-preview', 'usr');
                img.src = result;
                var range = window.getSelection().getRangeAt(0);
                range.deleteContents();
                range.insertNode(img);
                range.collapse();
                var sel = window.getSelection();
                sel.removeAllRanges();
                sel.addRange(range);
                //undoManager.get(el).pushHistory(true);
                //commitValue(document.activeElement);
            });
        }
//endregion

        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         */
        function initInputElement(el, elForm) {
            el.setAttribute('hidden', '');
            el.type = 'hidden';
            el.insertAdjacentHTML('afterend', getBuiltHTML('editor'));
            var editor = el.closest('.input-control').querySelector('article');
            editor.form = elForm;
            el.editor = editor;
            editor.setAttribute('name', el.name);
            editor.name = el.name;
            //check whether the control was disabled before we got initiialised....
            editor.contentEditable = !el.closest('.input-control').classList.contains('disabled');
        }

        /**
         * Rem on mac buttons dont get focus....
         * The required behavior
         * 1) show button on editor focus in
         * 2) show button on active when toolbar is open
         * 3) show button as inactive when toolbar closed
         * 4) hide button on editor blur
         * 1 + 4 handle by css only
         * @param e
         */
        function onInputFocus(e) {
            //on a mac a button doesnt get focus so the related target is wrong
            var el = e.target;
            if(el.isContentEditable) {
                var ctl = e.target.closest('.input-control');
                if (!ctl.classList.contains('active')) {
                    if (ctl.append) {
                        ctl.append(domCache);
                    } else {
                        //support out of date browser
                        var children = domCache.childNodes;
                        for (var j = 0; j < children.length; j++) {
                            ctl.appendChild(children[j]);
                        }
                    }
                    ctl.classList.add('active');
                    //listen for paste events
                    el.addEventListener('paste', pasteHandler)
                }
            }
        }

        /**
         * Input control blur event handler
         * @param e
         */
        function onInputBlur(e) {
            //on a mac a button doesnt get focus so the related target is wrong
            var relatedTarget = e.relatedTarget;
            var el = e.target;
            if(el.isContentEditable) {
                var ctl = e.target.closest('.input-control');
                //if there was markup and the field is cleared we should clear the markup
                if (el.innerText.trim() === '') el.innerHTML = '';
                if (e.force) {
                    if(e.willBlur){
                        //just trying to commit the active element
                        //really losing focus - not on a editor or toolbar btn click - close the toolbar and deactivate the btn
                        if (ctl.classList.contains('active')) {
                            ctl.classList.remove('active', 'toolbar-active');
                            ctl.querySelector('.active-on-toggle').classList.remove('active');
                            domCache.appendChild(ctl.lastChild);
                            commitValue(el);
                        }
                    }else commitValue(el);
                } else {
                    if (ctl.contains(relatedTarget) || dlgFont === relatedTarget || dlgFont.contains(relatedTarget)) {
                        //main controller will have removed focus but we need to add it back here as we arent really losing focus
                        ctl.classList.add('focus');
                        if (e.force) {
                            //just trying to commit the active element
                            //really losing focus - not on a editor or toolbar btn click - close the toolbar and deactivate the btn
                            if (ctl.classList.contains('active')) {
                                ctl.classList.remove('active', 'toolbar-active');
                                ctl.querySelector('.active-on-toggle').classList.remove('active');
                                domCache.appendChild(ctl.lastChild);
                                commitValue(el);
                            }
                        }
                    } else {
                        //really losing focus - not on a editor or toolbar btn click - close the toolbar and deactivate the btn
                        if (ctl.classList.contains('active')) {
                            ctl.classList.remove('active', 'toolbar-active');
                            ctl.querySelector('.active-on-toggle').classList.remove('active');
                            domCache.appendChild(ctl.lastChild);
                            el.removeEventListener('paste', pasteHandler);
                            commitValue(el);
                        }
                    }
                }
            }
        }
        /**
         * Commit editor content in an article element
         * We only need to clean on injection of foreign markup (paste, drag/drop, insert image/link)
         * and at start (setFieldValue) for none clean markup
         * On commit we should only need to normalize
         *@param {HTMLElement} el the editor element
         */
        function commitValue(el){
            //check we are stil in a active form
            if(!el.form || !el.form.ctx || el.form.ctx.mode==='close') return;
            //the real input control
            var ctl = el.previousElementSibling;
            //the field we are bound to
            var field = ctl.getAttribute('name');
            //the current row data
            var data = Forms.getFormData(ctl);

            var html = el.innerHTML;

            //now compare, wrap and save.... the origianl field value maybe unwrapped or wrapped...so get the value and unwrap to compare...
            var fieldValue = (data[field] || '').replace(/<span class="usr">/, '').replace(/<\/span>$/,'');
            // innerHTML will return '' for empty
            // so if the original was udf or null or '' we should accept as a match for editor value of ''
            // so only commit if changed but do a soft compare
            if(html != fieldValue) {
                console.log('RTE COMMIT CHANGE')
                //update the field but wrap before save
                data[field] = '<span class="usr">' + html + '</span>';
                //this is just to ensure the data change event fires on the row - it doesnt cause anything to update
                Forms.commitValue(el);
            }
        }
        /**
         * Handler for all the callbacks from the main input controller
         * @param {Event} e
         * @returns {boolean} returns true for the Forms Controller's default processing
         */
        function onInputEvent(e) {
            switch(e.type) {
                case 'input': return onInputInput(e);
                case 'change': return false;
                case 'focus': return onInputFocus(e);
                case 'blur': return onInputBlur(e);
                case 'keyup': return true;
                case 'keydown': return true;
                default:
                    return true;
            }
        }
        /**
         * Handler for the input event
         * @param {Event} e
         */
        function onInputInput(e){
            var el = e.target;
            if(el.isContentEditable) {
                if(el.lastElementChild.tagName === 'P'){
                    if(el.lastElementChild.childNodes.length===1){
                        if(el.lastElementChild.childNodes[0].tagName === 'BR'){
                            return true;
                        }
                    }
                }
                el.appendChild(blankLine.cloneNode(true));
            }
        }
        /**
         * Update a form controls value
         * @param {HTMLInputElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         */
        function setFieldValue(el, data) {
            //insert into the editor BUT only if this is not the active element and not the lockFocus element ...
            if(el.nextElementSibling !== document.activeElement){
                var html = data[el.name] || '<span class="usr"><p><br></p></span>';
                if(html.startsWith('<span class="usr">')) {
                    // its pre cleaned so now prune the leading and trailing flags
                    html = html.replace(/<span class="usr">/, '').replace(/<\/span>$/, '');
                }else if(html.startsWith('<div class="flexText usr">')){
                    // its old format pre cleaned which means it could be dirty so prune the leading and trailing flags and retest
                    html = html.replace(/<div class="flexText usr">/, '').replace(/<\/div>$/, '');
                    //its not clean so lets clean it
                    html = HTMLSanitizer.cleanPastedHTML(html, '', false);
                }else {
                    //its not clean so lets clean it
                    html = HTMLSanitizer.cleanPastedHTML(html, '', false);
                }
                //so now its clean and not wrapped.... just add the tail
                if(!html.endsWith('<p><br></p>'))html += '<p><br></p>';
                // set the value into the editor - this is now unflagged and tailed....
                el.nextElementSibling.innerHTML = html;
            }
        }

        /**
         * Per gg we replace the built in editing API 'removeFormat' command with our own version
         * that does more than just remove formatting and strips all markup
         */
        function removeFormatting() {
            var selection = window.getSelection();
            if(selection.isCollapsed) {
                var elAnchor = selection.anchorNode.nodeType === Node.TEXT_NODE ? selection.anchorNode.parentNode : selection.anchorNode;
                var editor = elAnchor.closest('[contenteditable]');
                var text = HTMLSanitizer.htmlToText(editor);
                editor.innerHTML = '<p>' + text+ '</\p>';
            }else{
                document.execCommand('removeFormat', true, true);
            }
        }

        /**
         *
         * @param cmd
         * @param userInterface
         * @param value
         */
        function execCommand(cmd, userInterface, value) {
            try {
                //TODO:FF will throw an exception if try to buller an empty list or the first item
                var node = document.getSelection().anchorNode;
                var editor = (node.nodeType === 1 ? node : node.parentNode).closest('[contenteditable]');
                if(cmd === 'removeFormat') {
                    //replace built in cmd with our own version
                    removeFormatting();
                } else {
                    document.execCommand(cmd, false, value);
                }
                //find the editor and refocus it...the button is in a div in a menu and the next el is the editor
                editor.focus();
                Forms.commitValue(editor);
            } catch(ex) {
                console.log(ex)
            }
        }
        init();
        self.onInputEvent = onInputEvent;
        self.setFieldValue = setFieldValue;
        self.initInputElement = initInputElement;
        self.execCommand = execCommand;
        return self;
    };
    //shortcut
    window.TextEditor = controls.TextEditorController = new TextEditorController();
    Widget.setDefaultOptions("texteditor", {});
    //endregion


//endregion
})(jQuery);


(function ($){

//region RichTextUtils
    /**
     * RichText Utilities
     */
    var RichTextUtils = function(){
        //var blockElementNames = ['P', 'LI', 'TD', 'TH', 'DIV', 'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 'PRE'];
        const blockElements = new Set(['ADDRESS', 'ARTICLE', 'ASIDE', 'AUDIO', 'BLOCKQUOTE', 'CANVAS', 'DD',
            'DIV', 'FIELDSET', 'FIGCAPTION', 'FIGURE', 'FOOTER', 'FORM', 'H1',
            'H2', 'H3', 'H4', 'H5', 'H6', 'HEADER', 'HGROUP', 'HR', 'LI',
            'NOSCRIPT', 'OL', 'OUTPUT', 'P', 'PRE', 'SECTION', 'TABLE', 'TR', 'TD',
            'TH', 'THEAD','TFOOT','CAPTION', 'UL', 'VIDEO', 'CODE']);
        //var inlineElementNames = ['A', 'B', 'STRONG', 'I', 'EM', 'SUB', 'SUP', 'U', 'STRIKE'];
        const inlineElements = new Set(['B', 'BIG', 'I', 'SMALL', 'TT',
            'ABBR', 'ACRONYM', 'CITE', 'XCODE', 'DFN', 'EM', 'KBD', 'STRONG', 'SAMP', 'VAR',
            'A', 'BDO', 'BR', 'IMG', 'MAP', 'OBJECT', 'Q', 'SCRIPT', 'SPAN', 'SUB', 'SUP',
            'BUTTON', 'INPUT', 'LABEL', 'SELECT', 'TEXTAREA']);
        const voidElements = new Set(['AREA', 'BASE', 'BR', 'COL', 'COMMAND', 'EMBED', 'HR', 'IMG', 'INPUT', 'KEYGEN',
            'LINK', 'META', 'PARAM', 'SOURCE', 'TRACK', 'WBR']);
        //nodes that just style their contents
        const styleNodes = new Set(['FONT', 'B', 'I', 'STRONG', 'EM', 'MARK']);
        const otherElements = new Set(['THEAD','TFOOT','TR', 'TBODY', 'COLGROUP','CAPTION']);

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isBlockElement(node) {
            return blockElements.has(node.nodeName);
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isInlineElement(node) {
            return inlineElements.has(node.nodeName);
        }
        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isVoidElement(node) {
            return voidElements.has(node.nodeName);
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function hasContent(node) {
            return (node && node.children && node.children.length > 0)  || (node && node.nodeName === 'BR');
        }

        /**
         * return true if nested inline tags ultimately just contain <br> or ""
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isEmptyInlineElement(node) {
            if( node.children.length > 1 ) return false;
            if( node.children.length === 1 && node.textContent.trim() !== '' ) return false;
            if( node.children.length === 0 ) return node.textContent.trim() === '';
            return isEmptyInlineElement(node.children[0]);
        }

        /**
         *
         * @param node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isText(node) {
            return node.nodeType === Node.TEXT_NODE;
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isEmptyTextNode(node) {
            return isText(node) && node.data === '';
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isFragment(node) {
            return node.nodeType === Node.DOCUMENT_FRAGMENT_NODE;
        }

        /**
         *
         * @param {Node} node1
         * @param {Node} node2
         * @return {number}
         * @memberOf RichTextUtils#
         */
        function isBefore(node1, node2) {
            return node1.compareDocumentPosition(node2) & Node.DOCUMENT_POSITION_FOLLOWING;
        }

        /**
         *
         * @param {Node} node
         * @param {string} className
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function elementHasClass(node, className) {
            return (node.nodeType === Node.ELEMENT_NODE && node.className === className)
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isSelectionMarkerNode(node) {
            return elementHasClass(node,'rte-marker');
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isCaretPositionNode(node) {
            return elementHasClass(node,'caret-position');
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isWhitespaceOnlyTextNode(node) {
            return node.nodeType === Node.TEXT_NODE && /^\s*$/.test(node.nodeValue);
        }

        /**
         *
         * @param {Node} node
         * @return {boolean}
         * @memberOf RichTextUtils#
         */
        function isTextNodeWithContent(node) {
            return node.nodeType === Node.TEXT_NODE && !isWhitespaceOnlyTextNode(Node, node);
        }

        /**
         *
         * @param {Node} node
         * @return {Node}
         * @memberOf RichTextUtils#
         */
        function firstDeepestChild(node) {
            var fc = node.firstChild;
            return !fc || fc.nodeName === 'BR' ? node : firstDeepestChild(fc);
        }

        /**
         *
         * @param newNode
         * @param referenceNode
         * @return {Node}
         * @memberOf RichTextUtils#
         */
        function insertAfter(newNode, referenceNode) {
            return referenceNode.parentNode.insertBefore(newNode, referenceNode.nextSibling);
        }

        /**
         *
         * @param {Node} node
         * @return {Node}
         * @memberOf RichTextUtils#
         */
        function removeNode(node) {
            return node.parentNode.removeChild(node);
        }
        /**
         *
         * @param {Node} node
         * @param rootElement
         * @param nodeFilter
         * @return {Node}
         * @memberOf RichTextUtils#
         */
        function getAncestor(node, rootElement, nodeFilter) {
            if (node !== rootElement) {
                var currentNode = node.parentNode;
                // If it's a `contenteditable` then it's likely going to be the editor
                // instance, so stop traversing there.
                while (currentNode && currentNode !== rootElement) {
                    if (nodeFilter(currentNode)) {
                        return currentNode;
                    }
                    currentNode = currentNode.parentNode;
                }
            }
            return null;
        }

        /**
         * get toe root level containing bloack
         * @param {Node} node
         * @return {Node}
         */
        function getRootContainingBlock(node){
            var currentNode = node;
            while (currentNode.parentNode && currentNode.parentNode.tagName !== 'ARTICLE') {
                currentNode = currentNode.parentNode;
            }
            return currentNode.parentNode ? currentNode : null;
        }
        /**
         *
         * @param {Node} node
         * @return {Array}
         * @memberOf RichTextUtils#
         */
        function nextSiblings(node) {
            var all = [];
            while (node = node.nextSibling) {
                all.push(node);
            }
            return all;
        }

        /**
         *
         * @param {Array} nodes
         * @param {Node} parentNode
         * @return {Node}
         * @memberOf RichTextUtils#
         */
        function wrap(nodes, parentNode) {
            nodes[0].parentNode.insertBefore(parentNode, nodes[0]);
            nodes.forEach(function (node) {
                parentNode.appendChild(node);
            });
            return parentNode;
        }

        /**
         *
         * @param {Node} node
         * @param childNode
         * @memberOf RichTextUtils#
         */
        function unwrap(node, childNode) {
            while (childNode.childNodes.length > 0) {
                node.insertBefore(childNode.childNodes[0], childNode);
            }
            node.removeChild(childNode);
        }

        /**
         * Chrome: If a parent node has a CSS `line-height` when we apply the
         * insertHTML command, Chrome appends a SPAN to plain content with
         * inline styling replicating that `line-height`, and adjusts the
         * `line-height` on inline elements.
         *
         * As per: http://jsbin.com/ilEmudi/4/edit?css,js,output
         * More from the web: http://stackoverflow.com/q/15015019/40352
         * @param {HTMLElement} parentElement
         * @memberOf RichTextUtils#
         */
        function removeChromeArtifacts(parentElement) {
            function isInlineWithStyle(parentStyle, element) {
                return window.getComputedStyle(element).lineHeight === parentStyle.lineHeight;
            }

            var nodes = Array.from(parentElement.querySelectorAll((['B','BIG','I','SMALL','TT']
                .map(function(tagName) { return tagName + '[style*="line-height"]' }))
                .join(',')));

            nodes = nodes.filter(isInlineWithStyle.bind(null, window.getComputedStyle(parentElement)));

            var emptySpans = [];

            nodes.forEach(function(node) {
                node.style.lineHeight = null;
                if (node.getAttribute('style') === '') {
                    node.removeAttribute('style');
                }
                if (node.nodeName === 'SPAN' && node.attributes.length === 0) {
                    emptySpans = emptySpans.push(node);
                }
            });

            emptySpans.forEach(function(node) {
                unwrap(node.parentNode, node);
            });
        }

        /**
         *
         * @param {string} text
         * @memberOf RichTextUtils##
         */
        function insertText(text){
            var node = document.createTextNode(text);
            var selection = window.getSelection();
            if (selection.rangeCount) {
                var range = selection.getRangeAt(0);
                range.collapse(false);
                range.insertNode(node)
                range.selectNode(node);
                range.selectNodeContents(node);
                range.collapse(false);
                selection.removeAllRanges();
                selection.addRange(range);
            }
        }
        /**
         * Insert an html string
         * @param {string} html
         * @param {boolean} block
         * @param {boolean} block
         * @memberOf RichTextUtils#
         */
        function insertHTML(html, block, selectStart) {
            // find the root of current block
            // if we are in an empty root level p tag we want to replace it otherwise
            var frag = $.fragmentFromString(html);
            var node = frag.firstElementChild;
            var selection = window.getSelection();

            if (selection.rangeCount) {
                var range = selection.getRangeAt(0);
                range.collapse(true);
                if(block){
                    ///if we are inside a p li or textnode we need to split it
                    //document.execCommand('insertParagraph', false, null);
                    //var rootBlock = range.commonAncestorContainer;
                    var rootBlock = RTEUtils.getRootContainingBlock(range.commonAncestorContainer);
                    node = rootBlock.insertAdjacentElement("afterend", node);
                    var isEmpty = node.innerHTML==='<br>';
                    range.selectNode(node);
                    range.selectNodeContents(node);
                    range.collapse(isEmpty);
                    selection.removeAllRanges();
                    selection.addRange(range);
                }else {
                    range.insertNode(frag);

                    // Move the caret immediately after the inserted span
                    range.setStartAfter(node);
                    range.collapse(true);
                    selection.removeAllRanges();
                    selection.addRange(range);
                }
            }
            return node;
        }
        /**
         *
         * @param {Node} node
         * @return {HTMLElement}
         * @memberOf RichTextUtils#
         */
        function getElementFromNode(node){
            while(node && node.nodeType!==Node.ELEMENT_NODE) node = node.parentElement;
            return node;
        }

        /**
         * @returns Closest ancestor Node satisfying nodeFilter. Undefined if none exist before reaching rte container.
         * @memberOf RichTextUtils#
         */
        function getContainingNode(range, nodeFilter, root) {
            var node = range.commonAncestorContainer;
            return  node  && nodeFilter(node) ? node :  getAncestor(node, root, nodeFilter);
        }

        /**
         *
         * @param {boolean} focus
         * @param {Node} node
         * @return {HTMLElement}
         * @memberOf RichTextUtils#
         */
        function getTextEditor(focus, node){
            if(!node){
                //find a node in the selection
                var selection = getSelection();
                node = selection.focusNode;
            }
            //find nearest element for the current node
            var el = getElementFromNode(node);
            //find the editor
            var editor = el.closest('[contenteditable]');
            if(focus){
                //make sure it has focus.....
                if(editor !== document.activeElement)editor.focus();
            }
            return editor;
        }

        return {
            isInlineElement: isInlineElement,
            isBlockElement: isBlockElement,
            isVoidElement:isVoidElement,
            isEmptyInlineElement: isEmptyInlineElement,
            isText: isText,
            isEmptyTextNode: isEmptyTextNode,
            isWhitespaceOnlyTextNode: isWhitespaceOnlyTextNode,
            isTextNodeWithContent: isTextNodeWithContent,
            isFragment: isFragment,
            isBefore: isBefore,
            isSelectionMarkerNode: isSelectionMarkerNode,
            isCaretPositionNode: isCaretPositionNode,
            firstDeepestChild: firstDeepestChild,
            insertAfter: insertAfter,
            removeNode: removeNode,
            getAncestor: getAncestor,
            nextSiblings: nextSiblings,
            wrap: wrap,
            unwrap: unwrap,
            removeChromeArtifacts: removeChromeArtifacts,
            elementHasClass: elementHasClass,
            hasContent: hasContent,
            getRootContainingBlock:getRootContainingBlock,
            insertText:insertText,
            insertHTML:insertHTML,
            getElementFromNode:getElementFromNode,
            getContainingNode:getContainingNode,
            getTextEditor:getTextEditor
        };
    };
    /**@type RichTextUtils*/
    window.RTEUtils = new RichTextUtils();

//endregion

})(jQuery);
(function ($) {
    "use strict";
    //region Imports
    const tags = {
        A:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true, href:true, target:true, rel:true, class:true}},
        ABBR:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        ACRONYM:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        ADDRESS:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        APPLET:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        AREA:{whitelist:true, block:false, inline:false, void:true, meta:false, attr:{style:true}},
        ARTICLE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        ASIDE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        AUDIO:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        B:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        BASE:{whitelist:false, block:false, inline:false, void:true, meta:true, attr:{}},
        BASEFONT:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        BDI:{whitelist:false, block:false, inline:true, void:false, meta:false, attr:{}},
        BDO:{whitelist:false, block:false, inline:true, void:false, meta:false, attr:{}},
        BGSOUND:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{style:true}},
        BIG:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        BLINK:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        BLOCKQUOTE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        BODY:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        BR:{whitelist:true, block:false, inline:true, void:true, meta:false, attr:{}},
        BUTTON:{whitelist:false, block:true, inline:true, void:false, meta:false, attr:{}},
        CANVAS:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        CAPTION:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        CENTER:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        CITE:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        CODE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        COL:{whitelist:true, block:false, inline:true, void:true, meta:false, attr:{style:true,width:true,span:true,align:true,valign:true}},
        COLGROUP:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{}},
        COMMAND:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        CONTENT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DATA:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        DATALIST:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        DD:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DEL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DETAILS:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DFN:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        DIALOG:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        DIR:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DIV:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,class:true}},
        DL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        DT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        ELEMENT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        EM:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        EMBED:{whitelist:false, block:false, inline:false, void:true, meta:false, attr:{style:true}},
        FIELDSET:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FIGCAPTION:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FIGURE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FONT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{}},
        FOOTER:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FORM:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FRAME:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        FRAMESET:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H1:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H2:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H3:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H4:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H5:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        H6:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        HEAD:{whitelist:false, block:false, inline:false, void:false, meta:true, attr:{}},
        HEADER:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        HGROUP:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        HR:{whitelist:true, block:false, inline:false, void:true, meta:false, attr:{style:true}},
        HTML:{whitelist:false, block:true, inline:false, void:false, meta:true, attr:{style:true}},
        I:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{}},
        IFRAME:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        IMAGE:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        IMG:{whitelist:true, block:false, inline:true, void:true, meta:false, attr:{style:true, height: true, width: true, src:true}},
        INPUT:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        INS:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        ISINDEX:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        KBD:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        KEYGEN:{whitelist:false, block:false, inline:false, void:true, meta:false, attr:{style:true}},
        LABEL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        LEGEND:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        LI:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,class:true}},
        LINK:{whitelist:false, block:false, inline:false, void:true, meta:false, attr:{}},
        LISTING:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MAIN:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MAP:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MARK:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        MARQUEE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MATH:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MENU:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MENUITEM:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        META:{whitelist:false, block:false, inline:false, void:false, meta:true, attr:{style:true}},
        METER:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        MULTICOL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NAV:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NEXTID:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NOBR:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NOEMBED:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NOFRAMES:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        NOSCRIPT:{whitelist:false, block:true, inline:false, void:false, meta:true, attr:{}},
        OBJECT:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        OL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true, start:true, reversed:true, type:true, compact:true}},
        OPTGROUP:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        OPTION:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        OUTPUT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        P:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,align:true,class:true}},
        PARAM:{whitelist:true, block:false, inline:false, void:true, meta:false, attr:{}},
        PICTURE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        PLAINTEXT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        PRE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        PROGRESS:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        Q:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        RB:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        RBC:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        RP:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        RT:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        RTC:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        RUBY:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        S:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SAMP:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SCRIPT:{whitelist:false, block:false, inline:false, void:false, meta:true, attr:{}},
        SECTION:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        SELECT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        SHADOW:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        SLOT:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SMALL:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SOURCE:{whitelist:true, block:false, inline:false, void:true, meta:false, attr:{}},
        SPACER:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        SPAN:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true, class:true}},
        STRIKE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        STRONG:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        STYLE:{whitelist:true, block:false, inline:false, void:false, meta:true, attr:{style:true, title:true}},
        SUB:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SUMMARY:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        SUP:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        SVG:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        TABLE:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,class:true,cellspacing:true,cellpadding:true,border:true,align:true,bgcolor:true,frame:true,rules:true,width:true}},
        TBODY:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        TD:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,class:true, colspan:true, rowspan:true, nowrap:true, scope:true, valign:true, width:true, height:true, bgcolor:true, align:true}},
        TEMPLATE:{whitelist:false, block:true, inline:false, void:false, meta:false, attr:{}},
        TEXTAREA:{whitelist:false, block:false, inline:true, void:false, meta:false, attr:{}},
        TFOOT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        TH:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true,class:true,colspan:true, rowspan:true, nowrap:true, scope:true, valign:true, width:true, height:true, bgcolor:true, align:true}},
        THEAD:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        TIME:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        TITLE:{whitelist:false, block:true, inline:false, void:false, meta:true, attr:{}},
        TR:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        TRACK:{whitelist:true, block:false, inline:false, void:true, meta:false, attr:{}},
        TT:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        U:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        UL:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true, type:true, compact:true}},
        VAR:{whitelist:true, block:false, inline:true, void:false, meta:false, attr:{style:true}},
        VIDEO:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
        WBR:{whitelist:false, block:false, inline:true, void:true, meta:false, attr:{}},
        XMP:{whitelist:true, block:true, inline:false, void:false, meta:false, attr:{style:true}},
//FLASH HTML
        TEXTFORMAT:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
//FRAMEWORK
        DATAGRID:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        CMVIEWSTACK:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        LISTVIEW:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        PAGE:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        PANEL:{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}},
        "PAGE-BODY":{whitelist:false, block:false, inline:false, void:false, meta:false, attr:{}}
    };
    const encodeEntities = {
        '‘' : '&lsquo;',
        '’' : '&rsquo;',
        '“' : '&ldquo;',
        '”' : '&rdquo;',
        '–' : '&ndash;',
        '—' : '&mdash;',
        '…' : '&hellip;',
        '™' : '&trade;',
        '£' : '&pound;',
        '€' : '&euro;',
        '©' : '&copy;',
        '®' : '&reg;',
        '°' : '&deg;',
        '¼' : '&frac14;',
        '½' : '&frac12;',
        '¾' : '&frac34;'
    };
    var inheritedCSS = {
        "border-collapse":true,
        "border-spacing":true,
        "caption-side":true,
        "color":true,
        "cursor":true,
        "direction":true,
        "empty-cells":true,
        "font-family":true,
        "font-size":true,
        "font-style":true,
        "font-variant":true,
        "font-weight":true,
        "font-size-adjust":true,
        "font-stretch":true,
        "font":true,
        "letter-spacing":true,
        "line-height":true,
        "list-style-image":true,
        "list-style-position":true,
        "list-style-type":true,
        "list-style":true,
        "orphans":true,
        "quotes":true,
        "tab-size":true,
        "text-align":true,
        "text-align-last":true,
        "text-decoration-color":true,
        "text-indent":true,
        "text-justify":true,
        "text-shadow":true,
        "text-transform":true,
        "visibility":true,
        "white-space":true,
        "widows":true,
        "word-break":true,
        "word-spacing":true,
        "word-wrap":true
    };
    var whitespaceCSS = {
        "font-size":true,
        "letter-spacing":true,
        "line-height":true,
        "white-space":true,
        "margin":true,
        "padding":true,
    };
    //endregion
    //region Sanitizer
    /**
     * Document structure
     * - body container: <article contenteditable="true">, <td>, <blockquote>, <code>, <ul>, <ol>
     * - block node: <div>, <p>, <li>, <h1-6>, <table>
     * - void block node: <hr>
     * - inline node: <span>, <b>, <font>, <a>, ...
     * - void inline node: <img>
     * - text node: #text
     * A body container has block node, but <ul> has only <li> nodes.
     * A body container also has inline nodes sometimes. This inline nodes will be wrapped with <p>.
     * A block node only has inline nodes.
     * A inline nodes has another inline nodes
     * #text and void inline node doesn't have children.
     * Note:
     * Word structure OutlineElement ListContainerWrapper and div.SCXW******* with nested OutlineElement ListContainerWrapper
     * Word all lists are 'fake', lists with LI converted to block elements, some dont display bullets
     * Some Excel content contains dbl whitespace and linefeeds in the opening tag, so we cant easily clean it up with regex.
     * Excel/Sheets can contain a styles tag and make use of class attribute. So excluding them will screw things up.
     * But not excluding them has potential to affect the framework
     * and also not excluding class/styles from other pasted content like the framework has potential to mess with existing framework functionality
     * OpenOffice all tags are upperCase - no distinction between doc and sheet lots of superfluous tabs and linefeed- leaves most attributes unquoted...
     *
     * We have normalized on \n and \u0A00
     * @return {HTMLSanitizer}
     * @constructor
     */
    var HTMLSanitizer = function() {
        var self = this;
        //the config for the current sanitize run
        var config;
        //coalesce all types of ws to a single space
        var nbspCharRegExp = /(\s|&nbsp;)+/g;
        var rxEmptyEditor = /^<p>(<br>)*(<em class="rte-marker"><\/em>)?<br><\/p>/i;
        var rxNewLine = /(\r\n|\n|\r)/g;
        var rxDblWS = /  +/g;
        var rxTab = /\t/g;
        var rxPTag = /<p><\/p>/g;
        // White space codes.
        var tab = 0x9;
        var space = 0x20;
        var zeroWidthSpace = 0x200b;
        var zeroWidthNonJoiner = 0x200c;
        var nbsp = '\u00A0';

        // Characters.
        var tabChar = '\t';
        var lineFeedChar = '\n';
        var spaceChar = ' ';

        //VML Image extraction
        var rxLinkProtocol = /^(https?:|ftps?:|)\/\//i;
        var rxLinkExtURL = /^((ftp|http|https):\/\/[a-z\u0080-\u009f\u00a1-\uffff0-9-_\\.]{1,})((:[0-9]{1,5})|)(((\/|\\?|#)[a-z\u00a1-\uffff0-9@?\\|!^=%&amp;\/~+#-\\'*-_{}]*)|())$/gi;
        var mapVSID = {};


        //flex cleaner - this has to be uppercase as only flex html is uppercase and this kills any style attribute
        var rxFontAttr = /(face|size|color|letterspacing|kerning|leading)="([^"]*?)"/gi;
        var rxFontAttr2 = /(style)='([^']*?)'/gi;
        var rxFontSize = /size=(\d)/i;
        var fontSizeMap = ['', 'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large',  'xx-large'];

        //Excel wrappers
        var excelStart = '<html><body><!--StartFragment--><div>';
        var excelEnd = '</div><!--EndFragment--></body></html>';
        //doc type identifier
        var rxDetectGDocs = /id=(["'])?docs-internal-guid-/;
        var rxDetectWord = /(class="?OutlineElement|class="?Mso|style=[^>]*?\bmso-|w:WordDocument|<o:\w+>)/;
        //"OpenOffice 4.1.7  (Win32)"//"Sheets"//"Microsoft Excel 15"//"Microsoft Word 15"
        var rxDetectGenerator = /<meta name="?generator"? content="([^"]*?)"\/?>/i;
        //IE11 doesnt support dotALL
        var cleaners = {
            None:[
                //outlook appears as None has no markings, no styles, no rtf..has no head has body contains fragment
                //check for fragment and/or body
                {rx: /^([\s\S]*?)<!--StartFragment-->([\s\S]*?)<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                {rx: /^([\s\S]*?)(<body(?:[^>]*?)>([\s\S]*?)<\/body[^>]*>)([\s\S]*)$/g, str:'$3'},
                //anything either side of the body tag is gone now remove any remaining meta tag
                {rx: /<meta([^>]*?)>/gi, str:''},
                //we nuke styles in unknown source
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/i, str:''},
            ],
            GDocs: [
                //has fragment marker, no generator and no style and no html/head/body
                {rx: /^([\s\S]*?)<!--StartFragment-->([\s\S]*?)<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                //anything either side of the body tag is gone now remove any remaining meta tag
                {rx: /<meta([^>]*?)>/gi, str:''},
                //google docs wraps everything in a b tag with the font weight normal
                {rx: /^<b[^>]*>(.*)<\/b>(<br[^>]*>)?$/, str:'$1'},
            ],
            Sheets: [
                //has fragment marker, generator, style and no html/head/body, style is in the fragment
                {rx: /^([\s\S]*?)<!--StartFragment-->([\s\S]*?)<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                // Remove anything but the contents within the styles element from excel
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/i, str:extractStylesheet},
                //anything either side of the body tag is gone now remove any remaining meta tag
                {rx: /<meta([^>]*?)>/gi, str:''},
            ],
            OpenOffice: [
                // No fragment marker, has generator, has style use BODY, adds random returns and tabs into the text content these need replacing with a space
                // Remove anything but the contents within the styles element from excel
                // open office doesnt terminate the last style
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/i, str:extractStylesheet},
                // Remove anything but the contents within the BODY element from word/google
                {rx: /^([\s\S]*?)(<BODY(?:[^>]*?)>([\s\S]*?)<\/BODY[^>]*>)([\s\S]*)$/g, str:'$3'},
                // returns between tags all need removing
                {rx: />\n</g, str:'><'},
                {rx: /\n/g, str:' '},
                // Seems to be a lot of tabs where there should be just a space i.e. 'Google\tDocs'
                {rx: /\t+/g, str:' '},
                {rx: /  +/g, str:' '},
            ],
            Cocoa:[
                //no fragment marker std html/head/body has generator 'Cocoa HTML Writer'
                // 1 style tag all need scopeing as usual
                // keep all classes
                //get rid of all returns so other regex works
                //{rx: /\r/g, str:''},
                //{rx: /\b(\n)\b/gm, str:' '},
                {rx: /\n/g, str:''},
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/i, str:extractStylesheet},
                //extract everything inside the body
                {rx: /^([\s\S]*?)(<body(?:[^>]*?)>([\s\S]*?)<\/body[^>]*>)([\s\S]*)$/g, str:'$3'},
            ],
            MSOffice:[
                //has fragment marker inside the body and generator no secondary but lots of ws and style tag is outside the frag
                // 3 style tags behavior/font/styles
                // most attributes are not quoted but some use " or '
                // keep all classes
                //get rid of all returns so other regex works
                //{rx: /\r/g, str:''},
                {rx: /\b(\n)\b/gm, str:' '},
                {rx: /\n/g, str:''},
                //remove vml 1+
                {rx: /v:shape\s*?id="([^"]*?)" o:spid="([^"]*?)"/g, str:extractVMLID},
                {rx: /<!--\[if gte vml 1]>(.*?)<!\[endif]-->/g, str:''},
                //remove mso wrappers
                {rx: /<!--\[if gte mso \d+\]>/gi, str:''},
                {rx: /<!\[endif]-->/gi, str:''},
                //remove none vml wrappers
                {rx: /<!\[if !vml]>/g, str:''},
                {rx: /<!\[endif]>/gi, str:''},
                //extract styles
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/gi, str:extractStylesheet},
                //now fetch the fragment content
                {rx: /^([\s\S]*)?<!--StartFragment-->([\s\S]*)?<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                //get rid of all whitespace so the stylesheet regex works
                {rx: /\t+/g, str:' '},
                {rx: /  +/g, str:' '},

                // Remove anything but the contents within the BODY element from word/google
                //{rx: /^(.*)?(<body(?:[^>]*)?>(.*)?<\/body[^>]*>)(.*)$/g, str:'$3'},
                //{rx: /<o:p>\s*<\/o:p>/g, str:''},
                //{rx: /<o:p>.*?<\/o:p>/g, str:''},
                //structured document tag
                {rx: /<w:Sdt[^>]*?>/g, str:''},
                {rx: /<\/w:Sdt>/g, str:''},
                /**
                 * Replaces last space preceding elements closing tag with `&nbsp;`. Such operation prevents spaces from being removed
                 * This method also takes into account Word specific `<o:p></o:p>` empty tags.
                 * Additionally multiline sequences of spaces and new lines between tags are removed (see #39 and #40).
                 */
                {rx: /<span(?: style='mso-spacerun:yes')?>(\s+)<\/span>/g, str:normalizeSafariSpaceSpans},
                // Remove all \r\n from "spacerun spans" so the last replace line doesn't strip all whitespaces.
                {rx: /(<span style='mso-spacerun:yes'>[\s]*?)[\r\n]+(\s*<\/span>)/g, str:'$1$2'},
                // strip empty spaceruns
                {rx: /<span style='mso-spacerun:yes'><\/span>/g, str:''},
                // Remove <o:p> block filler from empty paragraph.
                {rx: / <o:p><\/o:p>/g, str:'\u00A0<o:p></o:p>'},
                {rx: /<o:p>\u00A0<\/o:p>/g, str:'\u00A0'},
                // Microsoft Word makes these odd tags, like <o:p></o:p>
                {rx: /<\/?o:[a-z]*>/gi, str:''},

                // Microsoft Word adds some special elements around list items
                {rx: /<!\[if !supportLists\]>(((?!<!).)*)<!\[endif]\>/gi, str:'$1'},

                //{cleaner:'style', rx:false, str:''},
                //{rx: /(\n|\r| class=(")?Mso[a-zA-Z]+(")?)/g, str:' '},
                //{rx: /mso-[^;]*?;/g, str:''},
                //more word crap
                {rx: /<(\/)*(link|\\\\?xml:|st1:|o:)(.*?)>/gi, str:''},
            ],
            MSOneNote:[],
            MSExcel:[
                //{rx: /(\r\n|\r)/g, str:'\n'},
                // online has fragment marker plus inside a secondary html/head/body and generator and stylesheet and further unneeded fragment markers
                // desktop has only one fragment and its wrong and has html/head/body its the same as the inner secondary content of online
                {rx: /^([\s\S]*?)<!--StartFragment-->([\s\S]*?)<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                // Remove anything but the contents within the styles element from excel
                //{rx: /<style[^>]*?>([\s\S]*?)<\/style>/i, str:extractStylesheet},
                // Remove anything but the contents within the body element from word/google
                {rx: /^(\s\S*?)(<body(?:[^>]*?)>(\s\S*?)<\/body[^>]*>)(\s\S*)$/g, str:'$3'},
                //get rid of all returns and whitespace so the stylesheet regex works
                {rx: /\n/g, str:''},
            ],
            MSWord: [
                // has fragment marker, no generator and no style and no html/head/body
                // we can strip all classes
                // all attributes are quoted
                {rx: /^([\s\S]*?)<!--StartFragment-->([\s\S]*?)<!--EndFragment-->([\s\S]*)$/m, str:'$2'},
                // Remove anything but the contents within the BODY element from word/google
                //{rx: /^(.*?)(<body(?:[^>]*?)>(.*?)<\/body[^>]*>)(.*)$/g, str:'$3'},
                //{rx: /<o:p>\s*<\/o:p>/g, str:''},
                //{rx: /<o:p>.*?<\/o:p>/g, str:''},
                {rx: /\n/g, str:''},
                /**
                 * Replaces last space preceding elements closing tag with `&nbsp;`. Such operation prevents spaces from being removed
                 * This method also takes into account Word specific `<o:p></o:p>` empty tags.
                 * Additionally multiline sequences of spaces and new lines between tags are removed (see #39 and #40).
                 */
                {rx: /<span(?: style='mso-spacerun:yes')?>(\s+)<\/span>/g, str:normalizeSafariSpaceSpans},
                // Remove all \r\n from "spacerun spans" so the last replace line doesn't strip all whitespaces.
                {rx: /(<span style='mso-spacerun:yes'>[\s]*?)[\r\n]+(\s*<\/span>)/g, str:'$1$2'},
                // strip empty spaceruns
                {rx: /<span style='mso-spacerun:yes'><\/span>/g, str:''},
                // Remove <o:p> block filler from empty paragraph.
                {rx: / <o:p><\/o:p>/g, str:'\u00A0<o:p></o:p>'},
                {rx: /<o:p>\u00A0<\/o:p>/g, str:'\u00A0'},
                // Microsoft Word makes these odd tags, like <o:p></o:p>
                {rx: /<\/?o:[a-z]*>/gi, str:''},

                // Microsoft Word adds some special elements around list items
                {rx: /<!\[if !supportLists\]>(((?!<!).)*)<!\[endif]\>/gi, str:'$1'},
                //{rx: /(class="Mso[a-zA-Z]+)"/g, str:' '},

                {rx: /<span[^>]*?>\u00A0<\/span><\/span>/g, str:' </span>'},
                //word live doesnt have a stylesheet so strip all classes but we need some to find lists
                //{rx: /(class="([^"]*?)")/g, str:''},
                //more word crap
                {rx: /<(\/)*(link|\\\\?xml:|st1:|o:)(.*?)>/gi, str:''},
            ],
            Other: [//not run
                // Convert dbl ws to single ws - CANT USE THIS WITH MSO IN CASE THERE IS ANY dbl ws IN A TAG
                //{rx: /  +/g, str:'&nbsp;'},
                // Convert linefeed to br tag - CANT USE THIS WITH MSO IN CASE THERE IS ANY linefeed IN A TAG
                //{rx: /(\r\n|\n|\r)/g, str:'<br>'},
            ],
            cssPre:[
                //for cleaning the style tag contents
                //strip all the returns out for easy parsing
                {rx: /\n/g, str:'' },
                //remove comment tags
                {rx: /<!--(.*?)-->/g, str:'$1'},
                //remove comment tags
                {rx: /\/\*(?:(?!\*\/)[\s\S])*\*\//g, str:'' },
                //strip all font face out
                {rx: /@font-face\t{([^}]*?)}/g, str:patchFontFace},
                //make sure all rules are terminated - OpenOffice
                {rx: /{([^}]*?)}/g, str:patchRule},
                //strip out behaviors
                {rx: /(\.shape|[vow]\\:\*)\s*\{behavior:url\(#default#VML\);}\s?/g, str:''},
            ],
            cssMain:[
                //for cleaning the style tag contents
                //remove comment tags
                {rx: /[\n\t]+/g, str:'' },
                //add returns to put everything on seperate lines
                {rx: /([{};])/g, str:'$1\n'},
                {rx: /{/g, str:'\n{'},
                //fix word sans/serif error
                {rx: /font-family:(.*?)Sans",serif;$/gm, str:'font-family:$1Sans",sans-serif;\n'},
                //remove mso rules
                {rx: /mso-.*\n/g, str:''},
                //remove office lists
                {rx: /@list[^{]*?{[^}]*?}\n/g, str:''},
            ],
            style:[
                //remove word selection highlight background and spellcheck underline

                //for cleaning style attributes
                //make sure  leading '" ' and trailing ';"' and the attribute is lowercased
                {rx: /style='/g, str:'style=\' '},
                {rx: / -webkit-font-smoothing:[^;]*?;/g, str:''},
                {rx: / -webkit-text-stroke-width:[^;]*?;/g, str:''},
                {rx: / -webkit-user-drag:[^;]*?;/g, str:''},
                {rx: / -webkit-tap-highlight-color:[^;]*?;/g, str:''},
                {rx: / visibility:[^;]*?;/g, str:''},
                {rx: / user-select:[^;]*?;/g, str:''},
                {rx: / cursor:[^;]*?;/g, str:''},
                {rx: / direction:[^;]*?;/g, str:''},
                {rx: / position:[^;]*?;/g, str:''},
                {rx: / letter-spacing: ?normal;/g, str:''},
                {rx: / font-kerning:[^;]*?;/g, str:''},
                {rx: / font-variant-ligatures:[^;]*?;/g, str:''},
                {rx: / font-variant-caps:[^;]*?;/g, str:''},
                {rx: / orphans:[^;]*?;/g, str:''},
                {rx: / widows:[^;]*?;/g, str:''},
                {rx: / word-spacing:[^;]*?;/g, str:''},
                {rx: / vertical-align: ?baseline;/g, str:''},
                {rx: / overflow:[^;]*?;/g, str:''},
                {rx: / page-break-after:[^;]*?;/g, str:''},
                {rx: / padding: ?0px;/g, str:''},
                {rx: / margin: ?0px;/g, str:''},
                {rx: / margin-left:0in;/g, str:''},
                {rx: / text-indent:-[^;]*?;/g, str:''},
                {rx: / z-index:[^;]*?;/g, str:''},
                {rx: / text-indent:-[^;]*?;/g, str:''},
                {rx: / tab-stops:[^;]*?;/g, str:''},
                {rx: / text-underline:[^;]*?;/g, str:''},
                {rx: / panose:[^;]*?;/g, str:''},
                {rx: / font-family:\s?"([^\s]*?)"/g, str:'font-family: $1'},
                //word copies in the selection highlight background on paste
                {rx: / background-color: ?rgba\(198, 198, 198, 0.7\) \!important;/g, str:''},
                {rx: / background-repeat: ?repeat-x; background-position: ?left bottom; background-image: ?url\("data:image\/gif; ?base64,R0lGODlhBQAEAJECAP\/\/\/\/8AAAAAAAAAACH5BAEAAAIALAAAAAAFAAQAAAIIlGAXCCHrTCgAOw=="\); border-bottom: ?1px solid transparent; background-color: ?inherit;/g, str:''},
                //{rx: / border-bottom: 1px solid transparent;/g, str:''},
                //{rx: / background-color: inherit;/g, str:''},
                //strip office prefixed styles - run last as this will remove ws and trim
                {rx: / mso-[^:]*?:[^;]*?;/g, str:''},
                {rx: /style=' /g, str:'style=\''},
                {rx: /style='\s*'/g, str:''},
            ],
            commonPre:[
                //common pre clean mainly normalize and pull out any stylesheets
                //convert all \r to \n
                {rx: /(\r\n|\r)/g, str:'\n'},
                //safari uses \u00AA0 windows uses &nbsp; normalize here
                {rx: /&nbsp;/g, str:'\u00A0'},
                //strip out the stylesheet
                {rx: /<style[^>]*?>([\s\S]*?)<\/style>/gmi, str:extractStylesheet},
                //now process stylesheet
                {run:getStylesheet, rx: null, str:''},
                //convert return between words to spaces.....
                {rx: /\b(\n)\b/gm, str:' '},
                //convert office '' quoted style attributes, fix spacing and quoted fonts
                {rx: /STYLE=/g, str:' style='},
                {rx: /style="([^"]*?)"/g, str:normalizeQuotedStyle},
                {rx: /style='([^']*?)'/g, str:normalizeStyleAttribute},
            ],
            commonMain: [
                //by the time we get here we need all linefeeds removed.....
                //now clean all style attributes
                {cleaner:'style', rx:false, str:''},
                //clean classes
                {rx: /class="([^"]*?)"/g, str:cleanClasses},
                //remove empty classes
                {rx: /class=""/g, str:''},

                //the rest are all tag based cleaners
                //remove all other comments
                {rx: /<!--(.*?)-->/g, str:''},
                {rx: /<!\[(.*?)\]>/g, str:''},

                //top and tail of google docs
                {rx: /<[^>]*docs-internal-guid[^>]*>/gi, str:''},
                //convert any no std br to std br in google and word (Apple-interchange-newline)
                {rx: /<br([^>]+)?>/gi, str:'<br>'},

                // Google has a trailing BR element after each P/DIV tag but sets top/btm margin on the P tag to zero
                {rx: /<\/p><br>/gi, str:'<\/p><p><br></p>'},

                {rx: /<h[1-6]><\/h[1-6]>/gi, str:''},
                {rx: /<\/div><br>/gi, str:'<\/div><p><br></p>'},

                {rx: /<[auibs]([^>]*)?><\/[auibs]>/gi, str:''},
                {rx: /<([bius]) ([^>]*)?>/gi, str:'<$1>'},
                {rx: /<b>/gi, str:'<span style=\'font-weight:bold;\'>'},
                {rx: /<i>/gi, str:'<span style=\'font-style:italic;\'>'},
                {rx: /<u>/gi, str:'<span style=\'text-decoration:underline;\'>'},
                {rx: /<s>/gi, str:'<span style=\'text-decoration:line-through;\'>'},
                {rx: /<\/[bius]>/gi, str:'<\/span>'},


                // un-html spaces and newlines inserted by OS X
                {rx: /<span(?: class="Apple-converted-space"|)>(\s+)<\/span>/g, str:normalizeSafariSpaceSpans},

                // Remove all whitespaces when they contain any \r or \n.
                {rx: />(\s*[\r\n]\s*)</g, str:'><'},

                //replace space before closing tag with nbsp problem when we then unwrap the element in word.......
                //{rx: / <\//g, str:'\u00A0</'},

                //  FIXME: replace manually entered b/i/a tags with real ones
                //{rx: /&lt;(\/?)(i|b|a)&gt;/gi, str:'<$1$2>'},

                // FIXME: replace manually a tags with real ones, converting smart-quotes from google docs
                //{rx: /&lt;a(?:(?!href).)+href=(?:&quot;|&rdquo;|&ldquo;|"|“|”)(((?!&quot;|&rdquo;|&ldquo;|"|“|”).)*)(?:&quot;|&rdquo;|&ldquo;|"|“|”)(?:(?!&gt;).)*&gt;/gi, str:'<a href="$1">'},

                // Newlines between paragraphs in html have no syntactic value,
                // but then have a tendency to accidentally become additional paragraphs down the line
                {rx: /<\/p>\n+/gi, str:'</p>'},
                {rx: /\n+<p/gi, str:'<p'},


                // Convert tabs to 8 spaces - actualy google uses a tab but wraps in pre-wrap
                {rx: /\t/g, str:'\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0'},
                // sometimes word doesnt pass the tab but the list delimiter and a single space,
                // gdocs sometimes uses delimeter and spaces in a 'pre' style container
                // sometimes the user just uses a delimeter and space because thats • what • they • want
                {rx: /• /g, str:'•\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0'},

                //make sure any empty p tags get populated
                {rx: /<p><\/p>/g, str:'<p><br></p>'},


                //framework and form related
                {rx: /<(fieldset)/g, str:'<$1 class="ignore" '},
                {rx: /<(form)/gi, str:'<div'},
                {rx: /<\/(form)/gi, str:'</div'},
                {rx: /<(input)/gi, str:'<br'},
                {rx: /(class="[^"]*)bind/g, str:'$1'},

                {rx: /<font([^>]*?)>/gi, str:fontStyle},
                {rx: /<\/font>/gi, str:'</span>'},
                {rx: /<\/?textformat\W*>/gi, str:''},
                //strip empty spans
                {rx: /<span([^>]*?)><\/span>/gi, str:''},
                {rx: /<\/span><span[^>]*?>&nbsp;<\/span>/g, str:'&nbsp;</span>'},


                //  FIXME: convert any free url links to 'a' tags (a link that isnt in a src or preceeded by a >, its fooled by a span...)
                {rx: /originalsrc="([^"]*?)"/g, str:''},
                {rx: /title="[^"]*?((https?|sftp|ftp|file|nfs|smb|afp|git):\/\/[\-A-Za-z0-9+&@#\/%?=~_|!:,.;]*[\-A-Za-z0-9+&@#\/%=~_|??])[^"]*?"/g, str:''},
                {rx: /(^|[^>"=])((https?|sftp|ftp|file|nfs|smb|afp|git):\/\/[\-A-Za-z0-9+&@#\/%?=~_|!:,.;]*[\-A-Za-z0-9+&@#\/%=~_|??])/gi, str:'$1<a href="$2" target="_blank" rel="noopener noreferer">$2</a>'},
                //find any possible link inside a tag and flag it....
                {rx: /(>)((https?|sftp|ftp|file|nfs|smb|afp|git):\/\/[\-A-Za-z0-9+&@#\/%?=~_|!:,.;]*[\-A-Za-z0-9+&@#\/%=~_|??])/gi, str:' data-link="$2">$2'},
            ]
        };

        var emptyEditorElement;

        /**
         * set up empty editor element so we can clone it
         */
        function init(){
            emptyEditorElement = document.createElement('p');
            emptyEditorElement.appendChild(document.createElement('br'));
        }
        init();
        /**
         * Extract shape spid's and map to img id
         * @param $0
         * @param $1
         * @param $2
         * @return {*}
         */
        function extractVMLID($0, $1, $2){
            mapVSID[$1] = $2;
            return $0;
        }
        /**
         * regex helper to parse out the attributes on a font tag
         * @param {string} $0
         * @param {string} $1 the attributes of a font tag
         * @return {string}
         */
        function fontStyle($0, $1){
            var style = $1.replace(rxFontSize,'size="$1"').replace(rxFontAttr, fontAttr2Style).replace(rxFontAttr2, fontAttr2Style);
            return '<span style=\''+style+'\'>';
        }
        /**
         * Internal Regex Helper for flash html
         * @param {string} $0
         * @param {string} $1 the attribute name
         * @param {string} $2 the attribute value
         * @returns {string}
         */
        function fontAttr2Style($0, $1, $2) {
            switch($1.toLowerCase()) {
                case 'face': return 'font-family:' + $2 + ';';
                case 'size':
                    var size = parseInt($2);
                    //openoffce 10pt, also 15px also 1 also x-small
                    //if NaN its large/small etc otherwise if 1-7 it is a text size otherwise its a size in pixel
                    var fontSize = isNaN(size) ? $2 : size > 7 ? size + 'px' : fontSizeMap[size];
                    return 'font-size:' + fontSize + ';';
                case 'color': return 'color:' + $2 + ';';
                case 'letterspacing': return '';
                case 'kerning': return '';
                case 'leading': return '';
                case 'style': return ($2 + ';').replace(';;',';').trim();
                default:
                    return '';
            }
        }
        /**
         * patch css font family declartion from word
         * @param $0
         * @param $1
         * @return {string}
         */
        function normalizeQuotedStyle($0, $1){
            return 'style=\'' + $1.replace(rxFixQuot, '"') + '\'';
        }
        var rxStripSemiSpace = /;\s/g;
        var rxSemi = /;/g;
        var rxColonSpace = /: /g;
        var rxFixQuot = /&quot;|'/g;
        var rxWords = /[\w-]+/g;
        /**
         * check class validity
         * @param {string} $0
         * @return {boolean}
         */
        function validClassHelper($0){
            return config.validClasses.indexOf($0)===-1 ? '' : $0;
        }
        /**
         * normalize style attribute
         * @param $0
         * @param $1
         * @return {string}
         */
        function cleanClasses($0, $1){
            var s = $1.replace(rxWords, validClassHelper).trim();
            return s ? 'class="' + s + '"' : '';
        }
        /**
         * normalize style attribute
         * @param $0
         * @param $1
         * @return {string}
         */
        function normalizeStyleAttribute($0, $1){
            //normalize leading and trailing spaces and ';'
            return 'style=\' ' + (($1.trim() + ';').replace(rxNewLine,'').replace(';;', ';').replace(rxStripSemiSpace, ';').replace(rxSemi, '; ').replace(rxColonSpace, ':').replace(rxFixQuot, '\'')).trim() + '\'';
        }
        var rxIsBadFontFamily1 = /^\\?@/;
        var rxIsBadFontFamily2 = /\\?;.*$/;
        var rxFontFamilyU = /font-family:([^;]*?);/;
        var rxFontFamilyQ = /font-family:(?:[^"]*?)"([^"]*?)";/;
        var rxIsFontFamilyQ = /font-family:\s*"/;
        /**
         * patch css font family declartion from word
         * the font-family may or may not be quoted - if quoted it may or may not be corrupt
         * look for quoted first ..
         * @param $0
         * @param $1
         * @return {string}
         */
        function patchRule($0, $1){
            return '{ ' + ($1.trim() + ';').replace(';;',';') + '}' ;
        }
        /**
         * patch css font family declartion from word
         * the font-family may or may not be quoted - if quoted it may or may not be corrupt
         * look for quoted first ..
         * @param $0
         * @param $1
         * @return {string}
         */
        function patchFontFace($0, $1){
            var value = '';
            if(rxIsFontFamilyQ.test($1)){
                value = rxFontFamilyQ.exec($1)[1];
                if(rxIsBadFontFamily1.test(value)){
                    //bad prefix
                    var rxBad1 = new RegExp(value.replace(/\\/, '\\\\'),'g');
                    var rxBad2 = new RegExp(value.replace(/\\/, ''),'g');
                    var good = value.replace(rxIsBadFontFamily1,'');
                    config.fixfonts.push({bad1:rxBad1, bad2:rxBad2, good:good});
                    value = good;
                }else if(rxIsBadFontFamily2.test(value)){
                    //bad suffix
                    var rxBad1 = new RegExp(value.replace(/\\/, '\\\\'),'g');
                    var rxBad2 = new RegExp(value.replace(/\\/, ''),'g');
                    var good = value.replace(rxIsBadFontFamily2,'');
                    config.fixfonts.push({bad1:rxBad1, bad2:rxBad2, good:good});
                    value = good;
                }else{
                    value = value;
                }
                value = '"' + value.trim() + '"';
            }else{
                value = rxFontFamilyU.exec($1)[1];
                if(rxIsBadFontFamily1.test(value)){
                    //bad prefix
                    var rxBad1 = new RegExp(value.replace(/\\/, '\\\\'),'g');
                    var rxBad2 = new RegExp(value.replace(/\\/, ''),'g');
                    var good = value.replace(rxIsBadFontFamily1,'');
                    config.fixfonts.push({bad1:rxBad1, bad2:rxBad2, good:good});
                    value = good;
                }else if(rxIsBadFontFamily2.test(value)){
                    //bad suffix
                    var rxBad1 = new RegExp(value.replace(/\\/, '\\\\'),'g');
                    var rxBad2 = new RegExp(value.replace(/\\/, ''),'g');
                    var good = value.replace(rxIsBadFontFamily2,'');
                    config.fixfonts.push({bad1:rxBad1, bad2:rxBad2, good:good});
                    value = good;
                }else{
                    value = value;
                }
                value = '' + value.trim() + '';
            }
            config.fonts.push('@font-face{font-family: ' + value + ';}');
            return '';
        }
        /**
         * Internal regex helper
         * Extract a stylesheet from the markup and store in the styles array to inject after cleaning is complete
         * @param {string} $0 the style tag
         * @param {string} $1 the style rules
         * @return {string} the replacemnt an empty string
         */
        function extractStylesheet($0, $1) {
            config.styles.push($1);
            return '';
        }
        /**
         * Compile the stylesheets into a single style tag
         * @param {string} html
         * @returns {string}
         */
        function getStylesheet(html){
            //concat the styles and process rules - remove comments and strip fontface
            var rules = cleaners['cssPre'].reduce(execCleaner, config.styles.join('\n'));
            //now fix andy bad word fonts...
            for (var j = 0, jlen = config.fixfonts.length; j < jlen; j++) {
                var fix = config.fixfonts[j];
                rules = rules.replace(fix.bad1, fix.good).replace(fix.bad2, fix.good);
                html = html.replace(fix.bad1, fix.good).replace(fix.bad2, fix.good);
            }
            //run main css cleaners
            rules = cleaners['cssMain'].reduce(execCleaner, rules);
            //all classes must start with an alpha
            var pfx = config.pfx = 'pfx-' + $.getGUID(true);
            //add the pfx as a valid class
            config.validClasses.push(pfx);
            //find all the classes and add as valid classes for check later
            config.validClasses = config.validClasses.concat(parseClasses(rules)).filter(onlyUnique);
            //prefix all selectors
            var ss = config.fonts.join('\n') + '\n' + prefixCSSSelectors(rules, '.usr .' + pfx) + '\n';

            //add the stylesheet
            config.stylesheet = ss ? '<style data-title="' + pfx + '">' + ss + '</style>' : '';
            //return cleaned up markup
            return html;
        }
        /**
         * Prefix all rules with the usr class so that the stylesheet is scoped
         * @param rules
         * @param className
         * @return {string}
         */
        function prefixCSSSelectors(rules, className) {
            //@ rules font-face, media, import, font-feature-values
            // makes sure the className will not concatenate the selector
            className += ' ';
            var items = rules.split('}\n');
            for (var i = 0, ilen = items.length; i < ilen; i++) {
                var rule = items[i].split('\n{\n');
                if(rule.length===2) {
                    var selectors = rule[0].trim().split(', ');
                    for (var j = 0, jlen = selectors.length; j < jlen; j++) {
                        selectors[j] = className + selectors[j];
                    }
                    rule[0] = selectors.join(', ');
                    items[i] = rule.join('\n{\n');
                }
            }
            return items.join('}\n');
        }
        /**
         * Array helper remov duplicates from array
         * @param {*} value
         * @param {number} index
         * @param {Array} self
         * @return {boolean}
         */
        function onlyUnique(value, index, self) {
            return self.indexOf(value) === index;
        }

        /**
         * We want to extract only the elements with classes defined in the stylesheets:
         *
         * @param rules
         * @return {string[]}
         */
        function parseClasses( rules) {
            // rules are just the different rules in the style sheets
            // We want to merge them and them split them by commas, so we end up with only
            // the selectors
            return rules.replace(/{[^}]*?}/g,'')
                .replace(/@font-face/g,'')
                .replace(/@page/g,'')
                //lookbehind not supported xcpt chrome
                //.replace(/(?<!\.)\b\w*\b/g,'')
                // Remove selectors splitting the elements, leave only the class selector (.)
                .replace( /(,|>|\+|~)/g, ' ' )
                // Remove attribute selectors: table[border="0"]
                .replace( /\[[^\]]*/g, '' )
                // Remove Ids: div#main
                .replace( /#[^\s]*/g, '' )
                // Remove pseudo-selectors and pseudo-elements: :hover :nth-child(2n+1) ::before
                .replace( /\:{1,2}[^\s]*/g, '' )
                .replace(/\n/g,' ')
                //strip elements out
                .replace(/\s\b\w*\b/g,' ')
                .replace(/\./g,' ')
                .replace( /\s+/g, ' ' )
                .trim()
                .split(' ').filter(onlyUnique);
        }
        /**
         * find all possible links in the doc and check if they are inside an a tag if not wrap them
         * @param node
         */
        function fixLinks(node){
            var links = node.querySelectorAll('[data-link]');
            links.forEach(function(link){
                if(!link.closest('a')){
                    //not embedded in a link so lets wrap it.....
                    var tag = document.createElement('a');
                    tag.href = link.dataset.link;
                    tag.target = '_blank';
                    tag.setAttribute('rel', 'noopener noreferer');
                    link.parentNode.insertBefore(tag, link);
                    tag.append(link);
                }
            });
        }
        function copyToParentNodeStyles(node, toNode, excludes) {
            copyTextNodeStyles(node, toNode);
            excludes.display = true;
            for (var i = 0, len = node.style.length; i<len; i++) {
                var name = node.style[i];
                if(!excludes[name]){
                    var value = node.style.getPropertyValue(name);
                    if(value !== 'inherit') {
                        toNode.style.setProperty(name, value, node.style.getPropertyPriority(name));
                    }
                }
            }
            return node;
        }
        function copyFromParentNodeStyles(node, toNode, excludes) {
            copyTextNodeStyles(node, toNode);
            excludes.display = true;
            for (var i = 0, len = node.style.length; i<len; i++) {
                var name = node.style[i];
                if(!excludes[name]){
                    var value = node.style.getPropertyValue(name);
                    toNode.style.setProperty(name, value, node.style.getPropertyPriority(name));
                }
            }
            return node;
        }
        function copyOtherNodeStyles(node, toNode, excludes) {
            copyTextNodeStyles(node, toNode);
            excludes.display = true;
            for (var i = 0, len = node.style.length; i<len; i++) {
                var name = node.style[i];
                if(!excludes[name]){
                    var value = node.style.getPropertyValue(name);
                    toNode.style.setProperty(name, value, node.style.getPropertyPriority(name));
                }
            }
            return node;
        }
        /**
         * check whether the styles on a node are required or they just duplicate styles on the parent
         * @param {Node} node
         */
        function checkRequiredStyles(node){
            var parentNode = node.parentNode;
            if(node.hasAttribute('class')) {
                if (node.style && parentNode.style) {
                    var styles = Array.from(node.style);
                    for (var i = 0, len = styles.length; i < len; i++) {
                        var styleName = styles[i];
                        if (parentNode.style[styleName] === node.style[styleName]) {
                            node.style.removeProperty(styleName);
                        }
                    }
                }
                if(node.classList.length===0)node.removeAttribute('class');
            }
        }
        /**
         * Check if all styles in target exist in src
         * @param {Node} src
         * @param {Node} tgt
         * @param {boolean} strict
         * @return {boolean}
         */
        function checkStyleContains(src, tgt, strict){
            var srcStyles = src.style;
            var tgtStyles = tgt.style;
            if(srcStyles && tgtStyles){
                var names = strict ? Array.from(tgtStyles).concat(Array.from(srcStyles)).filter(onlyUnique) : tgtStyles;
                for (var i = 0, len = names.length; i < len; i++) {
                    var styleName = names[i];
                    if(srcStyles[styleName] !== tgtStyles[styleName]){
                        if(!strict && tgtStyles[styleName] === 'inherit'){
                            //ignore inherit rules when comparing to partent
                        }else {
                            return false;
                        }
                    }
                }
                return true;
            }
            return false;
        }

        /**
         * creat a copy excludes object
         * find all the styles in node that are the same or different in tgt or inherited insrc
         * @param srcNode
         * @param tgtNode
         * @param include if 'diff' 'eq' 'inherit
         * @param {Object} [diff]
         * @return {Object}
         */
        function styleExcludes(srcNode, tgtNode, include, exclude, diff){
            var result = diff || {};
            for (var i = 0, len = srcNode.style.length; i<len; i++) {
                var name = srcNode.style[i];
                if(include === 'diff' && srcNode.style[name] !== tgtNode.style[name]) result[name] = true;
                if(include === 'eq' && srcNode.style[name] === tgtNode.style[name]) result[name] = true;
                if(srcNode.style[name] === 'inherit'){
                    if(include === 'inherit')result[name] = true;
                    if(exclude === 'inherit')result[name] = false;
                }
            }
            return result;
        }

        function copyTextNodeStyles(node, toNode) {
            //copy style attributes from the node to its parent...
            if(node.nodeName==='B' || node.nodeName==='STRONG') toNode.style.setProperty('font-weight', 'bold');
            if(node.nodeName==='I' || node.nodeName==='EM') toNode.style.setProperty('font-style', 'italic');
            if(node.nodeName==='S' || node.nodeName==='STRIKE') toNode.style.setProperty('text-decoration', 'line-through');
            if(node.nodeName==='U') toNode.style.setProperty('text-decoration', 'underline');
        }
        /**
         * copy styles from node to another node, defaults to parent node
         * @param {Node} node
         * @param {Node} toNode
         * @param {Object} excludes
         * @return {Node} return from node for chaining
         */
        function copyNodeStyles(node, toNode, excludes){
            copyTextNodeStyles(node, toNode);
            excludes.display = true;
            for (var i = 0, len = node.style.length; i<len; i++) {
                var name = node.style[i];
                if(!excludes[name]){
                    var value = node.style.getPropertyValue(name);
                    toNode.style.setProperty(name, value, node.style.getPropertyPriority(name));
                }
            }
            return node;
        }
        //if not found return empty tag - always blacklist
        var emptyTag = {whitelist:false};
        /**
         *
         * @param {Node} node
         * @returns {Object}
         */
        function isNode(node){
            return node ? tags[node.nodeName] || emptyTag : emptyTag;
        }

        /**
         *
         * @param {Node} node
         * @returns {Object}
         */
        function isNodeParent(node){
            return node ? isNode(node.parentNode) : emptyTag;
        }
        /**
         *
         * @param {DocumentFragment} document
         * @param {Node} parentNode
         */
        function sanitize(document, parentNode) {
            var treeWalker = createTreeWalker(document, parentNode);
            var node = treeWalker.firstChild();

            if (!node) { return; }
            //walk the parentNode siblings untilwe hit the buffer
            do {
                if (node.nodeType === Node.TEXT_NODE) {
                    // If this text node is just whitespace and the previous or next element
                    // sibling is a block element, remove it
                    // N.B.: This heuristic could change. Very specific to a bug with
                    // `contenteditable` in Firefox: http://jsbin.com/EyuKase/1/edit?js,output
                    if (node.data.trim() === '' &&
                        ((node.previousElementSibling && tags[node.previousElementSibling.tagName].block) ||
                            (node.nextElementSibling && tags[node.nextElementSibling.tagName].block))
                    ) {
                        parentNode.removeChild(node);
                        sanitize(document, parentNode);
                        break;
                    } else {
                        continue;
                    }
                }

                // Remove all comments
                if (node.nodeType === Node.COMMENT_NODE) {
                    parentNode.removeChild(node);
                    sanitize(document, parentNode);
                    break;
                }

                // Drop tag if not whitelist, but save contents as text
                if (isNode(node).whitelist === false) {
                    var span = document.createElement('span');
                    parentNode.insertBefore(span, node);
                    span.innerText = node.innerText;
                    parentNode.removeChild(node);
                    sanitize(document, parentNode);
                    break;
                }
                // Drop tag if not whitelist but copy children up to parent.
                if (isNode(node).whitelist === null) {
                    //copy children to parent dont copy styles
                    unwrapElement( node , 'none');
                    sanitize(document, parentNode);
                    break;
                }

                if(node.nodeName === 'STYLE'){
                    continue;
                }

                //replace fake inline with a real inline
                if (isNode(node).block && node.style.display === 'inline') {
                    var span = document.createElement('span');
                    parentNode.insertBefore(span, node);
                    span.append(node);
                    //copy children to parent copy styles to parent
                    unwrapElement(node, 'parent');
                    //continue withe the newlys inserted span but after carry on with the current siblings
                    sanitize(document, span.parentNode);
                    break;
                }

                //table inside span inside div or div inside b inside div
                //BLOCK TABLE/OL/UL cant be unwrapped - INLINE A tags cant either
                if (isNode(node).block && isNodeParent(node).inline) {
                    //hoist the node - remove parent if then empty
                    if(parentNode.childNodes.length === 1 || parentNode.firstChild === node) {
                        parentNode.parentNode.insertBefore(node, node.parentNode);
                        sanitize(document, parentNode.parentNode);
                        break;
                    }
                }
                //unwrap nested inlines - dont if the tags are different and it has a class
                if (isNode(node).inline && isNodeParent(node).inline && node.nodeName !== 'A' && !isNode(node).void) {
                    var copyStyles = !checkStyleContains(node.parentNode, node);
                    if(node.parentNode.childNodes.length===1){
                        if(node.hasAttribute('class') || parentNode.hasAttribute('class')){
                            //dont unwrap because class attribute may cause styling difference
                        }else {
                            if(node.nodeName==='SPAN' && parentNode.nodeName !== 'A' && parentNode.nodeName !== 'SPAN'){
                                parentNode.parentNode.insertBefore(node, parentNode);
                                copyFromParentNodeStyles(parentNode, node, {});
                                parentNode.remove()
                                sanitize(document, parentNode);
                                break;
                            }else {
                                //copy children to parent and copy styles to parent continue with the parent
                                unwrapElement(node, 'parent');
                                sanitize(document, parentNode);
                                break;
                            }
                        }
                    }else if(!copyStyles){
                        //reset treewalker
                        treeWalker.previousSibling();
                        //unwrap nested inlines - copy node children to parent copy styles to child but no need if parent has all styles already
                        unwrapElement(node, 'child');
                        sanitize(document,parentNode);
                        break;
                    }
                }
                if(node.nodeName === 'SPAN'){
                    if(node.previousSibling && node.previousSibling.nodeName === 'SPAN'){
                        //roll up adjacent sibling
                        if(checkStyleContains(node.previousSibling, node, true)){
                            //reset treewalker
                            treeWalker.previousSibling();
                            //unwrap
                            unwrapSiblingElement( node, false);
                            sanitize(document,parentNode);
                            break;
                        }
                        //roll up adjacent spans if styles are the same and no class diff
                        // prev node styles are clean so we need to clean this nodes styles before comparing
                        checkRequiredStyles(node);

                        if(checkStyleContains(node.previousSibling, node, true)){
                            //reset treewalker
                            treeWalker.previousSibling();
                            //only unwrap if styles are the same
                            unwrapSiblingElement( node, 'none');
                            sanitize(document,parentNode);
                            break;
                        }
                    }
                    if(node.parentNode.nodeName === 'P' && node.parentNode.childNodes.length===1){
                        //copy node children to partent and copy styles on to parent - continue with the parent
                        unwrapElement(node, 'parent');
                        sanitize(document,parentNode);
                        break;
                    }
                }
                // Drop tag entirely but copy children up to parent.
                if (node.nodeName === 'P' && parentNode.nodeName === 'LI' && parentNode.childNodes.length === 1) {
                    //copy style attributes from the node to its parent...
                    if(!checkStyleContains(parentNode, node))copyToParentNodeStyles(node, parentNode, {display:true});
                    node.parentNode.style.display = 'list-item';
                    //copy children to parent and copy styles to parent
                    sanitize(document, unwrapElement( node, 'parent' ));
                    break;
                }
                //word sometimes sets padding to 0 and then the listmarker doesnt show up
                if(node.nodeName === 'UL' && parseInt(node.style.padding) === 0) node.style.padding = '';
                if(node.nodeName === 'OL' && parseInt(node.style.padding) === 0) node.style.padding = '';

                if(node.nodeName === 'DIV'){
                    // so word has fake lists - a list where the list-items are changed to block and dont display bullets
                    // structure is div.ListContainerWrapper(no style) > ul(no style) > li(has style) > p (real stuff)
                    if(node.classList.contains('ListContainerWrapper')){
                        var list = node.firstChild;
                        var li = list.firstChild;
                        if(li.dataset.leveltext===''){
                            //get each child and insert into current node
                            //unwrap....
                            do{
                                //copy li styles to firstchild exclude margin and display
                                //merge styles from li - remove block and margin (margin usually has a negative value to offset the list indent)
                                copyNodeStyles(li, li.firstChild, {margin:true, "margin-left":true, display:true});
                                //move node to top level parent div
                                node.appendChild(li.firstChild);
                            }while(li = li.nextSibling);
                            list.remove();
                        }
                    }
                    //ignore ul/li ol/li table/tbody/tr/td
                    //best div/div div/
                    //copy node child to parent and copy styles on to child
                    if(node.childNodes.length === 1 && (node.firstChild.nodeName === 'P' || node.firstChild.nodeName === 'DIV')) {
                        unwrapElement(node,'child');
                        ///copyNodeStyles(node, node.firstChild, {});
                        //parentNode.insertBefore(node.firstChild, node);
                        //node.remove();
                        //continue with the parent
                        sanitize(document, parentNode);
                        break;
                    }
                }

                if(node.nodeName === 'LI'){
                    node.style.display = 'list-item';
                    node.style.removeProperty('margin');
                }

                if(node.nodeName === 'IMG'){
                    if(!setImageSrc(node)){
                        parentNode.removeChild(node);
                        sanitize(document, parentNode);
                        break;
                    }
                }


                //were done shifting ownership so lets make sure we really need all these styles
                checkRequiredStyles(node);

                var allowedAttrs = getAllowedAttrs(node.nodeName, node);
                // Sanitize attributes
                for (var a = 0; a < node.attributes.length; a++) {
                    var attr = node.attributes[a];

                    if (shouldRejectAttr(attr, allowedAttrs, node)) {
                        node.removeAttribute(attr.name);
                        // Shift the array to continue looping.
                        a--;
                    }
                }
                //check if the classes are valid
                //checkValidClasses(node);
                //patch up A tag
                if(node.nodeName==='A'){
                    //force a target unless its an in page jump
                    if(!node.href.startsWith('#'))node.target='_blank';
                    // for security on 'wild' links in usr content
                    node.setAttribute('rel', 'noopener noreferer');
                }

                // Sanitize children
                sanitize(document, node);

            } while ((node = treeWalker.nextSibling()));
        }
        /**
         *
         * @param document
         * @param {Node} node
         * @return {TreeWalker}
         */
        function createTreeWalker(document, node) {
            return document.createTreeWalker(node, NodeFilter.SHOW_TEXT | NodeFilter.SHOW_ELEMENT | NodeFilter.SHOW_COMMENT, null, false);
        }

        /**
         *
         * @param nodeName
         * @param {Node} node
         * @return {*}
         */
        function getAllowedAttrs(nodeName, node){
            if (typeof tags[nodeName].attr === 'function') {
                return tags[nodeName].attr(node);
            } else {
                return tags[nodeName].attr;
            }
        }

        /**
         * Is this an allowed attribute
         * @param {Object} attr
         * @param {boolean|Object} allowedAttrs
         * @param {Node} node
         * @return {boolean}
         */
        function shouldRejectAttr(attr, allowedAttrs, node){
            var attrName = attr.name.toLowerCase();

            if (allowedAttrs === true){
                return false;
            } else if (typeof allowedAttrs[attrName] === 'function'){
                return !allowedAttrs[attrName](attr.value, node);
            } else if (typeof allowedAttrs[attrName] === 'undefined'){
                return true;
            } else if (allowedAttrs[attrName] === false) {
                return true;
            } else if (typeof allowedAttrs[attrName] === 'string') {
                return (allowedAttrs[attrName] !== attr.value);
            }

            return false;
        }

//region CleanNormalize
        /**
         * Is it a texnode or inline node
         * @param {Node} node
         * @return {boolean}
         */
        function checkTextOrInlineNode(node){
            //non ws text or inline not void
            return node ? (node.nodeType === Node.TEXT_NODE) || (!tags[node.nodeName].block && !tags[node.nodeName].void) : false;
        }

        /**
         * Is it empty
         * @param {Node} node
         * @return {boolean|*}
         */
        function isEmpty(node) {
            // Do not insert BR in empty non block elements with parent containing text
            return (node.children.length === 0 && !tags[node.nodeName].void && !RTEUtils.isSelectionMarkerNode(node) &&
                        (tags[node.nodeName].block ||  node.textContent === '')) ||
                (node.children.length === 1 &&  RTEUtils.isSelectionMarkerNode(node.children[0]));
        }

        /**
         * traverse the dom tree and find any empty or whitespace nodes and add a br so it can be edited..
         * @param {Node} parentNode
         */
        function traverse(parentNode) {
            // Instead of TreeWalker, which gets confused when the BR is added to the dom,
            // we recursively traverse the tree to look for an empty node that can have childNodes
            var node = parentNode.firstElementChild;
            while (node) {
                // Find any node that contains no child *elements*, or just contains
                // whitespace, is not self-closing and is not a custom element
                // here a scenario - excel put the wrongs font size on an empty cell...
                // since its empty it has no size which is what they want to look right.
                // If you put a <br> tag in it, it will assume the font-size and throw the layout off.
                // If you don't put a <br> in it it wont be editable unless it has dimensions from elsewhere
                // but you don't know that --- so what do you do.
                if (isEmpty(node)) {
                    if(node.tagName !== 'TD')node.appendChild(document.createElement('br'));
                } else if (node.children.length > 0) {
                    traverse(node);
                }
                node = node.nextElementSibling;
            }
        }
        /**
         * Check a container ensure all children are wrapped in a block container or else wrap in a p tag
         * Wrap consecutive text nodes or inline elements in a P element.
         * @param {Node} parentNode
         */
        function wrapChildNodes(parentNode) {
            //coalesce textnodes
            parentNode.normalize();
            var node = parentNode.firstChild;
            var pNode;
            do{
                if (node.tagName === 'BLOCKQUOTE') {
                    //main point is to stop return in a BQ from creating a new BQ
                    wrapChildNodes(node);
                }else if(checkTextOrInlineNode(node)){
                    if(pNode){
                        pNode.appendChild(node);
                        node = pNode;
                    }else{
                        pNode = document.createElement('P');
                        parentNode.insertBefore(pNode, node);
                        pNode.appendChild(node);
                        node = pNode;
                    }
                }else{
                    pNode = null;
                }
            } while (node = node.nextSibling);
        }
        /**
         * make sure all nodes are wrapped and any empty nodes have content
         * @param {HTMLElement} el
         * @memberOf HTMLSanitizer#
         */
        function normalize(el){
            /**
             * 1)
             * Set the default content of the editor so that each carriage return creates a P.
             * We have to begin with the following HTML '<p><br></p>', because otherwise some
             * browsers(?) will position the caret outside of the P when the editor is focused.
             */
            /**
             * 2)
             * Chrome and Firefox: All elements need to contain either text or a `<br>` to
             * remain selectable. (Unless they have a width and height explicitly set with
             * CSS(?), as per: http://jsbin.com/gulob/2/edit?html,css,js,output)
             */
            /**
             * 3)
             * Chrome and Firefox: Upon pressing backspace inside of a P, the
             * browser deletes the paragraph element, leaving the caret (and any
             * content) outside of any P.
             *
             * Firefox: Erasing across multiple paragraphs, or outside of a
             * whole paragraph (e.g. by ‘Select All’) will leave content outside
             * of any P.
             *
             * Entering a new line in a pristine state state will insert
             * `<div>`s (in Chrome) or `<br>`s (in Firefox) where previously we
             * had `<p>`'s. This patches the behaviour of delete/backspace so
             * that we do not end up in a pristine state.
             *
             * Ensure P mode.
             * Wrap any orphan text nodes in a P element.
             */
            if (el.childNodes.length === 0) {
                el.appendChild(emptyEditorElement.cloneNode(true));
            }else if (el.childNodes.length === 1 && isEmptyEditorNode(el.childNodes[0])) {
                return;
            }else {
                // normalize the root and then traverse the root children wrapping any text or inline nodes in p tags.
                wrapChildNodes(el);
                //traverse the tree and make sure any empty or whitespace nodes have a br so they can be selected
                traverse(el);
                //add an extra line at the bottom of the doc to make it easier to select a new line
                if(!isEmptyEditorNode(el.lastChild)) el.appendChild(emptyEditorElement.cloneNode(true));
        }

        /**
             * check if node is the empty editor node
             * @param {Node} node
             * @return {boolean}
         */
            function isEmptyEditorNode(node){
                return node.tagName === 'P' && node.childNodes.length === 1 && node.childNodes[0].tagName === 'BR';
            }
        }

        /**
         * Array helper to exec cleaning regexp
         * @param {string} html
         * @param {Object} cleaner
         * @return {string}
         */
        function execCleaner(html, cleaner){
            if(cleaner.rx) {
                return html.replace(cleaner.rx, cleaner.str).trim();
            }else if(cleaner.cleaner){
                return cleaners[cleaner.cleaner].reduce(execCleaner, html);
            }else if(cleaner.run){
                return cleaner.run(html);
            }else return html;
        }

        /**
         * remove an element, including its parent, if it is the only element within its parent
         * @param {HTMLElement} el
         */
        function removeWithParent (el) {
            if (el && el.parentNode) {
                if (el.parentNode.parentNode && el.parentNode.childElementCount === 1) {
                    el.parentNode.parentNode.removeChild(el.parentNode);
                } else {
                    el.parentNode.removeChild(el);
                }
            }
        }

        /**
         * remove any empty spans tags
         * @param {HTMLElement} containerEl
         */
        function cleanup (containerEl) {
            //remove any empty spans
            Array.from(containerEl.querySelectorAll('span:empty,p:empty')).forEach(function(el){el.remove();});
        }

        /**
         * Detect the type of pasted content
         * @param {string} html
         */
        function detectDocType(html){
            config.docType = '';
            html.replace(rxDetectGenerator, detectGenerator);
            if(!config.docType) {
                config.docType = rxDetectGDocs.test(html) ? 'GDocs' : rxDetectWord.test(html) ? 'MSWord' : 'None';
            }
            console.log('DOCTYPE ' + config.docType)
        }

        /**
         * Detect MSExcel/GSheets/OpenOffice
         * @param {string} $0
         * @param {string} $1
         */
        function detectGenerator($0,$1){
            config.docType = $1.startsWith('Microsoft Word') ? 'MSOffice' : $1.startsWith('Microsoft Excel') ? 'MSExcel' : $1.startsWith('Microsoft OneNote') ? 'MSOffice' : $1.startsWith('OpenOffice') ? 'OpenOffice' : $1.startsWith('Cocoa') ? 'Cocoa' : $1;
        }
        /**
         * Regex helper
         * Normalizes spacing in special Word `spacerun spans` (`<span style='mso-spacerun:yes'>\s+</span>`) by replacing
         * all spaces with `&nbsp; ` pairs. This prevents spaces from being removed during further DOM/View processing
         * Normalizes specific spacing generated by Safari when content pasted from Word (`<span class="Apple-converted-space"> </span>`)
         * by replacing all spaces sequences longer than 1 space with `&nbsp; ` pairs. This prevents spaces from being removed during
         * further DOM/View processing.
         * @param {string} $0
         * @param {string} $1
         * @return {string}
         */
        function normalizeSafariSpaceSpans( $0, $1) {
            var spaces = $1;
            return spaces.length === 1 ? ' ' : Array( spaces.length + 1 ).join( '\u00A0 ' ).substr( 0, spaces.length );
        }

        /**
         * take children and insert into target them remove this node
         * take node children and move to node parent and set node styles on parent (if parent has only 1 child)
         * take node children and move to node parent and set node styles on children
         * @param {Node} node
         * @param {string} copyStyles none parent child
         * @return {Node} return toNode for chaining
         */
        function unwrapElement(node, copyStyles) {
            // default to the element's parent node
            var toNode = node.parentNode;
            // move all children out of the element
            while (node.firstChild){
                // copy to child as parent has already got other children
                // exclude if they already exist in new parent except inherited props and exclude if child override nod
                if(copyStyles === 'child' && node.firstChild.nodeType === Node.ELEMENT_NODE) copyNodeStyles(node, node.firstChild, styleExcludes(node.firstChild, node, 'diff', 'inherit', styleExcludes(node, toNode, 'eq', 'inherit')));
                toNode.insertBefore(node.firstChild, node);
            }
            // copy node styles to parent
            // exclude inherited props- this will only happen when the parent has only node as a child
            if(copyStyles === 'parent') copyNodeStyles(node, toNode, styleExcludes(node, toNode, 'inherit'));
            // remove the empty element
            node.remove();
            //for chaining
            return toNode;
        }
        /**
         * take children and insert into target them remove this node
         * @param {Node} node
         * @param {string} copyStyles
         * @return {Node} return toNode for chaining
         */
        function unwrapSiblingElement(node, copyStyles) {
            // default to the element's parent node
            var toNode = node.previousSibling;
            // move all children out of the element
            while (node.firstChild){
                toNode.append(node.firstChild);
            }
            //copy node styles to sibling
            if(copyStyles === 'sibling') copyOtherNodeStyles(node, toNode, {});
            // remove the empty element
            node.remove();
            //for chaining
            return toNode;
        }

        /**
         * find any VML shapes in rtf
         * @param {string} rtf rtf string
         */
        function getImageData(rtf){
            //empty the image data
            config.imageData = {};
            //find image pictures
            findVMLShapes(rtf, "i", "\\shppict");
            //find shapes
            findVMLShapes(rtf, "s", "\\shp{");
        }

        /**
         * convert hex to base64
         * @param {string} hex
         * @return {string}
         */
        function hexToBase64(hex) {
            for (var t = hex.match(/[0-9a-f]{2}/gi), n = [], i = 0; i < t.length; i++) {
                n.push(String.fromCharCode(parseInt(t[i], 16)));
            }
            var o = n.join('');
            return btoa(o);
        }

        /**
         * parse and walk a rtf string find vml shape
         * @param {string} rtf rtf string
         * @param {string} t prefix type image  or shape (i or s)
         * @param {string} n separator
         */
        function findVMLShapes(rtf, t, n) {
            for (var i = rtf.split(n), o = 1; o < i.length; o++) {
                var r = i[o].split("shplid");
                if (r.length > 1) {
                    r = r[1];
                    for (var s = "", a = 0; a < r.length && "\\" !== r[a] && "{" !== r[a] && " " !== r[a] && "\r" !== r[a] && "\n" !== r[a]; ) {
                        s += r[a];
                        a++;
                    }
                    var l = r.split("bliptag");
                    if (l && l.length < 2)
                        continue;
                    var c = (l[0].indexOf("pngblip") !== -1) ? "image/png" : (l[0].indexOf("jpegblip") !== -1) ? (c = "image/jpeg") : null;
                    if (!c) continue;
                    var d = l[1].split("}");
                    if (d && d.length < 2) continue;
                    var buffer = void 0;
                    if (d.length > 2 && -1 !== d[0].indexOf("blipuid")) {
                        buffer = d[1].split(" ");
                    }else {
                        if ((buffer = d[0].split(" ")) && buffer.length < 2) continue;
                        buffer.shift()
                    }
                    var hex = buffer.join("");
                    var dataurl = "data:" + c + ";base64," + hexToBase64(hex);

                    config.imageData[t + s] = { hex: hex, type: c, dataurl:dataurl};
                }
            }
        }
        /**
         * Moodify the image src from MSWord where it is set to a local file link
         * We have to parse the rtf file to locate the base64 image data
         * @param node the img tag
         * @return {boolean}
         */
        function setImageSrc(node) {
            if (config.imageData) {
                if (node.tagName === 'IMG') {
                    var src = node.getAttribute("src");
                    //we are only converting local file references
                    if (!src || src.indexOf("file://") === -1) return true;
                    //check whether there is a valid alt url we could use
                    if ((src.indexOf("file://") === 0)  && isExtURL(node.getAttribute("alt"))){
                        //switch to the alt url and return
                        node.setAttribute("src", node.getAttribute("alt"));
                        return true;
                    }
                    //node.parentNode && node.parentNode.parentNode && node.parentNode.parentNode.innerHTML.indexOf("msEquation") >= 0 && (i = null))
                    node.removeAttribute("height");
                    //get the vs id
                    var vsID = node.getAttribute("v:shapes");
                    if (vsID) {
                        var shapeID = (mapVSID[vsID] || vsID || '').substring(7);
                        var imgInfo = config.imageData[shapeID];
                        if (imgInfo) {
                            node.src = imgInfo.dataurl;
                            return true;
                        }else{
                            return false;
                        }
                    }else return false;
                }else return true;
            }else return true;
        }
        /**
         * Check whether this is a valid link to an external resource
         * @param {string} url
         * @return {boolean}
         */
        function isExtURL(url) {
            //check the protocol
            var result = !!rxLinkProtocol.test(url);
            if(result) {
                //escape the url before we test it
                url = url.replace(/</g, '%3C').replace(/>/g, '%3E').replace(/"/g, '%22').replace(/ /g, '%20');
                //now test it
                result = rxLinkExtURL.test(url);
            }
            return result;
        }

        /**
         * cleanup pasted html text
         * @param {string} html
         * @param {string} rtf
         * @param {boolean} asFrag
         * @return {string|DocumentFragment}
         * @memberOf HTMLSanitizer#
         */
        function cleanPastedHTML(html, rtf, asFrag){
            var result;
            //empty the styles array and docType before we start
            config = {src:html, html:html, rtf:rtf, asFrag:asFrag, docType:'', stylesheet:'', styles:[], fonts:[], fixfonts:[], imageData:null, validClasses:['pasted-content', 'usr', 'ListContainerWrapper']};
            try{
                if(rtf){
                    getImageData(rtf);
                }
                //check what type of doc was pasted
                detectDocType(config.html);
                //patch up Desktop Excel
                if(config.docType==='MSExcel'){
                    //wrap desktop excel to match live
                    if(!config.html.startsWith('<html>'))config.html = excelStart + config.html + excelEnd;
                }
                //generic pre cleaners - normalize
                config.html = cleaners['commonPre'].reduce(execCleaner, config.html);
                //run main cleaners
                config.html = cleaners[config.docType].reduce(execCleaner, config.html);
                //run main generic cleaners
                config.html = cleaners['commonMain'].reduce(execCleaner, config.html);
                //insert any stylesheets

                config.html = config.stylesheet ? '<div class="pasted-content ' + config.pfx + '">' + config.stylesheet + config.html + '</div>': config.html;
                // create a temporary sandbox to cleanup elements
                const sandbox = document.implementation.createHTMLDocument('');
                const range = sandbox.createRange();
                //now create a frag, remove and selected content and insert the new markup this will already be normalized
                const frag = range.createContextualFragment(config.html);

                //fix links
                fixLinks(frag);
                //now sanitize - any ws between blocks get removed - comment tags removed - clean attr-remove blacklist-rollup p in li-collapse nested spans
                sanitize(sandbox, frag);
                //remove empty spans and paragraphs
                cleanup(frag);
                //wrap all root childnodes and children of a root blockquote in a p tag
                wrapChildNodes(frag);
                //make sure all empty nodes get a br so they can be edited.......
                traverse(frag);
                //return the frag or markup as requested finaly make sure any empty p tags get populated
                if(asFrag){
                    result = frag;
                }else{
                    var div = document.createElement("div");
                    div.appendChild(frag);
                    result = div.innerHTML;
                }
            } catch (e) {
               result = cleanPastedText(htmlToText(config.src));
            }
            config = null;
            return result;
        }

//endregion

//region TextToHTML

        //escape html entities helper
        var reET = /[&<>"']/g;
        var encodeHTMLRules = { "&": "&#38;", "<": "&#60;", ">": "&#62;", '"': '&#34;', "'": '&#39;', "/": '&#47;' };
        var rxURL = /(^|[^>"=])((https?|sftp|ftp|file|nfs|smb|afp|git):\/\/[\-A-Za-z0-9+&@#\/%?=~_|!:,.;]*[\-A-Za-z0-9+&@#\/%=~_|??])/g;
        var rxIsHTML = /(<\/|<IMG|<img|<BR|<br)/;

        /**
         * Internal Regex helper
         * @param $0
         * @returns {string}
         */
        function escapeTextHelper($0) {
            return encodeHTMLRules[$0];
        }
        /**
         * Escape text
         * @param v the string to escape
         * @returns {string}
         * @function external:String.escapeText
         */
       function escapeText(v) {
            return v ? v.toString().replace(reET, escapeTextHelper) : '';
        }
        /**
         * @param {string} str
         * @return {string}
         * @memberOf HTMLSanitizer#
         */
        function textToHTML(str){
            //escape all entities, replace tabs, collapse dblws, replace nl
            str = String.escapeText(str).replace(rxTab,'\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0\u00A0').replace(rxDblWS, '\u00A0').replace(rxNewLine,'<br>');
            str = str.replace(rxURL, '$1<a href="$2" target="_blank" rel="noopener noreferer">$2</a>');

            //finaly make sure any empty p tags get populated
            return str.replace(rxPTag,'<p><br></p>');
        }
        /**
         *
         * @param {string} text
         * @param {boolean} asFrag
         * @returns {DocumentFragment|string}
         * @memberOf HTMLSanitizer#
         */
        function cleanPastedText(text, asFrag){
            //is it html or text
            var isHTML = rxIsHTML.test(text);
            if(isHTML){
                return cleanPastedHTML(text, asFrag);
            }else {
                // create a temporary sandbox to cleanup elements
                const sandbox = document.implementation.createHTMLDocument('');
                const range = sandbox.createRange();
                //now create a frag, remove and selected content and insert the new markup this will already be normalized
                const frag = range.createContextualFragment(textToHTML(text));
                //return the frag or markup as requested finaly make sure any empty p tags get populated
                if (asFrag) {
                    return frag;
                } else {
                    var div = document.createElement("div");
                    div.appendChild(frag);
                    return div.innerHTML;
                }
            }
        }
//endregion

//region HTMLToText
        /**
         * Convert node to string - similar to calling innerText but innerText generates too many line feeds
         * @param {Node} node
         * @return {string}
         * @memberOf HTMLSanitizer#
         */
        function htmlToText(node) {
            var children = node.childNodes || [];
            var length = children.length;
            var block = tags[node.tagName].block;
            var whiteSpace = inferWhiteSpace(node, {});
            var index = -1;
            var results;
            var current;
            var result;
            var value;
            var count;

            // Treat "text" as having normal white-space.
            // All other nodes are handled as if they are "element"s
            // Nodes without children are treated as a void element and ignored
            if (node.nodeType === Node.TEXT_NODE) {
                return collectText(node, {
                    whiteSpace: whiteSpace,
                    breakBefore: true,
                    breakAfter: true
                })
            }

            // Let results be a new empty list.
            results = [];

            //For each child node node of this element:
            while (++index < length) {
                // Let current be the list resulting in running the inner text collection steps with node.
                // Each item in results will either be a string or a positive integer (a required line break count).
                current = innerTextCollection(children[index], index, node, {
                    whiteSpace: whiteSpace,
                    breakBefore: index === 0 ? block : false,
                    breakAfter: index === length - 1 ? block : children[index + 1].tagName=== 'BR'
                });

                // For each item item in current, append item to results.
                results = results.concat(current);
            }

            // Remove any items from results that are the empty string.
            // Remove any runs of consecutive required line break count items at the start or end of results.
            // Remove any runs of consecutive required breaks after an explicit break
            // Replace each remaining run of consecutive required line break count items with a string
            //  consisting of as many U+000A LINE FEED (LF) characters as the max of the values in the
            //  required line break count items.
            index = -1;
            length = results.length;
            result = [];
            var last;
            while (++index < length) {
                value = results[index];

                if (typeof value === 'number') {
                    if(last===lineFeedChar){
                        count = value - 1;
                    }else if (count !== undefined && value > count) {
                        count = value;
                    }
                } else if (value !== '') {
                    if (count) {// && value !== lineFeedChar
                        result.push(lineFeedChar.repeat(count));
                    }

                    count = 0;
                    result.push(value);
                    last = value;
                }
            }

            // Return the concatenation of the string items in results.
            return result.join('');
        }

        /**
         * collect the inner text
         * @param {Node} node
         * @param {number} index
         * @param {Node} parent
         * @param {Object}options
         * @return {(string|*)[] | *[]|string[]|Array}
         */
        function innerTextCollection(node, index, parent, options) {
            if (node.nodeType === Node.ELEMENT_NODE) {
                return collectElement(node, index, parent, options);
            }

            if (node.nodeType === Node.TEXT_NODE) {
                return [options.whiteSpace === 'normal' ? collectText(node, options) : collectPreText(node)];
            }

            return [];
        }

        /**
         * Collect an element
         * @param {Node} node
         * @param {number} index
         * @param {Node} parent
         * @param {Object}options
         * @return {Array}
         */
        function collectElement(node, index, parent, options) {
            // First we infer the `white-space` property.
            var whiteSpace = inferWhiteSpace(node, options);
            var children = node.childNodes || [];
            var length = children.length;
            var offset = -1;
            var items = [];
            var current;
            var prefix;
            var suffix;

            // Note: we first detect if there is going to be a break before or after the
            // contents, as that changes the white-space handling.

            // If node is a BR element, then append a string containing a single
            // U+000A LINE FEED (LF) character to items.
            if (node.tagName === 'BR') {
                suffix = lineFeedChar;
            }

            //   If node’s computed value of `display` is `table-row`, and node’s CSS
            //   box is not the last `table-row` box of the nearest ancestor `table`
            //   box, then append a string containing a single U+000A LINE FEED (LF)
            //   character to items.
            else if (row(node) && row (node.nextElementSibiling)) {
                suffix = lineFeedChar;
            }

            //   If node’s used value of `display` is block-level or `table-caption`,
            //  then append 1 (a required line break count) at the beginning and end of items.
            else if (tags[node.tagName].block) {
                prefix = 1;
                suffix = 1;
            }

            //    Let items be the result of running the inner text collection steps with
            //     each child node of node in tree order, and then concatenating the
            //     results to a single list.
            while (++offset < length) {
                current = innerTextCollection(children[offset], offset, node, {
                    whiteSpace: whiteSpace,
                    breakBefore: offset === 0 ? prefix : false,
                    breakAfter: offset === length - 1 ? suffix : children[offset + 1].tagName === 'BR'
                });

                items = items.concat(current);
            }

            //     If node’s computed value of `display` is `table-cell`, and node’s CSS
            //     box is not the last `table-cell` box of its enclosing `table-row` box,
            //     then append a string containing a single U+0009 CHARACTER TABULATION
            //     (tab) character to items.
            if (cell(node) && cell(node.nextElementSibiling)) {
                items.push(tabChar);
            }

            // Add the pre- and suffix.
            if (prefix) {
                items.unshift(prefix);
            }

            if (suffix) {
                items.push(suffix);
            }

            return items;
        }

        // 4.  If node is a Text node, then for each CSS text box produced by node,
        //     in content order, compute the text of the box after application of the
        //     CSS `white-space` processing rules and `text-transform` rules, set
        //     items to the list of the resulting strings, and return items.
        //     The CSS `white-space` processing rules are slightly modified:
        //     collapsible spaces at the end of lines are always collapsed, but they
        //     are only removed if the line is the last line of the block, or it ends
        //     with a br element.
        //     Soft hyphens should be preserved.
        //
        //     Note: See `collectText` and `collectPreText`.
        //     Note: we don’t deal with `text-transform`, no element has that by
        //     default.
        //
        // See: <https://drafts.csswg.org/css-text/#white-space-phase-1>
        function collectText(node, options) {
            var breakBefore = options.breakBefore;
            var breakAfter = options.breakAfter;
            var value = String(node.data);
            var index = -1;
            var length = value.length;
            var lines = [];
            var result = [];
            var lineStart;
            var lineEnd;
            var line;
            var nextLine;
            var queue;

            lineStart = 0;
            lineEnd = value.indexOf(lineFeedChar);
            lineEnd = lineEnd === -1 ? value.length : lineEnd;

            while (lineEnd !== -1) {
                line = value.slice(lineStart, lineEnd);

                // Any sequence of collapsible spaces and tabs immediately preceding or
                // following a segment break is removed.
                line = trimAndcollapseSpacesAndTabs(line, breakBefore, breakAfter);

                // Add the line.
                lines.push(line);

                // Stop.
                if (lineEnd === value.length) {
                    break
                }

                lineStart = lineEnd + 1;
                lineEnd = value.indexOf(lineFeedChar, lineStart);
                lineEnd = lineEnd === -1 ? value.length : lineEnd;
            }

            index = -1;
            length = lines.length;
            queue = '';

            // Collapsible segment breaks are transformed for rendering according to the
            // segment break transformation rules.
            // So here we jump to 4.1.2 of [CSSTEXT]:
            // Any collapsible segment break immediately following another collapsible
            // segment break is removed
            while (++index < length) {
                line = lines[index];
                nextLine = lines[index + 1] || '';

                // *   If the character immediately before or immediately after the segment
                //     break is the zero-width space character (U+200B), then the break is
                //     removed, leaving behind the zero-width space.
                if (
                    line.charCodeAt(line.length - 1) === zeroWidthSpace ||
                    nextLine.charCodeAt(0) === zeroWidthSpace
                ) {
                    result.push(line);
                    queue = '';
                    continue;
                }

                // *   Otherwise, the segment break is converted to a space (U+0020).
                if (line) {
                    if (queue) {
                        result.push(queue);
                    }

                    result.push(line);
                    queue = spaceChar;
                }
            }

            return result.join('');
        }

        /**
         *
         * @param {Node} node
         * @return {string}
         */
        function collectPreText(node) {
            return String(node.data);
        }


        //   Every collapsible tab is converted to a collapsible space (U+0020).
        //   Any collapsible space immediately following another collapsible
        //   space—even one outside the boundary of the inline containing that
        //   space, provided both spaces are within the same inline formatting
        //   context—is collapsed to have zero advance width. (It is invisible,
        //   but retains its soft wrap opportunity, if any.)
        function trimAndcollapseSpacesAndTabs(value, breakBefore, breakAfter) {
            var start = 0;
            var end;
            var length = value.length;
            var result = [];
            var char;

            function next(index) {
                var spaceIndex = value.indexOf(spaceChar, index + 1);
                var tabIndex = value.indexOf(tabChar, index + 1);
                var fn = spaceIndex === -1 || tabIndex === -1 ? Math.max : Math.min;
                return fn(spaceIndex, tabIndex);
            }

            // Move forward past initial white space.
            while (start <= length) {
                char = value.charCodeAt(start);

                if (char !== space && char !== tab) {
                    break;
                }

                start++;
            }

            // If we’re not directly after a segment break, but there was white space,
            // add an empty value that will be turned into a space.
            if (start !== 0 && !breakBefore) {
                result.push('');
            }

            end = next(start - 1);

            while (start < length) {
                end = end === -1 ? length : end;
                result.push(value.slice(start, end));
                start = end;

                while (start <= length) {
                    char = value.charCodeAt(start);

                    if (char !== space && char !== tab) {
                        break;
                    }
                    start++;
                }

                // If we reached the end, there was trailing white space, and there’s no
                // segment break after this node, add an empty value that will be turned
                // into a space.
                if (start === length && start !== end && !breakAfter) {
                    result.push('');
                }

                end = next(start);
            }

            return result.join(' ');

        }

        /**
         * get whitespace mode for this node
         * @param {Node} node
         * @param {Object} options
         * @return {string}
         */
        function inferWhiteSpace(node, options) {
            var inherit = options.whiteSpace || 'normal';
            switch (node.tagName) {
                case 'LISTING':
                case 'PLAINTEXT':
                case 'XMP':
                    return 'pre';
                case 'NOBR':
                    return 'nowrap';
                case 'PRE':
                    var whitespace = getComputedStyle(node).whiteSpace;
                    return whitespace === 'wrap' ? 'pre-wrap' : 'pre';
                case 'TD':
                case 'TH':
                    var whitespace = getComputedStyle(node).whiteSpace;
                    return whitespace === 'nowrap' ? 'nowrap' : inherit;
                case 'textarea':
                    return 'pre-wrap';
                default:
                    return inherit;
            }
        }

        /**
         * is the node a table cell
         * @param {Node} node
         * @return {boolean}
         */
        function cell(node) {
            return node ? (node.tagName === 'TH' || node.tagName === 'TD') : false;
        }

        /**
         * Is the node a table row
         * @param {Node} node
         * @return {boolean}
         */
        function row(node) {
            return node ? node.tagName === 'TR' : false;
        }
//endregion
        self.cleanPastedText = cleanPastedText;
        self.cleanPastedHTML = cleanPastedHTML;
        self.normalize = normalize;
        self.textToHTML = textToHTML;
        self.htmlToText = htmlToText;
        return self;
    };
    /**@type HTMLSanitizer*/
    window.HTMLSanitizer = new HTMLSanitizer();

//endregion
})(jQuery);